
/* DWARF debugging format support for GDB.
   Copyright (C) 1991, 1992, 1993, 1994 Free Software Foundation, Inc.
   Written by Fred Fish at Cygnus Support.  Portions based on dbxread.c,
   mipsread.c, coffread.c, and dwarfread.c from a Data General SVR4 gdb port.

This file is part of GDB.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */

/*

  Important Note: This file is an adaptation of the original gdb dwarfread.c
  file which supported dwarf version 1.  This file supports version 2 using
  an SGI supplied libdwarf library.  Because of the interface change brought
  on by using the library, and because of changes in the format, this file
  is substantially changed from the original.  It was written by someone with
  little knowledge of dwarf or gdb, so it is highly non-optimal and somewhat
  convoluted due to its history.  In addition, all comments should be taken
  with a large grain of salt as they may be left over from the original and
  no longer apply.


  One open question is whether it would be better to make use of the types
  and globals sections to collect certain key information.

  Note: need to check for places where dwarf_dealloc should be called but
  isn't (for all child and sibling dies in particular).

  Note: we use sym->global is a flag for typecasts since they don't follow
  the normal conventions for our types, since they are constructed on the
  fly and don't have a proper dwarf entry to anchor them.
  
*/

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#ifdef sgi
#include <dem.h>
#endif
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <sys/file.h>

#include "file_formats.h"
#include "symbols.h"
#include "simutil.h"
#include "symfile.h"
#include "assoctab.h"
#include "tcl_init.h"
#include "sim_error.h"

    /* we're not depending on elf library, so provide some dummy defn */
typedef void * Elf;
#include "dwarf2.h"
#include "libdwarf2.h"


#define DIE_NAME(dip)	(((dip)->at_name != NULL) ? (dip)->at_name : "")

#define IS_CAST(sym)	(sym->global != 0)


int info_verbose;
   
typedef Dwarf_Off DIE_REF;	/* Reference to a DIE */

typedef char BLOCK;

struct dieinfo {
  Dwarf_Die             ddie;           /* pointer to dwarf library die data */
  Dwarf_Die             sibdie;         /* pointer to dwarf library die data */
  Dwarf_Die             childdie;       /* pointer to dwarf library die data */
  DIE_REF		die_ref;	/* Offset of this DIE in debug sect */
  Dwarf_Half	        die_tag;	/* Tag for this DIE */
  char *		at_name;
  Dwarf_Off		at_type;
};


typedef struct Execfile {
    ATHeader     head;
    AssocTab    *srcs;
    AssocTab    *globals;
    AssocTab    *gStructs;
    Dwarf_Debug  dwarf;
} Execfile;

typedef struct Srcfile {
    ATHeader     head;
    AssocTab    *globals;
    VA           lopc;
    VA           hipc;
    DIE_REF      dieref;
} Srcfile;

typedef struct Global {
    ATHeader   head;
    DIE_REF    dieref;
    Dwarf_Half dietag;
} Global;



static void scan_partial_symbols(Dwarf_Die, Execfile *exec, Srcfile *src );
static void scan_compilation_units(Execfile *exec);
static void basicdieinfo ( struct dieinfo *dip, Dwarf_Die ddie,
			   DIE_REF dieref, Execfile *exec );
static void freedieinfo( struct dieinfo *di, Execfile *exec );
static void completedieinfo(struct dieinfo *, Execfile *exec, int verbose);
static void decode_line_numbers(Dwarf_Die, VA *, VA *, Execfile *exec);
static VA locval (Dwarf_Locdesc *);


static VA
lookup_line ( Srcfile *src, Execfile *exec, Dwarf_Unsigned targetline )
{
    struct dieinfo  di;
    int             err;
    Dwarf_Line     *linebuf;
    Dwarf_Unsigned  lineno;
    Dwarf_Unsigned  bestlineno;
    Dwarf_Addr      lineaddr;
    Dwarf_Addr      bestlineaddr;
    Dwarf_Signed    cnt;
    Dwarf_Error     rc;
    int             i;

    bestlineno = ~0;
    bestlineaddr = 0;
    
    basicdieinfo (&di, NULL, src->dieref, exec);
    err = dwarf_srclines( di.ddie, &linebuf, &cnt, &rc );
    freedieinfo(&di,exec);

    if( err == DW_DLV_ERROR ) {
	CPUWarning("dwarf_srclines failed: %s\n", dwarf_errmsg( rc ) );
	ASSERT(0);
    } else if( err == DW_DLV_OK ) {

	for( i = 0; i < cnt; i++ ) {

	    if( dwarf_lineno( linebuf[i], &lineno, &rc ) != DW_DLV_OK) {
		CPUWarning("dwarf_lineno failed: %s\n", dwarf_errmsg( rc ) );
		exit(-1);
	    }
	    if( dwarf_lineaddr( linebuf[i], &lineaddr, &rc) != DW_DLV_OK) {
		CPUWarning("dwarf_lineaddr failed: %s\n", dwarf_errmsg( rc ) );
		exit(-1);
	    }

	    if( (lineno <= bestlineno) && (lineno >= targetline) ) {

		/* It doesn't seem like there is an easy way to deal with
		 * line numbers of other files included in this file, such
		 * as inlined source in .h's and that sort of thing.  The
		 * best thing may be just to ignore line number info for
		 * file other than our own
		 */
		char *name;
		if( dwarf_linesrc( linebuf[i], &name, &rc) != DW_DLV_OK) {
		    CPUWarning("dwarf_linesrc failed: %s\n", dwarf_errmsg(rc));
		    exit(-1);
		}
		if( strcmp( src->head.label, StripPath(name) ) == 0 ) {
		    bestlineno = lineno;
		    bestlineaddr = lineaddr;
		}
		dwarf_dealloc( exec->dwarf, name, DW_DLA_STRING );
	    }

	    dwarf_dealloc( exec->dwarf, linebuf[i], DW_DLA_LINE );
	}
	dwarf_dealloc( exec->dwarf, linebuf, DW_DLA_LIST );
    }

    return bestlineaddr;
}

static void
decode_line_numbers ( Dwarf_Die ddie, VA *lopc, VA *hipc, Execfile *exec )
{
    int             err;
    Dwarf_Line     *linebuf;
    Dwarf_Unsigned  lineno;
    Dwarf_Addr      lineaddr;
    Dwarf_Signed    cnt;
    Dwarf_Error     rc;
    int             i;
    VA       lo, hi;

    hi = 0;
    lo = ~hi;

    err = dwarf_srclines( ddie, &linebuf, &cnt, &rc );

    if( err == DW_DLV_ERROR ) {
	CPUWarning("dwarf_srclines failed: %s\n", dwarf_errmsg( rc ) );
    } else if( err == DW_DLV_OK ) {

	for( i = 0; i < cnt; i++ ) {

	    if( dwarf_lineno( linebuf[i], &lineno, &rc ) != DW_DLV_OK) {
		CPUWarning("dwarf_lineno failed: %s\n", dwarf_errmsg( rc ) );
		exit(-1);
	    }
	    if( dwarf_lineaddr( linebuf[i], &lineaddr, &rc) != DW_DLV_OK) {
		CPUWarning("dwarf_lineaddr failed: %s\n", dwarf_errmsg( rc ) );
		exit(-1);
	    }

	    /* double purpose function: either record line numbers, or find low
	     * and high pc values.  We can't do both because we need pc values
	     * before it appears safe to record line numbers.
	     */
	    if( lopc == NULL ) {
		/* we don't use this feature yet */
		ASSERT(0);
#if 0
		/* It doesn't seem like there is an easy way to deal with
		 * line numbers of other files included in this file, such
		 * as inlined source in .h's and that sort of thing.  The
		 * best thing may be just to ignore line number info for
		 * file other than our own
		 */
		char *name;
		if( dwarf_linesrc( linebuf[i], &name, &rc) != DW_DLV_OK) {
		    CPUWarning("dwarf_linesrc failed: %s\n", dwarf_errmsg(rc));
		    exit(-1);
		}
		if( !strcmp( current_subfile->name, name ) ) {
		    record_line (current_subfile, lineno, lineaddr);
		}
		dwarf_dealloc( dwarf, name, DW_DLA_STRING );
#endif
	    } else {
		if( lineaddr < lo )
		    lo = lineaddr;
		if( lineaddr > hi )
		    hi = lineaddr;
	    }

	    dwarf_dealloc( exec->dwarf, linebuf[i], DW_DLA_LINE );
	}
  
	dwarf_dealloc( exec->dwarf, linebuf, DW_DLA_LIST );
    }

    if( lopc != NULL )
	*lopc = lo;
    if( hipc != NULL )
	*hipc = hi;
}

static VA
locval ( Dwarf_Locdesc *loc )
{
    VA  stack[64];
    int        stacki;
    int        loc_atom_code;
    int        i;
    Dwarf_Loc *locs;
    int        numlocs;

    stacki = 0;
    stack[stacki] = 0;
/*    isreg = 0; */
/*    offreg = 0; */
    locs = loc->ld_s;
    numlocs = loc->ld_cents;
    for( i = 0; i < numlocs; i++ ) {
	loc_atom_code = locs[i].lr_atom;
	switch (loc_atom_code) {
	case DW_OP_addr:
	    /* push address (relocated address) */
	    stack[++stacki] = locs[i].lr_number;
	    break;
	case DW_OP_consts:
	    stack[++stacki] = locs[i].lr_number;
	    break;
	case DW_OP_fbreg:
#if 1
#if 0
	    offreg = 1;
	    basereg = 29; /* FIXME: MIPS stack pointer hard coded */
#endif
#endif
	    stack[++stacki] = locs[i].lr_number;
	    break;
	default:
	    CPUWarning("Unimplemented location atom: %x\n", loc_atom_code);
	    break;
	}
    }
    return (stack[stacki]);
}

static DIE_REF
GetTypeAbsOffset( Dwarf_Die ddie, Execfile *exec )
{
    Dwarf_Attribute  attr;
    Dwarf_Off        type, abs, rel;
    Dwarf_Error      rc;

    if (dwarf_attr(ddie,DW_AT_type,&attr,&rc) != DW_DLV_OK) {
	return 0;
    }
    if( dwarf_formref( attr, &type, &rc ) != DW_DLV_OK) {
	CPUWarning("dwarf_formref failed: %s\n", dwarf_errmsg( rc ) );
	ASSERT(0);
    }
    dwarf_dealloc( exec->dwarf, attr, DW_DLA_ATTR);
    if( dwarf_dieoffset( ddie, &abs, &rc ) != DW_DLV_OK) {
	CPUWarning("dwarf_dieoffset failed: %s\n", dwarf_errmsg( rc ) );
	ASSERT(0);
    }
    if( dwarf_die_CU_offset( ddie, &rel, &rc ) != DW_DLV_OK) {
	CPUWarning("dwarf_die_CU_offset failed: %s\n", dwarf_errmsg( rc ) );
	ASSERT(0);
    }

    /* convert from cu-relative to absolute offset */
    return (abs - rel) + type;

}

static VA
GetLocation( Dwarf_Die ddie, Execfile *exec )
{
    Dwarf_Error      rc;
    int              err;
    Dwarf_Locdesc   *loclist;
    Dwarf_Signed     signedtmp;
    Dwarf_Attribute  attr;
    VA               location;

    if (dwarf_attr(ddie,DW_AT_location,&attr,&rc) != DW_DLV_OK) {
	/* try member location */
	if (dwarf_attr(ddie,DW_AT_data_member_location,&attr,&rc)!=DW_DLV_OK) {
	    CPUWarning("no location list: %s\n", dwarf_errmsg( rc ) );
	    ASSERT(0);
	}
    }

    dwarf_loclist( attr, &loclist, &signedtmp, &rc );
    if( signedtmp != 1 ) {
	CPUWarning("location list size %lld != 1\n", signedtmp );
	ASSERT(0);
    }

    location = locval(loclist);

    dwarf_dealloc( exec->dwarf, loclist, DW_DLA_LOCDESC);
    dwarf_dealloc( exec->dwarf, attr, DW_DLA_ATTR);

    return location;
}

static int
add_global(struct dieinfo *di, AssocTab *tab)
{
    Global *global;

    global = (Global*)malloc(sizeof(Global));
    global->head.label = SaveString(di->at_name);
    global->dieref     = di->die_ref;
    global->dietag     = di->die_tag;

    return AssocTabEnter(tab, (ATHeader*)global);
}


/* release resources associated with die; not all are currently handled */
static void
freedieinfo( struct dieinfo *di, Execfile *exec )
{
    if( di->ddie ) dwarf_dealloc( exec->dwarf, di->ddie, DW_DLA_DIE );
    if( di->sibdie ) dwarf_dealloc( exec->dwarf, di->sibdie, DW_DLA_DIE );
    if( di->childdie ) dwarf_dealloc( exec->dwarf, di->childdie, DW_DLA_DIE );
    if( di->at_name ) dwarf_dealloc( exec->dwarf, di->at_name, DW_DLA_STRING );
}

static void
FillInBasicType( Symb *sym, Dwarf_Die ddie, Dwarf_Unsigned *bitsize,
		 Execfile *exec )
{
    Dwarf_Attribute  attr;
    Dwarf_Unsigned   encoding, bytesize;
    Dwarf_Off        type;
    Dwarf_Error      rc;
    struct dieinfo   di;

    type = GetTypeAbsOffset( ddie, exec );
    if( type == 0 ) {
	/* no type, so do nothing */
	return;
    }

    basicdieinfo (&di, NULL, type, exec);
    ddie = di.ddie;

    switch (di.die_tag) {
    case DW_TAG_base_type:
	/* for this we need the size and encoding to translate it to a
	 * form that simos expects
	 */
	if( dwarf_bytesize( ddie, &bytesize, &rc ) != DW_DLV_OK) {
	    CPUWarning("dwarf_bytesize failed: %s\n", dwarf_errmsg( rc ) );
	    ASSERT(0);
	}
	*bitsize = 8*bytesize;
	if (dwarf_attr(ddie,DW_AT_encoding,&attr,&rc) != DW_DLV_OK) {
	    CPUWarning("dwarf_attr for type failed: %s\n", dwarf_errmsg(rc));
	    ASSERT(0);
	}
	if( dwarf_formudata( attr, &encoding, &rc ) != DW_DLV_OK) {
	    CPUWarning("dwarf_formref failed: %s\n", dwarf_errmsg( rc ) );
	    ASSERT(0);
	}
	dwarf_dealloc( exec->dwarf, attr, DW_DLA_ATTR);
	switch( encoding ) {
	case DW_ATE_float:
	    switch( bytesize ) {
	    case 4:
		sym->type[sym->nextt++] = tFloat;
		break;
	    case 8:
		sym->type[sym->nextt++] = tDouble;
		break;
	    default:
		ASSERT(0);
		break;
	    }
	    break;

	case DW_ATE_signed:
	case DW_ATE_signed_char:
	    switch( bytesize ) {
	    case 1:
		sym->type[sym->nextt++] = tChar;
		break;
	    case 2:
		sym->type[sym->nextt++] = tShort;
		break;
	    case 4:
		sym->type[sym->nextt++] = tInt;
		break;
	    case 8:
		sym->type[sym->nextt++] = tLong;
		break;
	    default:
		ASSERT(0);
		break;
	    }
	    break;

	case DW_ATE_boolean:
	    /* we treat as a general unsigned value */
	case DW_ATE_unsigned:
	case DW_ATE_unsigned_char:
	    switch( bytesize ) {
	    case 1:
		sym->type[sym->nextt++] = tUChar;
		break;
	    case 2:
		sym->type[sym->nextt++] = tUShort;
		break;
	    case 4:
		sym->type[sym->nextt++] = tUInt;
		break;
	    case 8:
		sym->type[sym->nextt++] = tULong;
		break;
	    default:
		ASSERT(0);
		break;
	    }
	    break;

	case DW_ATE_complex_float:
	case DW_ATE_address:		/* not sure what this is */
	default:
	    ASSERT(0);
	    break;
	}
	break;

    case DW_TAG_array_type:
	sym->type[sym->nextt++] = tArray;
        FillInBasicType(sym, di.ddie, bitsize, exec);
	*bitsize = 0;
	break;
	    
    case DW_TAG_entry_point:
	ASSERT(0);
	break;

    case DW_TAG_enumeration_type:
	sym->type[sym->nextt++] = tEnum;
	if( dwarf_bytesize( ddie, &bytesize, &rc ) != DW_DLV_OK) {
	    CPUWarning("dwarf_bytesize failed: %s\n", dwarf_errmsg( rc ) );
	    ASSERT(0);
	}
	*bitsize = 8*bytesize;
	break;
	
    case DW_TAG_imported_declaration:
    case DW_TAG_lexical_block:
    case DW_TAG_member:
    case DW_TAG_variable:
	ASSERT(0);
	break;

    case DW_TAG_label:
	sym->type[sym->nextt++] = tLabel;
	*bitsize = 0;
	break;
	
    case DW_TAG_pointer_type:
	sym->type[sym->nextt++] = tPtr;
	if( dwarf_bytesize( ddie, &bytesize, &rc ) != DW_DLV_OK) {
	    CPUWarning("dwarf_bytesize failed: %s\n", dwarf_errmsg( rc ) );
	    ASSERT(0);
	}
	*bitsize = 8*bytesize;
	break;

    case DW_TAG_reference_type:
    case DW_TAG_compile_unit:
    case DW_TAG_string_type:
    case DW_TAG_formal_parameter:
	ASSERT(0);
	break;

    case DW_TAG_class_type:
    case DW_TAG_structure_type:
	sym->type[sym->nextt++] = tStruct;
	if( dwarf_bytesize( ddie, &bytesize, &rc ) != DW_DLV_OK) {
	    CPUWarning("dwarf_bytesize failed: %s\n", dwarf_errmsg( rc ) );
	    ASSERT(0);
	}
	*bitsize = 8*bytesize;
	break;

    case DW_TAG_subroutine_type:
    case DW_TAG_inlined_subroutine:
    case DW_TAG_subprogram:
	sym->type[sym->nextt++] = tPtr;
	*bitsize = 0;
	break;

    case DW_TAG_typedef:
#ifdef notdef
	sym->type[sym->nextt++] = tTypedef;
	*bitsize = 0;
#endif
        FillInBasicType(sym, di.ddie, bitsize, exec);

	break;

    case DW_TAG_union_type:
	sym->type[sym->nextt++] = tUnion;
	/* we possibly should recurse to get the bitsize */
	*bitsize = 0;
	break;

    case DW_TAG_unspecified_parameters:
    case DW_TAG_variant:
    case DW_TAG_common_block:
    case DW_TAG_common_inclusion:
    case DW_TAG_inheritance:
    case DW_TAG_module:
    case DW_TAG_ptr_to_member_type:
    case DW_TAG_set_type:
    case DW_TAG_subrange_type:
    case DW_TAG_with_stmt:
    case DW_TAG_access_declaration:
    case DW_TAG_catch_block:
    case DW_TAG_const_type:
    case DW_TAG_constant:
    case DW_TAG_enumerator:
    case DW_TAG_file_type:
    case DW_TAG_friend:
    case DW_TAG_namelist:
    case DW_TAG_namelist_item:
    case DW_TAG_packed_type:
    case DW_TAG_template_type_param:
    case DW_TAG_template_value_param:
    case DW_TAG_thrown_type:
    case DW_TAG_try_block:
    case DW_TAG_variant_part:
    case DW_TAG_volatile_type:
    case DW_TAG_lo_user:
    case DW_TAG_MIPS_loop:
    case DW_TAG_hi_user:
	ASSERT(0);
	break;

    default:
	ASSERT(0);
	break;
    }
    
    freedieinfo(&di,exec);
}

static int
FillInSym( Symb *sym, DIE_REF dieref, Execfile *exec )
{
    Dwarf_Error    rc;
    struct dieinfo di;
    int            err;
    Dwarf_Addr     atmp;
    Dwarf_Locdesc *loclist;
    Dwarf_Signed   signedtmp;
    Dwarf_Unsigned bitsize;
    Dwarf_Off      type;

    
    basicdieinfo (&di, NULL, dieref, exec);
	
    /* we use the isym field for the dieref field, even though it's not
     * strictly speaking big enough.
     */
    sym->global	= 0;
    sym->isym	= (int)dieref;
    sym->exec	= exec;
    sym->nextt	= 1;
    sym->value	= 0xbeeebaaabeeebaaaLL;	/* to help catch errors */
    sym->index	= DWARF_INDEX;		/* special tornado marker */

    switch (di.die_tag) {
    case DW_TAG_entry_point:
	ASSERT(0);
	break;

    case DW_TAG_array_type:
        type = GetTypeAbsOffset(di.ddie, exec);
        ASSERT(type != 0);
        freedieinfo(&di,exec);
        return FillInSym(sym, type, exec);
	
    case DW_TAG_enumeration_type:
	sym->type[sym->nextt++] = tEnumDef;
	break;
	
    case DW_TAG_imported_declaration:
	ASSERT(0);
	break;

    case DW_TAG_lexical_block:
	sym->type[sym->nextt++] = tBlock;
	break;

    case DW_TAG_variable:
	FillInBasicType(sym, di.ddie, &bitsize, exec);
	sym->value = GetLocation(di.ddie,exec);
	break;

    case DW_TAG_member:
	FillInBasicType(sym, di.ddie, &bitsize, exec);
	sym->value = 8*GetLocation(di.ddie,exec); /* offset in structure */
	break;

    case DW_TAG_pointer_type:
    {
	Dwarf_Unsigned dummy;
	sym->type[sym->nextt++] = tPtr;
 	FillInBasicType( sym, di.ddie, &dummy, exec );
	break;
    }

    case DW_TAG_reference_type:
    case DW_TAG_compile_unit:
    case DW_TAG_string_type:
    case DW_TAG_formal_parameter:
	ASSERT(0);
	break;

    case DW_TAG_class_type:
    case DW_TAG_structure_type:
	sym->type[sym->nextt++] = tStructDef;
	if( dwarf_bytesize( di.ddie, &bitsize, &rc ) != DW_DLV_OK) {
	    CPUWarning("dwarf_bytesize failed: %s\n", dwarf_errmsg( rc ) );
	    ASSERT(0);
	}
	sym->value = bitsize;		/* ok, really bytesize here */
	break;

    case DW_TAG_label:
	sym->type[sym->nextt++] = tLabel;
	if( dwarf_lowpc( di.ddie, &atmp, &rc ) != DW_DLV_OK) {
	    CPUWarning("dwarf_lowpc failed: %s\n", dwarf_errmsg( rc ) );
	    ASSERT(0);
	}
	sym->value = atmp;
	break;
	
    case DW_TAG_subroutine_type:
    case DW_TAG_inlined_subroutine:
    case DW_TAG_subprogram:
	sym->type[sym->nextt++] = tProc;
	if( dwarf_lowpc( di.ddie, &atmp, &rc ) != DW_DLV_OK) {
	    CPUWarning("dwarf_lowpc failed: %s\n", dwarf_errmsg( rc ) );
	    ASSERT(0);
	}
	sym->value = atmp;
#ifdef TORNADO
	/* problems with constructor return values (refs) so just ignore
	 * for now, as I don't think we really care
	 */
#else
	FillInBasicType(sym, di.ddie, &bitsize, exec);
#endif
	break;

    case DW_TAG_typedef:
#ifdef notdef
	sym->type[sym->nextt++] = tTypedefDef;
	FillInBasicType(sym, di.ddie, &bitsize, exec);
#endif
        type = GetTypeAbsOffset(di.ddie, exec);
        ASSERT(type != 0);
        freedieinfo(&di,exec);
        return FillInSym(sym, type, exec);

    case DW_TAG_union_type:
	sym->type[sym->nextt++] = tUnionDef;
	break;

    case DW_TAG_unspecified_parameters:
    case DW_TAG_variant:
    case DW_TAG_common_block:
    case DW_TAG_common_inclusion:
    case DW_TAG_inheritance:
    case DW_TAG_module:
    case DW_TAG_ptr_to_member_type:
    case DW_TAG_set_type:
    case DW_TAG_subrange_type:
    case DW_TAG_with_stmt:
    case DW_TAG_access_declaration:
    case DW_TAG_base_type:
    case DW_TAG_catch_block:
    case DW_TAG_const_type:
    case DW_TAG_constant:
    case DW_TAG_enumerator:
    case DW_TAG_file_type:
    case DW_TAG_friend:
    case DW_TAG_namelist:
    case DW_TAG_namelist_item:
    case DW_TAG_packed_type:
    case DW_TAG_template_type_param:
    case DW_TAG_template_value_param:
    case DW_TAG_thrown_type:
    case DW_TAG_try_block:
    case DW_TAG_variant_part:
    case DW_TAG_volatile_type:
    case DW_TAG_lo_user:
    case DW_TAG_MIPS_loop:
    case DW_TAG_hi_user:
	ASSERT(0);
	break;

    default:
	ASSERT(0);
	break;
    }

    ASSERT(sym->nextt<MAX_TYPES);
    sym->type[sym->nextt] = tNil;
    sym->nextt = 1;

    freedieinfo(&di,exec);

    return SYM_OK;
}


static void
scan_partial_symbols ( Dwarf_Die ddie, Execfile *exec, Srcfile *src )
{
    char            *nextdie;
    char            *temp;
    struct dieinfo   di;
    Dwarf_Die        nextddie;
    Dwarf_Bool       btmp1, btmp2;
    Dwarf_Error      rc;
    
    while (ddie) {
	basicdieinfo (&di, ddie, 0, exec);

	nextddie = di.sibdie;
	di.sibdie = NULL;		/* ensure not freed */

	switch (di.die_tag) {
	case DW_TAG_subprogram:
	case DW_TAG_variable:
	    dwarf_hasattr(ddie,DW_AT_low_pc,&btmp1,&rc);
	    dwarf_hasattr(ddie,DW_AT_location,&btmp2,&rc);
	    if ( di.at_name && (btmp1 || btmp2) ) {
		(void)add_global(&di, exec->globals);
		(void)add_global(&di, src->globals);
	    }
	    break;
	case DW_TAG_typedef:
	    if (di.at_name) {
		(void)add_global(&di, exec->globals);
		(void)add_global(&di, src->globals);
	    }
	    break;
	case DW_TAG_class_type:
	case DW_TAG_structure_type:
	case DW_TAG_union_type:
	case DW_TAG_enumeration_type:
	    if (di.at_name) {
		(void)add_global(&di, exec->gStructs);
		(void)add_global(&di, src->globals);
	    }
	    break;
	}

	freedieinfo(&di,exec);
	ddie = nextddie;
    }
}


static void
scan_compilation_units (Execfile *exec)
{
    struct dieinfo di;
    int            culength;
    int            curoff;
    Dwarf_Error    rc;
    Dwarf_Unsigned length;
    Dwarf_Half     version;
    Dwarf_Off      offset;
    Dwarf_Half     addrsize;
    Dwarf_Unsigned next;
    Dwarf_Die      ddie;
    Dwarf_Die      nextddie;
    Dwarf_Half     tagval;
    int            err;
    int            fd;
    Srcfile       *src;
    VA             low_pc;
    VA             high_pc;

    
    while (1) {

	err = dwarf_next_cu_header( exec->dwarf, &length, &version, &offset,
				    &addrsize, &next, &rc );
	if( err == DW_DLV_NO_ENTRY ) {
	    break;
	} else if( err != DW_DLV_OK) {
	    CPUWarning("dwarf_next_cu_header failed: %s\n", dwarf_errmsg(rc) );
	    ASSERT(0);
	}
	if( version != 2 ) {
	    CPUWarning("Wrong dwarf version: %u != 2\n", version);
	    ASSERT(0);
	}
	if( dwarf_siblingof( exec->dwarf, NULL, &ddie, &rc ) != DW_DLV_OK) {
	    CPUWarning("dwarf_siblingof failed: %s\n", dwarf_errmsg( rc ) );
	    ASSERT(0);
	}

	basicdieinfo (&di, ddie, 0, exec);

	if (di.die_tag != DW_TAG_compile_unit) {
	    CPUWarning("Not a compilation unit! (%x)\n", di.die_tag);
	    ASSERT(0);
	} else {
	    /*CPUWarning("Found cu '%s'\n", di.at_name);*/
	    decode_line_numbers (di.ddie, &low_pc, &high_pc, exec);

	    /* fill in source structure information and add to table */
	    if (AssocTabLookup(exec->srcs, StripPath(di.at_name), NULL)
		== AT_OK) {
		/* nothing */
		CPUWarning("'%s' is a duplicate cu\n", di.at_name);
	    } else {
		src = (Srcfile*)malloc(sizeof(Srcfile));
		src->head.label = SaveString(StripPath(di.at_name));
		src->globals = AssocTabCreate();
		src->lopc = low_pc;
		src->hipc = high_pc;
		src->dieref = di.die_ref;

		if (AssocTabEnter(exec->srcs, (ATHeader*)src) != AT_OK) {
		    ASSERT(0);
		}

		/*CPUWarning("scanning '%s' for globals\n", src->head.label);*/
		/* Now look for partial symbols */
		scan_partial_symbols (di.childdie, exec, src);
		di.childdie = NULL;	/* avoid freeing here */
	    }
	}

	freedieinfo(&di,exec);
    }

}


static void
basicdieinfo ( struct dieinfo *dip, Dwarf_Die ddie, DIE_REF dieref,
	       Execfile *exec )

{
    Dwarf_Error      rc;
    Dwarf_Die        tmpdie;
    Dwarf_Bool       btmp;
    Dwarf_Attribute  attr;
    int              err;
    char            *stmp;

    memset (dip, 0, sizeof (struct dieinfo));

    /* if ddie not null, use it, otherwise we use dieref */
    if( ddie == NULL ) {
	err = dwarf_offdie(exec->dwarf,dieref, &ddie, &rc);
	if( err == DW_DLV_ERROR ) {
	    CPUWarning("dwarf_child failed: %s\n", dwarf_errmsg( rc ) );
	    ASSERT(0);
	} else if ( err != DW_DLV_OK ) {
	    CPUWarning("Couldn't find symbol ref %d\n",dieref);
	    ASSERT(0);
	}
    }
	
    dip->ddie = ddie;
    if( dwarf_tag( ddie, &dip->die_tag, &rc ) != DW_DLV_OK) {
	CPUWarning("dwarf_tag failed: %s\n", dwarf_errmsg( rc ) );
	ASSERT(0);
    }
    /* CPUWarning("tag: %x\n", dip->die_tag ); */
    err = dwarf_siblingof( exec->dwarf, ddie, &tmpdie, &rc );
    if( err == DW_DLV_ERROR ) {
	CPUWarning("dwarf_siblingof failed: %s\n", dwarf_errmsg( rc ) );
	ASSERT(0);
    } else if( err == DW_DLV_OK ) {
	dip->sibdie = tmpdie;
    }
    
    err = dwarf_child( ddie, &tmpdie, &rc );
    if( err == DW_DLV_ERROR ) {
	CPUWarning("dwarf_child failed: %s\n", dwarf_errmsg( rc ) );
	ASSERT(0);
    } else if( err == DW_DLV_OK ) {
	dip->childdie = tmpdie;
    }

    dwarf_dieoffset( ddie, &dieref, &rc );
    dip->die_ref = dieref;

    /* if there is a linkage name, we use that instead of the diename */
    
    if( dwarf_attr(ddie,DW_AT_MIPS_linkage_name,&attr,&rc) == DW_DLV_OK ) {
	if( dwarf_formstring( attr, &stmp, &rc ) != DW_DLV_OK) {
	    CPUWarning("dwarf_formstring failed: %s\n", dwarf_errmsg( rc ) );
	    ASSERT(0);
	}
	dip->at_name = stmp;
	dwarf_dealloc( exec->dwarf, attr, DW_DLA_ATTR);
    } else {
	err = dwarf_diename( ddie, &stmp, &rc );
	if( err == DW_DLV_ERROR ) {
	    CPUWarning("dwarf_diename failed: %s\n", dwarf_errmsg( rc ) );
	    ASSERT(0);
	} else if( err == DW_DLV_OK ) {
	    dip->at_name = stmp;
	}
    }

#if 0
    if( dip->at_name && (strcmp(dip->at_name,"main") == 0) ) {
	CPUWarning("basicdieinfo, main: ref %lld\n", (Reg64)dip->die_ref);
    }
    if( dip->at_name /*&& (dip->die_tag == DW_TAG_subprogram)*/ ) {
	static int counter = 0;
	if( counter++ < 20 ) {
	    CPUWarning("basicdieinfo, %s: ref %lld\n",
		       dip->at_name, (Reg64)dip->die_ref);
	}
    }
#endif
}


static void
PrintDieInfo( Dwarf_Die ddie, Execfile *exec )
{
    Dwarf_Unsigned   tmp;
    Dwarf_Addr       atmp;
    Dwarf_Bool       btmp;
    Dwarf_Off        otmp;
    Dwarf_Locdesc   *ldtmp;
    Dwarf_Signed     signedtmp;
    char            *stmp;
    Dwarf_Signed     atcnt;
    Dwarf_Attribute *atlist;
    Dwarf_Half       attr;
    Dwarf_Error      rc;
    Dwarf_Half       from;
    Dwarf_Die        tmpdie;
    int              i;
    int              err;

    
    if( (err = dwarf_attrlist( ddie, &atlist, &atcnt, &rc )) != DW_DLV_OK) {
	if( err != DW_DLV_NO_ENTRY ) {
	    CPUWarning("dwarf_attrlist failed: %s\n", dwarf_errmsg(rc));
	    return;
	}
    }

    for( i = 0; i < atcnt; i++ ) {

	if( dwarf_whatattr( atlist[i], &attr, &rc ) != DW_DLV_OK) {
	    CPUWarning("dwarf_whatattr failed: %s\n", dwarf_errmsg(rc) );
	    ASSERT(0);
	}

	/* CPUWarning("  attr: %x\n", attr ); */
	switch (attr) {
	case DW_AT_type:
 	    if( dwarf_formref( atlist[i], &otmp, &rc ) != DW_DLV_OK) {
		CPUWarning("dwarf_formref failed: %s\n", dwarf_errmsg(rc) );
		ASSERT(0);
	    }
	    CPUWarning("type %lld\n", (Reg64)otmp);
	    break;
	case DW_AT_ordering:
	    if( dwarf_arrayorder( ddie, &tmp, &rc ) != DW_DLV_OK) {
		CPUWarning("dwarf_arrayorder failed: %s\n", dwarf_errmsg(rc) );
		ASSERT(0);
	    }
	    CPUWarning("array order %lld\n", (Reg64)tmp);
	    break;
	case DW_AT_bit_offset:
	    if( dwarf_bitoffset( ddie, &tmp, &rc ) != DW_DLV_OK) {
		CPUWarning("dwarf_bitoffset failed: %s\n", dwarf_errmsg(rc) );
		ASSERT(0);
	    }
	    CPUWarning("bit off %lld\n", (Reg64)tmp);
	    break;
	case DW_AT_sibling:
	    /* we pick this up in the basicdieinfo routine */
	    break;
	case DW_AT_stmt_list:
	    /* we get our info using special line info routines */
	    break;
	case DW_AT_low_pc:
	    if( dwarf_lowpc( ddie, &atmp, &rc ) != DW_DLV_OK) {
		CPUWarning("dwarf_lowpc failed: %s\n", dwarf_errmsg(rc) );
		ASSERT(0);
	    }
	    CPUWarning("lowpc %llx\n", (Reg64)atmp);
	    break;
	case DW_AT_high_pc:
	    if( dwarf_highpc( ddie, &atmp, &rc ) != DW_DLV_OK) {
		CPUWarning("dwarf_highpc failed: %s\n", dwarf_errmsg(rc) );
		ASSERT(0);
	    }
	    CPUWarning("hipc %llx\n", (Reg64)atmp);
	    break;
	case DW_AT_language:
	    if( dwarf_srclang( ddie, &tmp, &rc ) != DW_DLV_OK) {
		CPUWarning("dwarf_srclang failed: %s\n", dwarf_errmsg(rc) );
		ASSERT(0);
	    }
	    CPUWarning("lang %lld\n", (Reg64)tmp);
	    break;
	case DW_AT_byte_size:
	    if( dwarf_bytesize( ddie, &tmp, &rc ) != DW_DLV_OK) {
		CPUWarning("dwarf_bytesize failed: %s\n", dwarf_errmsg(rc) );
		ASSERT(0);
	    }
	    CPUWarning("bytes %lld\n", (Reg64)tmp);
	    break;
	case DW_AT_bit_size:
	    if( dwarf_bitsize( ddie, &tmp, &rc ) != DW_DLV_OK) {
		CPUWarning("dwarf_bitsize failed: %s\n", dwarf_errmsg(rc) );
		ASSERT(0);
	    }
	    CPUWarning("bits %lld\n", (Reg64)tmp);
	    break;
	case DW_AT_location:
	    if( dwarf_loclist( atlist[i], &ldtmp, &signedtmp, &rc )
		!= DW_DLV_OK) {
		CPUWarning("dwarf_loclist failed: %s\n", dwarf_errmsg(rc) );
		ASSERT(0);
	    }
	    if( signedtmp != 1 ) {
		CPUWarning("location list size %lld != 1\n", signedtmp );
		ASSERT(0);
	    }
	    CPUWarning("location %p\n", ldtmp);
	    break;
	case DW_AT_data_member_location:
	    if( dwarf_loclist( atlist[i], &ldtmp, &signedtmp, &rc )
		!= DW_DLV_OK) {
		CPUWarning("dwarf_loclist failed: %s\n", dwarf_errmsg(rc) );
		ASSERT(0);
	    }
	    CPUWarning("dmloc %lld, %p\n", (Reg64)signedtmp, ldtmp);
	    break;
	case DW_AT_name:
	    if( dwarf_formstring( atlist[i], &stmp, &rc ) != DW_DLV_OK) {
		CPUWarning("dwarf_formstring failed: %s\n", dwarf_errmsg(rc) );
		ASSERT(0);
	    }
	    CPUWarning("name %s\n", stmp);
	    break;
	case DW_AT_comp_dir:
	    if( dwarf_formstring( atlist[i], &stmp, &rc ) != DW_DLV_OK) {
		CPUWarning("dwarf_formstring failed: %s\n", dwarf_errmsg(rc) );
		ASSERT(0);
	    }
	    /* For now, ignore any "hostname:" portion, since gdb doesn't
	       know how to deal with it.  (FIXME). */
	    CPUWarning("compdir %s\n", stmp);
	    break;
	case DW_AT_producer:
	    if( dwarf_formstring( atlist[i], &stmp, &rc ) != DW_DLV_OK) {
		CPUWarning("dwarf_formstring failed: %s\n", dwarf_errmsg(rc) );
		ASSERT(0);
	    }
	    CPUWarning("prod %s\n", stmp);
	    break;
	case DW_AT_start_scope:
	    if( dwarf_formudata( atlist[i], &tmp, &rc )	!= DW_DLV_OK) {
		CPUWarning("dwarf_formudata failed: %s\n", dwarf_errmsg(rc) );
		ASSERT(0);
	    }
	    CPUWarning("startscope %lld\n", (Reg64)tmp);
	    break;
	case DW_AT_stride_size:
	    CPUWarning("Dwarf stride size not handled yet\n");
	    ASSERT(0);
	    if( dwarf_formudata( atlist[i], &tmp, &rc )	!= DW_DLV_OK) {
		CPUWarning("dwarf_formudata failed: %s\n", dwarf_errmsg(rc) );
		ASSERT(0);
	    }
	    CPUWarning("stride %lld\n", (Reg64)tmp);
	    break;
	case DW_AT_prototyped:
	    if( dwarf_formflag( atlist[i], &btmp, &rc )	!= DW_DLV_OK) {
		CPUWarning("dwarf_formflag failed: %s\n", dwarf_errmsg(rc) );
		ASSERT(0);
	    }
	    CPUWarning("proto %lld\n", (Reg64)btmp);
	    break;
	case DW_AT_external:
	    if( dwarf_formflag( atlist[i], &btmp, &rc )	!= DW_DLV_OK) {
		CPUWarning("dwarf_formflag failed: %s\n", dwarf_errmsg(rc) );
		ASSERT(0);
	    }
	    CPUWarning("external %lld\n", (Reg64)btmp);
	    break;
	case DW_AT_encoding:
	    if( dwarf_formudata( atlist[i], &tmp, &rc )	!= DW_DLV_OK) {
		CPUWarning("dwarf_formudata failed: %s\n", dwarf_errmsg(rc) );
		ASSERT(0);
	    }
	    CPUWarning("encode %lld\n", (Reg64)tmp);
	    break;
	case DW_AT_upper_bound:
 	    if( dwarf_formudata( atlist[i], &tmp, &rc )	!= DW_DLV_OK) {
		CPUWarning("dwarf_formudata failed: %s\n", dwarf_errmsg(rc) );
		ASSERT(0);
	    }
	    CPUWarning("upperbound %lld\n", (Reg64)tmp);
	    break;
	case DW_AT_lower_bound:
 	    if( dwarf_formudata( atlist[i], &tmp, &rc )	!= DW_DLV_OK) {
		CPUWarning("dwarf_formudata failed: %s\n", dwarf_errmsg(rc) );
		ASSERT(0);
	    }
	    CPUWarning("lowerbound %lld\n", (Reg64)tmp);
	    break;
	case DW_AT_const_value:
 	    if( dwarf_formudata( atlist[i], &tmp, &rc )	!= DW_DLV_OK) {
		CPUWarning("dwarf_formudata failed: %s\n", dwarf_errmsg(rc) );
		ASSERT(0);
	    }
	    CPUWarning("const %lld\n", (Reg64)tmp);
	    break;
	case DW_AT_declaration:
	    if( dwarf_formflag( atlist[i], &btmp, &rc )	!= DW_DLV_OK) {
		CPUWarning("dwarf_formflag failed: %s\n", dwarf_errmsg(rc) );
		ASSERT(0);
	    }
	    CPUWarning("decl %lld\n", (Reg64)btmp);
	    break;
	case DW_AT_specification:
	    /* we may not need this... I'm not sure yet */
 	    if( dwarf_formref( atlist[i], &otmp, &rc ) != DW_DLV_OK) {
		CPUWarning("dwarf_formref failed: %s\n", dwarf_errmsg(rc) );
		ASSERT(0);
	    }
	    CPUWarning("specification %lld\n", (Reg64)otmp);
	    break;
	case DW_AT_MIPS_linkage_name:
	    /* It's not clear what name we should be using, but it seems gdb
	     * is happier using the full name, even though it has all the
	     * class context information if it wants it.
	     */
	    if( dwarf_formstring( atlist[i], &stmp, &rc ) != DW_DLV_OK) {
		CPUWarning("dwarf_formstring failed: %s\n", dwarf_errmsg(rc) );
		ASSERT(0);
	    }
	    CPUWarning("linkage %s\n", stmp);
	    break;
	case DW_AT_frame_base:
	case DW_AT_MIPS_fde:
	case DW_AT_identifier_case:
	case DW_AT_decl_file:
	case DW_AT_decl_line:
	case DW_AT_address_class:
	case DW_AT_artificial:
	case DW_AT_abstract_origin:
		/* these are currently not handled/needed */
	    break;
	default:
	    CPUWarning("Unhandled attribute: %x\n", attr);
	    /* Found an attribute that we are unprepared to handle.  However
	       it is specifically one of the design goals of DWARF that
	       consumers should ignore unknown attributes.  As long as the
	       form is one that we recognize (so we know how to skip it),
	       we can just ignore the unknown attribute. */
	    break;
	}
	
	dwarf_dealloc( exec->dwarf, atlist[i], DW_DLA_ATTR);
    }
    dwarf_dealloc( exec->dwarf, atlist, DW_DLA_LIST);
    
}





/*
 *   variables
 */

static AssocTab *exectab = NULL;


/* exported interface to simos */


int Dwarf_SymLoad(char *pathname, char *execname)
{
    int fd;
    char *filename;
    ATHeader *dummy;
    Execfile *exec;
    Dwarf_Error rc;

    if (!exectab) {
	exectab = AssocTabCreate();
    }

    /* check if execname already used */
    if (AssocTabLookup(exectab, execname, &dummy) == AT_OK) {
	SymReportError("execname already taken \"", execname, "\"", NULL);
	return SYM_ERROR;
    }      

    if (!(filename = SearchForFile(pathname))) {
	SymReportError("no such file \"", pathname, "\"", NULL);
	return SYM_ERROR;
    }

    fd = open( filename, O_RDONLY );
    if( fd < 0 ) {
	SymReportError("couldn't open \"", filename, "\"", NULL);
	return SYM_ERROR;
    }




    exec = (Execfile*)malloc(sizeof(Execfile));

    if( dwarf_init(fd,DW_DLC_READ,NULL,NULL,&exec->dwarf,&rc) != DW_DLV_OK) {
	free(exec);
	close(fd);
	SymReportError("dwarf_init for \"", filename, "\" failed:", 
		       dwarf_errmsg(rc), NULL);
	return SYM_ERROR;
    }

    exec->head.label = SaveString(execname);
    exec->srcs = AssocTabCreate();
    exec->globals = AssocTabCreate();
    exec->gStructs = AssocTabCreate();

    if (AssocTabEnter(exectab, (ATHeader*)exec) != AT_OK) {
	ASSERT(0);
    }

    scan_compilation_units (exec);

    return SYM_OK;
}

int Dwarf_SymAddr2Symbol(char *execname, int addr, Symb *sym)
{
    SymReportError("Dwarf_SymAddr2Symbol not supported", NULL);
    return SYM_ERROR;
}

int Dwarf_SymLine2Addr(char *execname, char *srcfile, int line, Symb *sym)
{
    Execfile *exec = NULL;
    Srcfile  *src = NULL;
    VA        bestaddr;
   
    CPUWarning("Dwarf_SymLine2Addr(%s,%s,%d,%p)\n",
	       execname, srcfile, line, sym);

    if (!exectab) exectab = AssocTabCreate();

    if (AssocTabLookup(exectab, execname, (ATHeader**)&exec) != AT_OK) {
	SymReportError("no exec named \"", execname, "\"", NULL);
	return SYM_ERROR;
    }


    if (AssocTabLookup(exec->srcs, StripPath(srcfile), (ATHeader**)&src)
	!= AT_OK) {
	SymReportError("no source file named \"", srcfile, "\"", NULL);
	return SYM_ERROR;
    }

    bestaddr = lookup_line(src,exec,line);

    sym->global = 0;
    sym->isym = src->dieref;
    sym->index = DWARF_INDEX;		/* special tornado marker */
    sym->exec = exec;
   
    sym->type[1] = tAddr;
    sym->type[2] = tNil;
    sym->nextt = 1;
    sym->value = bestaddr;
   
    return SYM_OK;
}

int Dwarf_SymLookup(char *execname, char *srcfile, char *symname,
		    int flags, Symb *sym)
{
    Dwarf_Error    rc;
    struct dieinfo di;
    Dwarf_Die      ddie;
    int            err;
    Execfile      *exec = NULL;
    Srcfile       *src = NULL;
    Global        *global = NULL;

    /*
    CPUWarning("Dwarf_SymLookup(%s,%s,%s,%x,%p)\n",
    execname, srcfile, symname, flags, sym);
    */

    if (!exectab) exectab = AssocTabCreate();

    if (AssocTabLookup(exectab, execname, (ATHeader**)&exec) != AT_OK) {
	SymReportError("no exec named \"", execname, "\"", NULL);
	return SYM_ERROR;
    }

    if( srcfile[0] ) {
	/* find matching source in exec->srcs, and then use src->globals */
	if (AssocTabLookup(exec->srcs,srcfile,(ATHeader**)&src) != AT_OK ) {
	    SymReportError("couldn't find src \"", srcfile, "\"", NULL);
	    return SYM_ERROR;
	}
	AssocTabLookup(src->globals,symname,(ATHeader**)&global);
    }
    
    /* we count on global != NULL to indicate we already found it */
    if( global == NULL ) {
	if (flags & SYM_TAGS) {
	    if ( AssocTabLookup(exec->gStructs,symname,(ATHeader**)&global)
		 != AT_OK ) {
		SymReportError("couldn't find symbol \"", symname, "\"", NULL);
		return SYM_ERROR;
	    }
	} else {
	    if ( AssocTabLookup(exec->globals,symname,(ATHeader**)&global)
		 != AT_OK ) {
		SymReportError("couldn't find symbol \"", symname, "\"", NULL);
		return SYM_ERROR;
	    }
	}
    }
    ASSERT(global!=NULL);

    /* fill in sym fields */
    return FillInSym( sym, global->dieref, exec );
}

int Dwarf_SymFind(Symb *sym, char *name, int flags)
{
    int st;
    int code;
    Dwarf_Off type;
    Dwarf_Die nextddie, ddie;
    struct dieinfo di;
    Execfile *exec;

    /*CPUWarning("Dwarf_SymFind(%p,%s,%x)\n", sym, name, flags);*/

    exec = sym->exec;
    
    basicdieinfo (&di, NULL, sym->isym, exec);
    ddie = di.childdie;
    di.childdie = NULL;
    freedieinfo(&di,exec);

    code = SYM_ERROR;
    
    while (ddie) {
	basicdieinfo (&di, ddie, 0, exec);
	nextddie = di.sibdie;
	di.sibdie = NULL;		/* ensure not freed */

	if ( di.die_tag == DW_TAG_inheritance ) {
	    /* we don't deal with all the visibility issues, but just
	     * do a depth-first search through parent classes
	     */
	    FillInSym(sym,GetTypeAbsOffset(di.ddie,exec),exec);
	    if ( Dwarf_SymFind(sym, name, flags) == SYM_OK ) {
		code = SYM_OK;
		break;
	    }
	}

	/* we effectively ignore blocks */
	if ( (di.die_tag == DW_TAG_lexical_block) ||
	     ((flags & SYM_RECURSE) && (di.die_tag == DW_TAG_subprogram)) ) {

	    FillInSym(sym,di.die_ref,exec);
	    if ( Dwarf_SymFind(sym, name, flags) == SYM_OK ) {
		code = SYM_OK;
		break;
	    }
	}

	if ( di.at_name && (strcmp(di.at_name,name) == 0) ) {
	    FillInSym(sym,di.die_ref,exec);
	    if ( (di.die_tag == DW_TAG_class_type) ||
		 (di.die_tag == DW_TAG_structure_type) ||
		 (di.die_tag == DW_TAG_enumeration_type) ) {
		if ( flags & SYM_TAGS ) {
		    code = SYM_OK;
		    break;
		}
	    } else if ( !(flags & SYM_TAGS) ) {
		code = SYM_OK;
		break;
	    }
	}

	freedieinfo(&di,exec);
	ddie = nextddie;
    }

    if( code != SYM_OK ) {
	SymReportError("couldn't find symbol \"", name, "\"", NULL);
    }
    return code;
}

int Dwarf_SymNext(Symb *sym)
{
    SymReportError("Dwarf_Symnext not supported", NULL);
    return SYM_ERROR;
}

int Dwarf_SymIn(Symb *sym)
{
    SymReportError("Dwarf_SymIn not supported", NULL);
    return SYM_ERROR;
}

int Dwarf_SymSubtype(Symb *sym, int flags)
{
    int st;
    int code;
    Dwarf_Off type;
    struct dieinfo di;
    int i;

    /*CPUWarning("Dwarf_SymSubtype(%p,%x)\n", sym, flags);*/

    if ( (sym->type[sym->nextt] == tPtr) ||
	 (sym->type[sym->nextt] == tAddr) ||
	 (sym->type[sym->nextt] == tArray)) {

	sym->nextt++;
	if( sym->type[sym->nextt] != tNil ) {
	    return SYM_OK;
	}
	if( IS_CAST(sym) ) {
	    /* this is a type cast and so our the dieref we have is for
	     * the end type itself rather than some base object that his
	     * this as it's type.
	     */
	    type = sym->isym;
	} else {
	    basicdieinfo (&di, NULL, sym->isym, sym->exec);
	    type = GetTypeAbsOffset( di.ddie, sym->exec );
	    ASSERT(type != 0);
#if 0
	    /* we assume if type is zero we have the type itself rather than
	     * an object of the given type; so just use objects dieref
	     */
	    if( type == 0 ) {
		type = sym->isym;
	    }
	    freedieinfo(&di,sym->exec);
#endif
	}
	code = FillInSym (sym, type, sym->exec);
	ASSERT(code == SYM_OK);
	sym->nextt++;
	ASSERT(sym->type[sym->nextt] != tNil);

	return SYM_OK;
    } 

    st = sym->type[sym->nextt];
      
    ASSERT((st == tStruct) || (st == tUnion) || (st == tEnum)
	   || (st == tTypedef));

    /* sym->isym is the DIE_REF of either some object whose type
     * we are interested in, or of the actual type descriptor.
     */
    if( IS_CAST(sym) ) {
	/* this is a type cast and so our the dieref we have is for
	 * the end type itself rather than some base object that his
	 * this as it's type.
	 */
	type = sym->isym;
    } else {
	basicdieinfo (&di, NULL, sym->isym, sym->exec);
	type = GetTypeAbsOffset( di.ddie, sym->exec );
	ASSERT(type != 0);
	freedieinfo(&di,sym->exec);
    }
#if 0
    if( (st == tTypedef) && ( di.die_tag == DW_TAG_typedef )) {
	/* special case: for casts, the DIE_REF we have is for the last
	 * type in the cast chain, not some object of the given type.
	 * for typedefs, because they have types themselves, we can
	 * fail to recognize this case, so we check whether the die
	 * it represents is a typedef itself and use that as the type
	 */
	/* FIXME: if we have typedef of typedef we can get this case even
	 * though it is not a cast case; we need a better check for this
	 * case
	 */
	type = sym->isym;
    } else {
	type = GetTypeAbsOffset( di.ddie, sym->exec );
	/* we assume if type is zero we have the type itself rather than
	 * an object of the given type; so just use objects dieref
	 */
	if( type == 0 ) {
	    type = sym->isym;
	}
    }
#endif
    code = FillInSym (sym, type, sym->exec);
    ASSERT(code == SYM_OK);

    if (st == tStruct) {
	ASSERT(sym->type[sym->nextt] == tStructDef);
         
    } else if (st == tUnion) {
	ASSERT(sym->type[sym->nextt] == tUnionDef);
         
    } else if (st == tEnum) {
	ASSERT(sym->type[sym->nextt] == tEnumDef);
         
    } else if (st == tTypedef) {
	ASSERT(sym->type[sym->nextt] == tTypedefDef);
	sym->nextt++;

	if (flags & SYM_RESOLVE) {
	    return SymSubtype(sym, flags);
	}
    } 
#if 1
else {
	ASSERT(0);
    }
#endif

    return SYM_OK;
}

int Dwarf_SymType(Symb *sym, int flags)
{
    Dwarf_Off type;
    struct dieinfo di;
    Symb tmp;

    /*CPUWarning("Dwarf_Symtype(%p,%x)\n", sym, flags);*/
    if ((sym->type[(sym)->nextt] != tTypedef) || !(flags & SYM_RESOLVE)) {
	return sym->type[(sym)->nextt];
    }

    ASSERT(sym->type[(sym)->nextt] == tTypedef);

    tmp = *sym;
    if( Dwarf_SymSubtype(&tmp, 0) != SYM_OK ) {
	return SYM_ERROR;
    }
    
    return Dwarf_SymType(&tmp, flags);
}

Reg64 Dwarf_SymSize(Symb *sym)
{
    /*CPUWarning("Dwarf_SymSize(%p)\n", sym);*/

    /* FIXME: this should really just ask dwarf what the size of the
     * various objects are.
     */
    switch (Dwarf_SymType(sym, SYM_RESOLVE)) {
    case tChar:
    case tUChar:
	return 1;
    case tShort:
    case tUShort:
	return 2;
    case tInt:
    case tUInt:
	return 4;
    case tLong:
    case tULong:
    case tPtr:
	return 8;
    case tFloat:
	return sizeof(float);
    case tDouble:
	return sizeof(double);
    case tLabel:
	return 4;
    case tProc: {
	Dwarf_Error    rc;
	struct dieinfo di;
	Dwarf_Die      ddie;
	int            err;
	Dwarf_Addr     atmp;
	DIE_REF        dieref;

	dieref = sym->isym;
	err = dwarf_offdie(sym->exec->dwarf,dieref, &ddie, &rc);
	if( err == DW_DLV_ERROR ) {
	    CPUWarning("dwarf_child failed: %s\n", dwarf_errmsg(rc) );
	    ASSERT(0);
	} else if ( err != DW_DLV_OK ) {
	    CPUWarning("Couldn't find symbol ref %d\n", dieref);
	    ASSERT(0);
	}
	if( dwarf_highpc( ddie, &atmp, &rc ) != DW_DLV_OK) {
	    CPUWarning("dwarf_highpc failed: %s\n", dwarf_errmsg(rc) );
	    ASSERT(0);
	}
	return atmp - sym->value;
    }

    /* FIXME: not sure about this; how exactly are array/addr types used */
    case tAddr:
	return 8;
    case tArray:
	return 8;
    case tStruct:
    case tUnion:
    case tEnum: {
	Symb tmp = *sym;
	if (Dwarf_SymSubtype(&tmp, SYM_RESOLVE) != SYM_OK) {
	    ASSERT(0);
	}
	return tmp.value;
    }
    case tStructDef:
    case tUnionDef:
    case tEnumDef:
	return sym->value;
    default:
	ASSERT(0);
	break;
    }

    /* never reached - for compiler */
    ASSERT(0);
    return 0;
}


int Dwarf_SymTypeCast(char *execname, char *srcfile, int *types,
		      char *name, Symb *sym)
{
    Symb tmp;
    int *tp;

    /*CPUWarning("Dwarf_SymTypeCast(%s,%s,%p,%s,%p)\n",
      execname, srcfile, types, name, sym); */

    sym->global = 1;
    sym->isym = -1;
    sym->filenum = -1;
    sym->name = NULL;
    sym->index = DWARF_INDEX;
    sym->exec = NULL;

    sym->nextt = 1;
    tp = types;

    while (*tp != tNil) {
	sym->type[sym->nextt++] = *tp;
	tp++;
    }

    ASSERT(sym->nextt<MAX_TYPES);
    sym->type[sym->nextt] = tNil;

    if (name[0]) {

	sym->nextt--; /* OK because we are going to reset it anyway */

	if ((sym->type[sym->nextt] == tStruct)
	    || (sym->type[sym->nextt] == tUnion)
	    || (sym->type[sym->nextt] == tEnum)) {
	    if (Dwarf_SymLookup(execname, srcfile, name, SYM_TAGS, &tmp) !=
		SYM_OK) {
		SymReportError("couldn't find symbol \"", name, "\"", NULL);
		return SYM_ERROR;
	    }

	} else if (sym->type[sym->nextt] == tTypedef) {
	    if (Dwarf_SymLookup(execname, srcfile, name, 0, &tmp) != SYM_OK) {
		SymReportError("couldn't find symbol \"", name, "\"", NULL);
		return SYM_ERROR;
	    }
	} else {
	    ASSERT(0);
	}

	sym->exec = tmp.exec;
	sym->index = tmp.index;
	sym->isym = tmp.isym;
    }
   
    sym->nextt = 1;
    sym->value = 0;

    return SYM_OK;
}

int Dwarf_SymMakeAddr(Symb *sym)
{
    /*CPUWarning("Dwarf_SymMakeAddr(%p)\n", sym);*/

    ASSERT(sym->nextt > 0);
    sym->nextt--;
    sym->type[sym->nextt] = tAddr;

    return SYM_OK;
}

