/*
 * Copyright (C) 1998 by the Board of Trustees
 *    of Leland Stanford Junior University.
 * Copyright (C) 1998 Digital Equipment Corporation
 *
 * This file is part of the SimOS distribution.
 * See LICENSE file for terms of the license.
 *
 */



/* ******************************************************
 * epsilon.c
 *
 * main fetch/decode/execute loop for the gamma instruction
 * set interpreter. To be integrated into SimOS. 
 * ******************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <setjmp.h>
#include <alpha/inst.h>

#include "simtypes.h"
#include "sim_error.h"

#include "gamma.h"
#include "epsilon.h"
#include "ev5.h"
#include "ev5_ipr.h"
#include "disassembler.h"
#include "alpha_params.h"
#include "alpha_regs.h"
#include "alpha_trace.h"
#include "ipr_decode.h"
#include "memref.h"
#include "alpha_memref.h"
#include "hw_events.h"
#include "tcl_init.h"
#include "print_insts.h"
#include "stats.h"
#include "cpu_stats.h"
#include "cpu_interface.h"
#include "registry.h"
#include "params.h"
#include "machine_params.h"
#include "qc.h"

/* @@@@ use of devices.h really should go somewhere else */
#include "../simos/devices.h"

#undef ASSERT
static int inAssert=0;
#define ASSERT(x)  {if (!(x)) \
{CPUWarning("%s:%d: assertion failed: %s\n",\
	    __FILE__, __LINE__, # x); \
  if(!inAssert)(inAssert=1,PrintAllInflightInstructions(EPE[0])); \
  CPUError("%s:%d: assertion failed (looping): %s\n",\
          __FILE__, __LINE__, # x);} }




#define CACHE_LINE_SIZE 64 /* elsewhere XXX */

static void EpsilonCPURun(void);
static void EpsilonDoneRunning(void);
static void EpsilonCleanupInstruction(int tpuNum, Inum inum);


#ifndef DEBUG_EPSILON
#define DEBUG_EPSILON 0
#endif


/* ********************************************
 * globals
 * ********************************************/
/* The instruction formats corresponding to the opcodes of instructions */

EpsilonState *epeArray;
EpsilonState **EPE;

EpsilonParams epsilonParams;

int savedDebugScheduler;


char *epsilonTTNames[] = {
  "Exit Simulator",
  "Register Value Changed Externally",
  "External Interrupt",
  "IMMU Exception",
  "Illegal Opcode",
  "FEN",
  "Arith",
  "DMMU HW Read Exception",
  "DMMU Read Exception",
  "DMMU Write Exception",
  "MTrap",
  "Mispredict",
  "Pal Call",
  "HW REI",
};

extern int numCPUs;
extern int64 gammaCurrentTime[SIM_MAXCPUS];
extern jmp_buf     jmpEnv;                    /* Used to exit the main loop */
extern int numLLActive;

static int prePCAnnsExist;
extern uint gdbInstr;
extern uint gdbInstrCommon;

extern int gamma_debug_mode;
extern int gamma_break_nexti;
extern int gamma_sigusr;

/* how many cpus are running */
static int notifyExit = 0;
static int nRunning = 0;
static int toCPUType = NO_CPU;

/*****************************************************************
 * ANNOTATION SUPPORT!
 *****************************************************************/
#define CHECK_LD_ANN(_vAddr, _pAddr) \
{  if (annLoads) {        \
      AnnPtr ptr = AnnFMLookup(_vAddr, ANNFM_LD_TYPE);\
      if (ptr)         \
         AnnExec(ptr); \
      ptr = AnnFMLookup(_pAddr, ANNFM_LD_TYPE);\
      if (ptr)         \
         AnnExec(ptr); \
   }                         \
}

#define CHECK_ST_ANN(_vAddr,_pAddr) \
{ if (annStores) {         \
      AnnPtr ptr = AnnFMLookup(_vAddr, ANNFM_ST_TYPE);\
      if (ptr)         \
         AnnExec(ptr); \
      ptr = AnnFMLookup(_pAddr, ANNFM_ST_TYPE);\
      if (ptr)         \
         AnnExec(ptr); \
 }                         \
}

/* ************************************************
 * Internal state handling
 * ************************************************/

static regname_t epsilonPhysregAlloc(EpsilonState *epe)
{
  regname_t physreg = epe->freereg;
  /* freelist chained through the registers themselves */
  ASSERT(physreg != -1);
  epe->freereg = (regname_t)epe->regs[physreg];
  EpsilonPhysicalRegisterSetAllocated(epe, physreg);
  if (0) CPUWarning("Physreg: allocated physreg=%d\n", physreg);
  ASSERT(physreg >= 0); ASSERT(physreg < N_PHYSICAL_REGS);
  return physreg;
}

static void epsilonPhysregFree(EpsilonState *epe, regname_t physreg)
{
  /* freelist chained through the registers themselves */
  ASSERT(physreg >= 0); ASSERT(physreg < N_PHYSICAL_REGS);
  EpsilonPhysicalRegisterClearAllocated(epe, physreg);
  epe->regs[physreg] = epe->freereg;
  epe->freereg = physreg;
  if (0) CPUWarning("Physreg: freed physreg=%d\n", physreg);
}

char *EpsilonStringifyRegisters(EpsilonState *epe)
{
  int i, j;
  static char buf[65536];
  int offset = 0;
  for (i = 0; i < N_PHYSICAL_REGS; i += 8) {
    for (j = 0; j < 8; j++) {
      int pr = i+j;
      int n;
      sprintf(buf+offset, " p%d=%s%s%lx%n",
	      pr, 
	      (EpsilonPhysicalRegisterIsAllocated(epe, pr) ? "a" : ""),
	      (EpsilonPhysicalRegisterIsValid(epe, pr) ? "v" : "i"),
	      epe->regs[pr],
	      &n);
      offset += n;
    }
    sprintf(buf+offset, "\n");
    offset++;
  }
  return strdup(buf);
}

static char *EpsilonPhysRegsAccess(ClientData clientData, Tcl_Interp *interp, char *name1, char *name2, int flags)
{
  Tcl_SetVar(interp, name1, EpsilonStringifyRegisters(EPE[CPUVec.CurrentCpuNum()]), TCL_GLOBAL_ONLY);
  return NULL;
}

static void EpsilonCheckRegmap(EpsilonState *epe)
{
  short inUse[N_PHYSICAL_REGS];
  short onList[N_PHYSICAL_REGS];
  regname_t *regmap = epe->regmap;
  regname_t physreg, lreg;
  for (physreg = 0; physreg < N_PHYSICAL_REGS; physreg++) {
    inUse[physreg] = 0;
    onList[physreg] = 0;
  }
  inUse[regmap[ZERO_REGISTER]] = 1;
  inUse[ZERO_REGISTER_DEST_PHYSREG] = 1;
  for (lreg = 0; lreg < N_LOGICAL_REGISTERS; lreg++) {
    regname_t physreg = regmap[lreg];
    if (lreg == ZERO_REGISTER || lreg == ZERO_REGISTER+32)
      continue;
    if (physreg < 0 || physreg > N_PHYSICAL_REGS)
      CPUError("Regmap corruption: regmap[%d] = %d\n", lreg, physreg);
    if (inUse[physreg])
      CPUError("Physical register is in use for two logical registers -- physreg=%d lreg=%d\n",
	       physreg, lreg);
    if (!EpsilonPhysicalRegisterIsAllocated(epe, physreg))
      CPUError("Physical register is unallocated -- physreg=%d lreg=%d\n", physreg, lreg);
    inUse[physreg] = 1;
  }
  /* now walk the free list */
  for (physreg = epe->freereg; physreg != -1; physreg = epe->regs[physreg]) {
    if (physreg < 0 || physreg > N_PHYSICAL_REGS)
      CPUError("Physreg freelist corruption: physreg=%d\n", physreg);
    if (inUse[physreg])
      CPUError("Physical register in use and on free list -- physreg=%d\n", physreg);
    if (onList[physreg]) {
      regname_t refphysreg;
      for (refphysreg = epe->freereg; refphysreg != -1; refphysreg = epe->regs[refphysreg]) {
	if (epe->regs[refphysreg] == physreg)
	  break;
      }
      CPUError("Physreg freelist cycle: physreg=%d (referring physreg)=%d\n", physreg, refphysreg);
    }
    if (EpsilonPhysicalRegisterIsAllocated(epe, physreg))
      CPUError("Free physical register is allocated -- physreg=%d\n", physreg);
    onList[physreg] = 1;
  }
  /* now make sure all the unused physregs are on the free list */
  for (physreg = 0; physreg < N_PHYSICAL_REGS; physreg++) {
    if (!EpsilonPhysicalRegisterIsAllocated(epe, physreg) && !onList[physreg])
      CPUError("Freelist leakage: physreg=%d is unallocated but not on the freelist\n", physreg);
  }
}

static int epsilonDefaultWidth = 4;

static EpsilonState *epeInit(int cpu)
{
  AlphaState *pe = PE[cpu];
  EpsilonState *epe = &epeArray[cpu];
  int  logreg, physreg;
  bzero((void*)epe, sizeof(EpsilonState));
  epe->myNum = cpu;
  epe->nextInum = 1; /* leave inum 1 as the guard inum */
  {
    InflightInstructionState *iis = &epe->iqueue[0];
    iis->done = 1;
  }
  pe->takeInterrupt = 0;
  pe->takeInterruptAtRei = epsilonParams.takeInterruptAtRei;

  epe->iqueue[0].nextpc = epe->nextpc = pe->PC;
  epe->iqueue[0].committed = 1;
  epe->iqueue[0].trapfree = 1;
  epe->iqueue[0].memopCompleted = 1;
  epe->iqueue[0].done = 1;
  epe->oldestTrapState = pe->trapState;
  epe->newestTrapState = pe->trapState;

  /* construct the physical register freelist */
  for (physreg = 0; physreg < N_PHYSICAL_REGS; physreg++) {
    epe->regs[physreg] = physreg+1;
  }
  epe->regs[N_PHYSICAL_REGS-1] = -1;
  epe->freereg = 0;

  /* construct the register map */
  for (logreg = 0; logreg < N_LOGICAL_REGISTERS; logreg++) {
    regname_t physreg = epsilonPhysregAlloc(epe);
    epe->regmap[logreg] = physreg;
    epe->committedRegmap[logreg] = physreg;
    EpsilonPhysicalRegisterSetValid(epe, physreg);
  }

  /* leave FP_ZERO's physregister to catch output to ZERO_REGISTER */
  /* use same physical register for source ZERO_REGISTER and FP_ZERO_REGISTER */
  epe->regmap[ZERO_REGISTER+32] = epe->regmap[ZERO_REGISTER];
  epe->committedRegmap[ZERO_REGISTER+32] = epe->committedRegmap[ZERO_REGISTER];
  epe->regs[epe->regmap[ZERO_REGISTER]] = 0;
  epe->regs[epe->regmap[ZERO_REGISTER+32]] = 0;
  ASSERT(epe->regs[epe->regmap[ZERO_REGISTER]] == 0);
  ASSERT(epe->regs[epe->regmap[ZERO_REGISTER+32]] == 0);

  if (epsilonParams.checkRegmap)
    EpsilonCheckRegmap(epe);
  if (epsilonParams.width == 0) 
    epsilonParams.width = epsilonDefaultWidth;
  return epe;
}

/* copy from P to epe */
static void EpsilonCopyStateIn (AlphaState *pe, EpsilonState *epe)
{
  int i;
  regname_t *regmap = epe->regmap;
  Reg *physregs = epe->regs;
  Reg *ireg = pe->reg;
  double *fpreg = pe->fp;
  Reg *sreg = pe->shadowRegs;
  Inum lastInumCommitted = epe->lastInumCommitted;
  InflightInstructionState *iis = &epe->iqueue[inumInstid(lastInumCommitted)];
  epe->useShadowRegisters = pe->useShadowRegisters;

  iis->nextpc = epe->nextpc = pe->PC;
  epe->oldestTrapState = pe->trapState;
  epe->newestTrapState = pe->trapState;
  /* copy in integer and FP registers */
  for (i = 0; i < 32; i++) {
    physregs[regmap[i]] = ireg[i];
    physregs[regmap[i+32]] = *(Reg*)&fpreg[i];
    /* leave a tracer */
    ireg[i] = 0xDEADBEEF0022;
    *(long*)&fpreg[i] = 0xDEADBEEF0023;
  }
  /* copy in shadow registers */
  for (i = 0; i < 16; i++) {
    physregs[regmap[64+i]] = sreg[i];
    /* leave a tracer */
    sreg[i] = 0xDEADBEEF0024;
  }

  ASSERT(epe->regs[epe->regmap[ZERO_REGISTER]] == 0);
  ASSERT(epe->regs[epe->regmap[ZERO_REGISTER+32]] == 0);

}

/* copy from epe back to pe */
static void EpsilonCopyStateOut (AlphaState *pe, const EpsilonState *epe)
{
  int i;
  const regname_t *regmap = epe->regmap;
  const Reg *physregs = epe->regs;
  Reg *ireg = pe->reg;
  double *fpreg = pe->fp;
  Reg *sreg = pe->shadowRegs;
  Inum lastInumCommitted = epe->lastInumCommitted;
  const InflightInstructionState *iis = &epe->iqueue[inumInstid(lastInumCommitted)];

  pe->PC = iis->nextpc;
  pe->trapState = epe->oldestTrapState;
  /* copy out integer and FP registers */
  for (i = 0; i < 32; i++) {
    ireg[i] = physregs[regmap[i]];
    fpreg[i] = *(double*)&physregs[regmap[i+32]];
  }
  /* copy out shadow registers */
  for (i = 0; i < 16; i++) {
    sreg[i] = physregs[regmap[i+64]];
  }
}

/* ************************************************
 * Print Inflight Instructions
 * ************************************************/

void PrintInflightInstruction(EpsilonState *epe, Inum inum, InflightInstructionState *iis)
{
  CPUWarning("  %ld: pc=%016lx nxinum=%ld instr=%08x opcode=%02x \t",
	     inum, iis->PC, iis->inorderNext, iis->instr.word, iis->instr.common.opcode);
  CPUWarning("%s%s%s%s%s%s%s%s%s%s%s",
	     iis->inflight ? "*" : " ",
	     iis->fetched ? "F" : " ",
	     iis->readyToIssue ? "i" : " ",
	     iis->issued ? "I" : " ",
	     iis->trapfree ? "s" : " ",
	     iis->untrappable ? "G" : " ",
	     iis->readyToRetire ? "r" : " ",
	     iis->committed ? "R" : " ",
	     iis->isMemop ? "m" : " ",
	     iis->memopCompleted ? "M" : " ",
	     iis->trapped ? "T" : " "
	     );
#ifdef EPSILON_USE_MEMREF  
  CPUWarning("(%d%s%d%s)",
	     iis->dmmuStatus,
	     iis->memopRequested ? "Q" : " ",
	     iis->memrefStatus,
	     iis->memopReadyToComplete ? "C" : " ");
#endif
  CPUWarning("%s%03d: %s%03d %s%03d %s%03d [%lx]",
	     (EpsilonPhysicalRegisterIsValid(epe, iis->prdest) ? "v" : " "), iis->prdest,
	     (EpsilonPhysicalRegisterIsValid(epe, iis->pra) ? "v" : " "), iis->pra,
	     (EpsilonPhysicalRegisterIsValid(epe, iis->prb) ? "v" : " "), iis->prb,
	     (EpsilonPhysicalRegisterIsValid(epe, iis->prc) ? "v" : " "), iis->prc,
	     epe->regs[iis->prdest]
	     );
  CPUWarning("\n");
  ASSERT(iis->prdest >= 0); ASSERT(iis->prdest < N_PHYSICAL_REGS);
  ASSERT(iis->pra >= 0); ASSERT(iis->pra < N_PHYSICAL_REGS);
  ASSERT(iis->prb >= 0); ASSERT(iis->prb < N_PHYSICAL_REGS);
  ASSERT(iis->prc >= 0); ASSERT(iis->prc < N_PHYSICAL_REGS);
}

void PrintAllInflightInstructions(EpsilonState *epe)
{
  Inum lastInumFetched = epe->lastInumFetched;
  Inum lastInumDone = epe->lastInumDone;
  Inum lastInumCommitted = epe->lastInumCommitted;
  Inum lastInumUntrapped = epe->lastInumUntrapped;
  Inum trapAbortInum = (epe->trapPending ? epe->trapAbortInum : -1);
  Inum lastStoreInum = epe->lastStoreInum;
  Inum trapPointInum = epe->trapPointInum;
  Inum inum;
  CPUWarning("InflightInstructions: cycles=%ld\n"
             "\tlastInumFetched=%ld lastInumUntrapped=%ld lastInumCommitted=%ld \n\t"
             "stallFetchUntilInum=%ld trapPointInum=%ld stallMemop=%ld\n", 
	     gammaCurrentTime[epe->myNum],
	     lastInumFetched, epe->lastInumUntrapped, lastInumCommitted, epe->stallFetchUntilInum,
	     trapPointInum, epe->stalledMemop);
  if (epe->trapPending) {
     CPUWarning("\t trapPending: trapCauseInum=%ld trapAbortInum=%ld trapType=%s\n",
                epe->trapCauseInum, epe->trapAbortInum, epsilonTTNames[epe->trapType]);
  }
  CPUWarning("  pending_irq=%lx\n", curPE->pending_irq);
  EpsilonCheckRegmap(epe);
  for (inum = lastInumFetched - N_INFLIGHT_INSTRS + 1; inum <= lastInumFetched; inum++) {
    if (inum < 0)
      continue;
    PrintInflightInstruction(epe, inum, &epe->iqueue[inumInstid(inum)]);
    if (inum == lastInumDone)
       CPUWarning("DDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDD\n");
    if (inum == lastInumCommitted)
       CPUWarning("RRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRR\n");
    if (inum == lastStoreInum)
       CPUWarning("SSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSS\n");
    if (inum == trapPointInum)
       CPUWarning("--------------------------------------------------------------------------------\n");
    if (inum == trapAbortInum)
       CPUWarning("AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA\n");
    if (inum == lastInumUntrapped)
       CPUWarning("UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU\n");
    if (inum == lastInumFetched)
       CPUWarning("FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF\n");
       
  }
}


/* ************************************************
 * effective address resolution
 * ************************************************/
static MMUStatus EpsilonResolveEffectiveAddress(AlphaState *pe, Inum inum,
						VA vAddr, int isWrite, int isEV5HW,
						PA *ppAddr, MA *pmAddr)
{
   MMUStatus status;
   PA pAddr = 0;
   MA mAddr = 0;
   if ((mAddr = QCLookup(pe->curDQC,vAddr,isWrite)) != (MA)0) {
      pAddr=MEMADDR_TO_PHYS(M_FROM_CPU(curPE->myNum),mAddr);
      status = MMU_SUCCESS;
   } else {
     status =EV5_DTranslateVirtual(curPE,vAddr,isWrite,0,&pAddr);
     if (status != MMU_EXCEPTION) {
       mAddr = PHYS_TO_MEMADDR(M_FROM_CPU(curPE->myNum),pAddr);
     } else {
       mAddr = 0;
       if (isEV5HW)
	 EpsilonTrapNotify(pe->myNum, inum, EpsilonTTDMMUHWReadException);
       else
	 EpsilonTrapNotify(pe->myNum, inum, (isWrite 
					    ? EpsilonTTDMMUWriteException 
					    : EpsilonTTDMMUReadException));
     }
   }
   *ppAddr = pAddr;
   *pmAddr = mAddr;
   if ((int)vAddr == (int)alphaDebugParams.debugVATranslation) {
     CPUWarning("EpsilonResolveEffectiveAddress: pc=%lx mmustatus=%d va=%lx pa=%lx ma=%lx isWrite=%d\n",
		pe->PC, status, vAddr, *ppAddr, *pmAddr, isWrite);
   }
   return status;
}

/* maps opcode to size, in bytes, of value read or written */
static char memopValueSize[64] = {
  0,	/* op_call_pal */
  0,	/* op_opc01 */
  0,	/* op_opc02 */
  0,	/* op_opc03 */
  0,	/* op_opc04 */
  0,	/* op_opc05 */
  0,	/* op_opc06 */
  0,	/* op_opc07 */
  0,	/* op_lda */
  0,	/* op_ldah */
  1,	/* op_ldbu */
  8,	/* op_ldq_u */
  2,	/* op_ldwu */
  2,	/* op_stw */
  2,	/* op_stb */
  8,	/* op_stq_u */
  0,	/* op_inta */
  0,	/* op_intl */
  0,	/* op_ints */
  0,	/* op_intm */
  0,	/* op_fltm */
  0,	/* op_fltv */
  0,	/* op_flti */
  0,	/* op_fltl */
  0,	/* op_misc */
  0,	/* op_pal19 */
  0,	/* op_jsr */
  0,	/* op_pal1b */
  0,	/* op_intmisc */
  0,	/* op_pal1d */
  0,	/* op_pal1e */
  0,	/* op_pal1f */
  4,	/* op_ldf */
  8,	/* op_ldg */
  4,	/* op_lds */
  8,	/* op_ldt */
  4,	/* op_stf */
  8,	/* op_stg */
  4,	/* op_sts */
  8,	/* op_stt */
  4,	/* op_ldl */
  8,	/* op_ldq */
  4,	/* op_ldl_l */
  8,	/* op_ldq_l */
  4,	/* op_stl */
  8,	/* op_stq */
  4,	/* op_stl_c */
  8,	/* op_stq_c */
  0,	/* op_br */
  0,	/* op_fbeq */
  0,	/* op_fblt */
  0,	/* op_fble */
  0,	/* op_bsr */
  0,	/* op_fbne */
  0,	/* op_fbge */
  0,	/* op_fbgt */
  0,	/* op_blbc */
  0,	/* op_beq */
  0,	/* op_blt */
  0,	/* op_ble */
  0,	/* op_blbs */
  0,	/* op_bne */
  0,	/* op_bge */
  0,	/* op_bgt */
};

static void EpsilonStoreBufferWrite(AlphaState *P, EpsilonState *epe, InflightInstructionState *iis, Reg mdata)
{
  VA vAddr = iis->eaddr;
  PA paddr = iis->paddr;

  int shift = ((paddr & 0x7)*8);
  int size = memopValueSize[iis->instr.common.opcode];
  Reg vmask = (1L << (8*size))-1L;
  if (size < 8)
    vmask = vmask << shift;
  else 
    vmask = -1L;
  iis->storebuffer.vmask = vmask;
  if (gammaParams.debugLoadStore || vAddr == alphaDebugParams.debugVATranslation)
    CPUWarning("EpsilonStoreBufferWrite: opc=%x size=%d shift=%d paddr=%016lx vmask=%016lx mdata=%016lx\n", 
	       iis->instr.common.opcode, size, shift, paddr, vmask, mdata);
}

static Reg EpsilonStoreBufferRead(AlphaState *P, EpsilonState *epe, InflightInstructionState *iis, Reg mdata)
{
  Inum lastInumDone = epe->lastInumDone;
  Inum inum;
  InflightInstructionState *siis = NULL;
  Reg value = 0;
  VA vAddr = iis->eaddr;
  PA paddr = iis->paddr;
  PA baddr = paddr & ~0x7l;
  int shift = ((paddr & 0x7)*8);
  int size = memopValueSize[iis->instr.common.opcode];
  Reg vmask = (1L << (8*size))-1L;
  if (size < 8)
    vmask = vmask << shift;
  else 
    vmask = -1L;
  /* walk back looking for matching addresses until all bytes are fetched or storebuffer is empty */
  for (inum = iis->lastStoreInum; (vmask != 0) && (inum > lastInumDone); inum = siis->lastStoreInum) {
    Instid instid = inumInstid(inum);
    VA paddr1;
    siis = &epe->iqueue[instid];
    paddr1 = siis->paddr;
    if (baddr == (paddr1 & ~0x7L)) {
      Reg mdata1 = siis->value << ((paddr1 & 0x7)*8);
      Reg vmask1 = siis->storebuffer.vmask;
      value |= (mdata1 & vmask1);
      vmask &= ~vmask1;
      ASSERT(siis->issued);
    }
  }
  mdata = mdata << shift;
  value |= (mdata&vmask);
  value = value >> shift;
  if (size == 4) /* need to sign extend */
    *(long *)&value = *(int *)&value;
  if (gammaParams.debugLoadStore || vAddr == alphaDebugParams.debugVATranslation)
    CPUWarning("EpsilonStoreBufferRead: opc=%x size=%d shift=%d paddr=%016lx vmask=%016lx mdata=%016lx\n", 
	       iis->instr.common.opcode, size, shift, paddr, vmask, mdata);
  return value;
}

/* ************************************************
 * main entry point
 * ************************************************/

void ExecuteEpsilon()
{
  int i;

  /*
   * setup original arguments
   */
  GammaInitTables();
  /* reuse most */
  GammaCPUVectorInit();
  /* override a few */
  EpsilonCPUVectorInit();

  if (!alphaLateInitDone) {
     /*
      * Stuff that needs late initalization.
      * XXX should use a register mechanism
      */
     GammaInitStats();

     AnnCommonSetup();
     AnnFMInit(128);
     InstallPoller();  /* devices.c */
     prePCAnnsExist = AnnFMTypesExist(ANNFM_PRE_PC_TYPE);
     InitSimulatorStats(4*1024*1024, GammaStats);
     HWEventsLateInit();
     for (i=0; i < SIM_MAXCPUS; i++) {
        gammaCurrentTime[i] = -1;
     }
  }

  curPE = PE[0];
  curPE->cpuState = cpu_running;
  numCPUs = NUM_CPUS(0);

  CPUWarning("P->PC entry = 0x%lx \n",curPE->PC);

  {
    int i;
    epeArray = realloc(epeArray, numCPUs*sizeof(EpsilonState));      
    EPE = realloc(EPE, numCPUs*sizeof(EpsilonState*));      
    for (i = 0; i < numCPUs; i++) {
      EPE[i] = epeInit(i);
    }
  }

  /*
   * simos enter annotation
   */
  if (!alphaLateInitDone) {
     int cpu;
     for (cpu=0;cpu<TOTAL_CPUS;cpu++) {
	curPE = PE[cpu];
	AnnExec(AnnFind("simos","enter"));
     }
     curPE = PE[0];

     alphaLateInitDone = 1;
  }

  GammaFPInit(); /* installs SIGFPE handler */

  /* 
   * If we're using another CPU scheduler, initialize it here
   */
  PMINT_INIT(NUM_CPUS(0));

  EpsilonCPURun();
  EpsilonDoneRunning();
  simosCPUType = toCPUType;
}

void EpsilonTrapNotify(int tpuNum, Inum trapCauseInum, int trapType)
{
  EpsilonState *epe = EPE[tpuNum];
  InflightInstructionState *iis = &epe->iqueue[inumInstid(trapCauseInum)];
  if (epsilonParams.debugScheduler
      || (trapType == EpsilonTTInterrupt && alphaDebugParams.debugInterrupts))
    CPUWarning("EpsilonTrapNotify: cpu=%d trapCauseInum=%ld trapType=%d:%s\n", 
	       tpuNum, trapCauseInum, trapType, (trapType < EpsilonTTNumTypes ? epsilonTTNames[trapType] : "??"));
  if (!epe->trapPending
      || trapCauseInum < epe->trapCauseInum
      || (trapCauseInum == epe->trapCauseInum && trapType < epe->trapType)) {
    Inum trapAbortInum;
    epe->trapPending = 1;
    epe->trapPC = iis->PC;
    epe->trapCauseInum = trapCauseInum;
    epe->trapType = trapType;

    /* determine which inums to abort */
    switch (epe->trapType) {
    case EpsilonTTIMMUException:
    case EpsilonTTDMMUReadException:
    case EpsilonTTDMMUHWReadException:
    case EpsilonTTDMMUWriteException:
      trapAbortInum = trapCauseInum;
      break;
    case EpsilonTTIllegalOpcode:
    case EpsilonTTFEN:
    case EpsilonTTArith:
      trapAbortInum = trapCauseInum;
      break;
    case EpsilonTTPalCall:
    case EpsilonTTHwRei:
      trapAbortInum = iis->inorderNext;
      break;
    case EpsilonTTInterrupt:
      if (epsilonParams.debugSchedulerDuringInterrupt) {
	savedDebugScheduler++;
	epsilonParams.debugScheduler = 1;
      }
      {
         Inum trapPointInum = epe->trapPointInum;
         trapAbortInum = trapPointInum + 1;
      }
      break;
    case EpsilonTTMispredict:
       iis->trapfree = 1; /* now that we've registered the mispredict, 
                           * this instruction cannot cause any more traps */
       trapAbortInum = iis->inorderNext;
       break;
    case EpsilonTTExitSimulator:
       trapAbortInum = iis->inorderNext;
       if (epe->stallFetchUntilInum < trapCauseInum) {
          epe->stallFetchUntilInum = trapCauseInum;
       }
       break;
    case EpsilonTTRegisterChangedExternally:
      trapAbortInum = iis->inorderNext;
      break;
    default:
      CPUError("EpsilonTrapNotify: undefined trapType=%d\n", epe->trapType);
      trapAbortInum = epe->lastInumFetched+1;
      NOTREACHED();
    }
    epe->trapAbortInum = trapAbortInum;
    ASSERT(trapAbortInum >= trapCauseInum);
    if ((epsilonParams.debugTrapMask & (1 << epe->trapType))
        || epsilonParams.debugScheduler
        || (trapType == EpsilonTTInterrupt && alphaDebugParams.debugInterrupts))
      CPUWarning("EpsilonTrapNotify: most significant trap is cpu=%d pc=%lx ea=%lx inum=%ld type=%d:%s abortInum=%ld\n",
		 tpuNum, iis->PC, iis->eaddr, trapCauseInum, trapType,
		 (trapType < EpsilonTTNumTypes ? epsilonTTNames[trapType] : "??"),
		 trapAbortInum);
  }
}

static Inum lic = 0;
static long lastCycleInstructionCommitted = 0;

void EpsilonPreCycle()
{
  int cpu;
  AlphaState *pe;
  if (epsilonParams.debugScheduler)
    CPUWarning("EpsilonPreCycle: cycles=%ld stallFetchUntilInum=%d\n", 
	       gammaCurrentTime[0], EPE[0]->stallFetchUntilInum);
  for (cpu = 0; (pe = PE[cpu]) != NULL; cpu++) {
     gammaCurrentTime[cpu]++;
     curPE = PE[cpu];
     EventPollSingleQueue(gammaCurrentTime[cpu]);
  }

  /* check to see if we need to exit */
  if (notifyExit) {
     for (cpu = 0; (pe = PE[cpu]) != NULL; cpu++) {
        EpsilonState *epe = EPE[cpu];
        curPE = PE[cpu];
        EpsilonTrapNotify(cpu, epe->trapPointInum, EpsilonTTExitSimulator);
     }
     /* PrintAllInflightInstructions(EPE[0]); */
     notifyExit = 0;
  }
  /* now restore the interrupt level and do a trap notify if there are any pending_irq */
  if (epsilonParams.trapNotifyInterrupts && ((gammaCurrentTime[0]&epsilonParams.interruptFrequencyMask)==0)) {
    for (cpu = 0; (curPE = PE[cpu]) != NULL; cpu++) {
      AlphaState *pe = curPE;
      AlphaTrapState *trapState = pe->trapState;
      EpsilonState *epe = EPE[cpu];
      Inum inum = 
	(epsilonParams.interruptTrapUncommitted) 
	? epe->lastInumCommitted
	: epe->lastInumUntrapped;
      InflightInstructionState *iis = &epe->iqueue[inumInstid(inum)];
      long irqMask = ~((1L<<(trapState->atr[ATR_IPL]+1))-1L);
      pe->PC = iis->nextpc;
      if ((pe->pending_irq&irqMask) && !IS_PAL(pe)) {
	if (alphaDebugParams.debugInterrupts)
	  CPUWarning("EpsilonPreCycle: external interrupt: cpu=%d inum=%ld pc=%lx IPL=%ld pending_irq=%x\n",
		     pe->myNum, inum, pe->PC, trapState->atr[ATR_IPL], pe->pending_irq,
		     irqMask);
	EpsilonTrapNotify(cpu, inum, EpsilonTTInterrupt);
	epe->pending_irq = pe->pending_irq;
      }
    }
  }
}

void EpsilonPostCycle()
{
   if (epsilonParams.debugScheduler)
      CPUWarning("EpsilonPostCycle: cycles=%ld stallFetchUntilInum=%d lastInumCommitted=%ld\n", 
                 gammaCurrentTime[0], EPE[0]->stallFetchUntilInum, EPE[0]->lastInumCommitted);
   if (EPE[0]->lastInumCommitted > lic) {
      /* making progress */
      lic = EPE[0]->lastInumCommitted;
      lastCycleInstructionCommitted = gammaCurrentTime[0];
   } else if ((gammaCurrentTime[0] - lastCycleInstructionCommitted) > 128) {
      CPUWarning("EpsilonPostCycle: no instructions committed for %ld cycles\n",
                 gammaCurrentTime[0] - lastCycleInstructionCommitted);
      PrintAllInflightInstructions(EPE[0]);
      CPUError("EpsilonPostCycle: looks like we're not making any forward progress\n");
   }
}


void EpsilonTrapCycle(AlphaState *pe, EpsilonState *epe)
{
  int cpu = pe->myNum;

  if (pe->cpuState != cpu_running)
    return;
  curPE = pe;
  if (epsilonParams.checkRegmap)
    EpsilonCheckRegmap(epe);
  /* @@@ process traps here */
  if (epe->trapPending) {
    Inum lastInumUntrapped = epe->lastInumUntrapped;
    Inum trapInum = epe->trapCauseInum;
    Inum abortInum = epe->trapAbortInum;
    InflightInstructionState *iis = &epe->iqueue[inumInstid(trapInum)];
    int currentUseShadowRegisters = EV5_ShadowRegistersUsed(pe, iis->PC);

    if (alphaDebugParams.debugTraps && ((epsilonParams.debugTrapMask >> epe->trapType)&1))
      CPUWarning("EpsilonCycle: TRAP cpu=%d pc=%lx [inum=%ld -- lastInumUntrapped=%ld] trapType=%d:%s\n",
		 cpu, iis->PC, trapInum, lastInumUntrapped, epe->trapType, epsilonTTNames[epe->trapType]);

    if (lastInumUntrapped >= abortInum) {
      /* abort the newer instructions */
      Inum inum;
      InflightInstructionState *iis = NULL;
      for (inum = lastInumUntrapped; inum >= abortInum; inum = lastInumUntrapped) {
	/* this does not clear the InflightInstructionState, 
	 * it just sets the trapped flag and does other bookkeeping 
	 */
	iis = &epe->iqueue[inumInstid(inum)];
	EpsilonTrapInstruction(cpu, inum);
	lastInumUntrapped = iis->inorderPrev;
      }
      {
	InflightInstructionState *untrappedIIS = &epe->iqueue[inumInstid(lastInumUntrapped)];
	if (EV5_ShadowRegistersUsed(pe, untrappedIIS->PC) != currentUseShadowRegisters) {
	  ASSERT(!untrappedIIS->trapped);
	  EV5_SwitchShadowRegisters(pe, untrappedIIS->PC, pe->ipr[IPR_ICSR]);
	}
      }
      epe->trapAbortInum = epe->lastInumFetched+1; /* no more to abort until next fetch */
    }


    if (epsilonParams.trapNoStall || epe->lastInumCommitted >= epe->stallFetchUntilInum) {
      int needsNewTrapState;

      switch (epe->trapType) {
      case EpsilonTTIMMUException:
      case EpsilonTTDMMUReadException:
      case EpsilonTTDMMUHWReadException:
      case EpsilonTTDMMUWriteException:
      case EpsilonTTIllegalOpcode:
      case EpsilonTTFEN:
      case EpsilonTTArith:
      case EpsilonTTPalCall:
      case EpsilonTTHwRei:
      case EpsilonTTInterrupt:
	needsNewTrapState = 1;
	break;
      case EpsilonTTMispredict:
      case EpsilonTTRegisterChangedExternally:
      case EpsilonTTExitSimulator:
	needsNewTrapState = 0;
	break;
      default:
	CPUError("EpsilonTrapNotify: undefined trapType=%d\n", epe->trapType);
	needsNewTrapState = 0;
	NOTREACHED();
      }

      if (epe->trapType == EpsilonTTIllegalOpcode
	  || (alphaDebugParams.debugTraps && ((epsilonParams.debugTrapMask >> epe->trapType)&1)))
	CPUWarning("EpsilonCycle: trapped at cycle=%ld cpu=%d pc=%lx inum=%ld instr=%x opcode=0x%x type=%d:%s\n",
		   gammaCurrentTime[cpu], cpu, iis->PC, epe->trapCauseInum, iis->instr.word, iis->instr.common.opcode, 
		   epe->trapType, epsilonTTNames[epe->trapType]);

      {
	/* allocate new trap state */
	AlphaTrapState *trapState = 
	  (needsNewTrapState && epsilonParams.newTrapStates)
	  ? allocTrapState(epe->newestTrapState)
	  : epe->newestTrapState;
	ASSERT(trapState != NULL);
	epe->newestTrapState = trapState;

	if (needsNewTrapState) {
	  if (epsilonParams.trapClearRegisterLocks && trapState->lockedRegisters) {
	    /* @@@@ clear the register locks here
	     */
	    NOTREACHED(); /* @@@ I think this is a bad idea now.  Will remove. -Jamey 10/28/97 */
	    CPUWarning("EpsilonCycle: @@@@ clearing lockedRegisters=%x pc=%lx inum=%ld\n", 
		       trapState->lockedRegisters, iis->PC, iis->inum);
	    trapState->lockedRegisters = 0;
	  }

	}
	  
	/* now update the trap state */
	pe->PC = iis->PC; /* for synchronous exceptions, point to PC of instruction causing exception */
	pe->instr = iis->instr;
	pe->trapState = trapState;
	switch (epe->trapType) {
	case EpsilonTTIMMUException: {
	  PA paddr;
	  /* performed for side-effect */
	  EV5_ITranslateVirtual(pe, &paddr, 1);
	} break;
	case EpsilonTTDMMUReadException:
	case EpsilonTTDMMUHWReadException:
	case EpsilonTTDMMUWriteException: {
	  PA paddr;
	  MMUStatus mmuStatus =
	    EV5_DTranslateVirtual(pe, iis->eaddr,
				  (epe->trapType == EpsilonTTDMMUWriteException ? 1 : 0), 1,
				  &paddr);
	  if (alphaDebugParams.debugTraps && ((epsilonParams.debugTrapMask >> epe->trapType)&1))
	    CPUWarning("\t @@ MMU Exception -- inum=%ld eaddr=%lx mmuStatus=%d type=%d:%s\n",
		       iis->inum, iis->eaddr, mmuStatus, 
		       epe->trapType, epsilonTTNames[epe->trapType]);
	  if (epe->trapType == EpsilonTTDMMUHWReadException
	      && mmuStatus==MMU_EXCEPTION) {
	    if (pe->PC != pe->ipr[IPR_PAL_BASE] + TRAP_NDTB_MISS) {
	      CPUError("HWLd: DTranslate failed with pc=0x%lx \n",
		       pe->PC);
	    }
	    if (pe->PC == pe->ipr[IPR_PAL_BASE] + TRAP_NDTB_MISS) {
	      pe->PC = pe->ipr[IPR_PAL_BASE] + TRAP_PDTB_MISS;
	    }
	  }
	} break;
	case EpsilonTTIllegalOpcode: {
	  EV5_Trap(pe, TRAP_OPDEC);
	} break;	  
	case EpsilonTTFEN: {
	  EV5_Trap(pe, TRAP_FEN);
	} break;	  
	case EpsilonTTArith: {
	  EV5_Trap(pe, TRAP_ARITH);
	} break;	  
	case EpsilonTTPalCall: {
	  EV5_PALOpcode(pe,iis->instr.pal_format.function);
	  iis->nextpc = epe->nextpc = pe->PC;
	} break;

	case EpsilonTTHwRei: {
	  EV5_HwRei(pe);
	  iis->nextpc = epe->nextpc = pe->PC;
	} break;

	case EpsilonTTInterrupt: {
	  int rc;
	  pe->PC = iis->nextpc; /* @@ asynchronous execption: we want to continue at the next instruction */
	  rc = EV5_CheckForInterrupts(pe,"EpsilonTrapCycle");
	  if (!rc) {
	    AlphaTrapState *ts = epe->newestTrapState;
	    CPUWarning("EpsilonTrapCycle: no interrupt\n");
	    epe->newestTrapState = ts->prev;
	    iis->trapState = ts->prev;
	    if (needsNewTrapState && epsilonParams.newTrapStates)
	      freeTrapState(ts);
	  }
	  epe->nextpc = pe->PC;
	} break;

	case EpsilonTTMispredict:
	case EpsilonTTRegisterChangedExternally: {
	  /* pick up the actual next PC that was computed at issue time */
	  pe->PC = iis->nextpc; /* @@ not really an exception, continue at next instruction */
	} break;
	case EpsilonTTExitSimulator: {
	  /* pick up the actual next PC that was computed at issue time */
	  pe->PC = iis->nextpc; /* @@ not really an exception, continue at next instruction */
          pe->cpuState = cpu_halted;
          nRunning--;
          /* PrintAllInflightInstructions(epe); */
	} break;
	default:
	  NOTREACHED();
	}

	epe->nextpc = pe->PC;
	epe->trapPending = 0;
      
      }
    }
  }
  ASSERT(epe->regs[epe->regmap[ZERO_REGISTER]] == 0);
  ASSERT(epe->regs[epe->regmap[ZERO_REGISTER+32]] == 0);
}

static void EpsilonDebugCycle(AlphaState *pe, EpsilonState *epe)
{
  /*
   * Single-stepping debugger
   */
  curPE = pe;
  if (gamma_debug_mode) {
    if (gamma_sigusr) {
      gamma_debug_mode = 0;
      gamma_sigusr = 0;
      AnnExec(AnnFind("simos", "sigusr"));
    } else if ((curPE->myNum == gamma_break_nexti) 
	       || (gamma_break_nexti == GAMMA_BREAKANYCPU)) {
      /* GammaDebug returns nonzero if the PC has changed */
      /* or the contents of it have been overwritten, so it */
      /* needs to be reloaded. */
      if (GammaDebug(curPE->myNum, 1)) {
	EpsilonTrapNotify(curPE->myNum, epe->lastInumCommitted, EpsilonTTRegisterChangedExternally);
      }
    }
  }
}

static void EpsilonFetchCycle(AlphaState *pe, EpsilonState *epe)
{
  int cpu = pe->myNum;
  int fetch;
  VA predictedPC = epe->nextpc;
  pe->PC = epe->nextpc;
  if (epe->lastInumCommitted >= epe->stallFetchUntilInum
      && (epe->lastInumFetched - epe->lastInumCommitted < epsilonParams.fetchAheadLimit)) {
    for (fetch = 0; fetch < epsilonParams.width; fetch++) {
    
      Inum inum = EpsilonFetchInstruction(cpu, predictedPC);
      Instid instid = inumInstid(inum);
      InflightInstructionState *iis = &epe->iqueue[instid];
      union alpha_instruction instr = iis->instr;
      switch (instr.common.opcode) {
      case op_br:
      case op_bsr:
	predictedPC = iis->PC + (iis->instr.b_format.branch_displacement<<2) + 4;
	break;
      case op_beq:
      case op_blt:
      case op_ble:
      case op_bne:
      case op_bge:
      case op_bgt:
      case op_fbeq:
      case op_fblt:
      case op_fble:
      case op_fbne:
      case op_fbge:
      case op_fbgt: {
	int disp = iis->instr.b_format.branch_displacement;
	if (disp > 0) {
	  predictedPC += 4;
	} else {
	  predictedPC = iis->PC + (disp<<2) + 4;
	}
      } break;
      case op_jsr:
	/* @@@ should use prediction stack here */
	predictedPC = iis->PC + (iis->instr.j_format.hint<<2) + 4;
	break;
      case op_misc:
	/* @@@@ in case we get an MB used as an IMB */
	epe->stallFetchUntilInum = inum;
	break;
      case PRIV_OP_HW_LD:
      case PRIV_OP_MFPR: /* @@@ some mfpr's have a side-effect */
      case PRIV_OP_HW_ST:
      case PRIV_OP_MTPR:
	predictedPC += 4;
	epe->stallFetchUntilInum = inum;
	fetch += epsilonParams.width; /* force end of fetch this cycle */
	break;
      case op_call_pal:
      case PRIV_OP_REI:
	/* @@@@ should use call stack predictor and do this at issue time? */
	predictedPC = predictedPC+4;
	epe->stallFetchUntilInum = inum;
	fetch += epsilonParams.width; /* force end of fetch this cycle */
	break;
      default:
	predictedPC += 4;
      }
      iis->nextpc = predictedPC;
    } 
    epe->nextpc = predictedPC;
  } else {
    if (epsilonParams.debugScheduler)
      CPUWarning("EpsilonFetch: stalled until inum=%ld retires\n",
		 epe->stallFetchUntilInum);
  }
}

static void EpsilonSchedCycle(AlphaState *pe, EpsilonState *epe)
{
  int cpu = pe->myNum;
  Inum inum;
  Inum lastInumFetched = epe->lastInumFetched;
  Inum lastInumCommitted = epe->lastInumCommitted;
  int unreadyLoadOrStore = 0;
  int issueCount = 0;
  for (inum = lastInumCommitted+1; inum <= lastInumFetched; inum++) {
    InflightInstructionState *iis = &epe->iqueue[inumInstid(inum)];
    ASSERT(inum == iis->inum);
    ASSERT(iis->fetched);
    /* @@@ issue loads and stores in order */
    if (!iis->readyToIssue && !iis->trapped) {
      if (epsilonParams.debugScheduler)
	CPUWarning("EpsilonSched: cpu=%d pc=%lx inum=%ld instr=%x opcode=0x%x valid = %d %d %d\n",
		   cpu, iis->PC, iis->inum, iis->instr.word, iis->instr.common.opcode, 
		   EpsilonPhysicalRegisterIsValid(epe, iis->pra),
		   EpsilonPhysicalRegisterIsValid(epe, iis->prb),
		   EpsilonPhysicalRegisterIsValid(epe, iis->prc));
    }
    if (!iis->trapped
	&& EpsilonPhysicalRegisterIsValid(epe, iis->pra)
	&& EpsilonPhysicalRegisterIsValid(epe, iis->prb)
	&& EpsilonPhysicalRegisterIsValid(epe, iis->prc)) {
      if (EpsilonInstructionIsMemop(cpu, inumInstid(inum))
	  && unreadyLoadOrStore)
	continue;
      iis->readyToIssue = 1;
      issueCount++;

      if (issueCount >= epsilonParams.width)
	break;
    } else {
      if (EpsilonInstructionIsMemop(cpu, inumInstid(inum)))
	unreadyLoadOrStore = 1;
    }
  }
}

static void EpsilonExecCycle(AlphaState *pe, EpsilonState *epe)
{
  int cpu = pe->myNum;
  Inum inum;
  Inum lastInumFetched = epe->lastInumFetched;
  Inum lastInumCommitted = epe->lastInumCommitted;
  int issuedCount = 0;
  for (inum = lastInumCommitted+1; inum <= lastInumFetched; inum++) {
    InflightInstructionState *iis = &epe->iqueue[inumInstid(inum)];
    if (iis->readyToIssue && !iis->issued && !iis->trapped) {
      VA predictedNextPC = iis->nextpc;
      EpsilonIssueInstruction(cpu, inum);
      issuedCount++;
      /* check next pc and trap if we fetched down badpath */
      if (iis->nextpc != predictedNextPC) {
	EpsilonTrapNotify(cpu, iis->inum, EpsilonTTMispredict);
        iis->trapfree = 0;
      }
      if (issuedCount > epsilonParams.width)
	break;
    }
  }
}

#ifdef EPSILON_USE_MEMREF
static void EpsilonRequestMemopsCycle(AlphaState *pe, EpsilonState *epe)
{
  int cpu = pe->myNum;
  Inum inum;
  Inum lastInumFetched = epe->lastInumFetched;
  Inum lastInumCommitted = epe->lastInumCommitted;
  Inum lastInumUntrapped = epe->lastInumUntrapped;
  Inum trapPointInum = epe->trapPointInum; /* @@@@ compute trap point */
  int issuedCount = 0;

  if (epsilonParams.debugScheduler && epe->pendingMemops)
    CPUWarning("EpsilonRequestMemopsCycle: cpu=%d pendingMemops=%d\n", cpu, epe->pendingMemops);
  if (epe->pendingMemops == 0) {
    for (inum = lastInumCommitted+1; inum <= lastInumFetched; inum++) {
      InflightInstructionState *iis = &epe->iqueue[inumInstid(inum)];
      if (iis->trapped | iis->memopRequested)
	continue;
      if (iis->isMemop) {
         if (!iis->issued) {
            goto memopLimit;
         }
	if (epsilonParams.debugScheduler)
	  CPUWarning("EpsilonRequestMemopsCycle: cpu=%d pc=%lx inum=%ld trapPoint=%ld opcode=0x%x dmmuStatus=%d\n",
		     cpu, iis->PC, inum, trapPointInum, iis->instr.common.opcode, 
		     iis->dmmuStatus);

	if (inum > trapPointInum && EpsilonInstructionIsStore(cpu, inumInstid(iis->inum)))
	  goto memopLimit;

	switch (iis->dmmuStatus) {
	case MMU_SUCCESS:
	issueRequest: {
	  EpsilonRequestMemop(cpu, inum);
	  if (epsilonParams.debugScheduler)
	    CPUWarning("EpsilonRequestMemopsCycle: \t memrefStatus=%d\n", iis->memrefStatus);
	  if (iis->memrefStatus == SUCCESS)
	    iis->memopReadyToComplete = 1;
	  else
	    epe->pendingMemops++;
	  issuedCount++;
	  goto memopLimit;
	} break;
	case MMU_UNCACHED:
	  /* have to wait until we know it's not going to trap */
	  if (inum <= trapPointInum)
	    goto issueRequest;
	  else
	    goto memopLimit;
	  break;
	case MMU_EXCEPTION:
	  /* do nothing */
	  break;
	}
	if (issuedCount > epsilonParams.width)
	  break;
      }
    }
  }
memopLimit:
  ;
}

static void EpsilonCompleteMemopsCycle(AlphaState *pe, EpsilonState *epe)
{
  int cpu = pe->myNum;
  Inum inum;
  Inum lastInumFetched = epe->lastInumFetched;
  Inum lastInumDone = epe->lastInumDone;
  Inum trapPointInum = epe->trapPointInum;
  /* 
   * @@@ very simple strategy:
   *   do the memops in order
   *   wait until all previous instructions have issued so we don't trap
   */
  for (inum = lastInumDone+1; inum <= lastInumFetched; inum++) {
    InflightInstructionState *iis = &epe->iqueue[inumInstid(inum)];

    if (iis->trapped | iis->done | iis->memopCompleted)
      continue;

    /* if we have any unissued instructions then we could potentially trap, 
     * so don't complete later memops 
     */
    if (inum > trapPointInum)
      break;
    /* complete memops in order */
    if (iis->isMemop) {

      if (!iis->memopReadyToComplete) {
	break;
      } else {
	EpsilonCompleteMemop(cpu, inum);
      }
    }
  }
}

#else /* EPSILON_USE_MEMREF */
static void EpsilonCompleteMemopsCycle(AlphaState *pe, EpsilonState *epe)
{
#if 0
  int cpu = pe->myNum;
  Inum inum;
  Inum lastInumFetched = epe->lastInumFetched;
  Inum lastInumDone = epe->lastInumDone;
  /* 
   * @@@ very simple strategy:
   *   do the memops in order
   *   do the stores after they are committed
   *   do the loads after issue and after all previous memops have issued
   */
  for (inum = lastInumDone+1; inum <= lastInumFetched; inum++) {
    InflightInstructionState *iis = &epe->iqueue[inumInstid(inum)];
    if (!iis->trapped && EpsilonInstructionIsMemop(cpu, inumInstid(inum))) {
      /* for stores, do them at commit time */ 
      if (EpsilonInstructionIsStore(cpu, inumInstid(inum))
	  && !iis->done) {
	if (iis->readyToRetire || iis->committed) {
	  EpsilonCompleteMemop(cpu, inum);
	} else {
	  /* can't process any more loads or stores this cycle unless we compare addresses */
	  break;
	}
      }
      if (EpsilonInstructionIsLoad(cpu, inumInstid(inum))
	  && !iis->done) {
	if (iis->readyToRetire || iis->committed)
	  EpsilonCompleteMemop(cpu, inum);
	else 
	  break;
      }
    }
  }
#endif
}
#endif /* EPSILON_USE_MEMREF */

/**
 * @@@ TrapPointInum Invariants:
 *  1. forall inum s.t. iis->untrappable, trapPointInum >= inum
 *  2. forall inum s.t. !iis->trapfree, trapPointInum < inum
 *  3. trapPointInum <= trapAbortInum if (trapPending)
 */

static void EpsilonRetireCycle(AlphaState *pe, EpsilonState *epe)
{
   int cpu = pe->myNum;
   Inum inum;
   Inum lastInumFetched = epe->lastInumFetched;
   Inum lastInumCommitted = epe->lastInumCommitted;
   Inum trapPointInum = lastInumFetched;
   int count = 0;
   int width = epsilonParams.width;
   int uncommittedInstruction = 0;
   for (inum = lastInumCommitted+1; inum <= lastInumFetched; inum++) {
      int nomore = 0;
      InflightInstructionState *iis = &epe->iqueue[inumInstid(inum)];
      /* @@@ traps can arise during EpsilonCommitInstruction because of annotations 
       * modifying the PC or other processor state 
       */
      if (epe->trapPending && epe->trapAbortInum <= inum)
         break;
      if (iis->issued)
         iis->readyToRetire = 1;
      if (iis->trapped) {
         epe->lastInumCommitted = inum;
      } else if (iis->readyToRetire && (!iis->isMemop || iis->memopCompleted)) {
#ifndef EPSILON_USE_MEMREF
         if (EpsilonInstructionIsMemop(cpu, inumInstid(inum)))
            EpsilonCompleteMemop(cpu, inum);
#endif /* EPSILON_USE_MEMREF */
         EpsilonCommitInstruction(cpu, inum);
         count++;
         epe->lastInumCommitted = inum;
         epe->trapPointInum = inum;
      } else {
         uncommittedInstruction = 1; /* force break */
      }
      if (!iis->trapfree && inum <= trapPointInum)
         trapPointInum = inum-1;
      if (uncommittedInstruction || count > width)
         break;
   }
   if (epe->trapPending && trapPointInum >= epe->trapAbortInum)
      trapPointInum = epe->trapAbortInum - 1;
   epe->trapPointInum = trapPointInum;
}

void EpsilonCPUCycle()
{
  int cpu;
  for (cpu = 0; (curPE = PE[cpu]) != NULL; cpu++) {
    EpsilonState *epe = EPE[cpu];
    AlphaState *pe = curPE;

    EpsilonTrapCycle(pe, epe);
    EpsilonDebugCycle(pe, epe);

    if (pe->cpuState != cpu_running) {
      continue;
    }

    if (epsilonParams.pipelined) {
      EpsilonCompleteMemopsCycle(pe, epe);
#ifdef EPSILON_USE_MEMREF
      EpsilonRequestMemopsCycle(pe, epe);
#endif /* EPSILON_USE_MEMREF */
      EpsilonRetireCycle(pe, epe);
      EpsilonExecCycle(pe, epe);
      EpsilonSchedCycle(pe, epe);
      EpsilonFetchCycle(pe, epe);
    } else {
      EpsilonFetchCycle(pe, epe);
      EpsilonSchedCycle(pe, epe);
      EpsilonExecCycle(pe, epe);
#ifdef EPSILON_USE_MEMREF
      EpsilonRequestMemopsCycle(pe, epe);
#endif /* EPSILON_USE_MEMREF */
      EpsilonCompleteMemopsCycle(pe, epe);
      EpsilonRetireCycle(pe, epe);
    }

  }
  cpu = 0; curPE = PE[cpu];
}


static void EpsilonCPURun(void)
{
  int cpu;
  nRunning = 0;
  for (cpu = 0; (curPE = PE[cpu]) != NULL; cpu++) {
    EpsilonState *epe = EPE[cpu];
    EpsilonCopyStateIn(curPE, epe);
    nRunning++;
  }
  lic = 0;
  lastCycleInstructionCommitted = gammaCurrentTime[0];
  PMINT_COPYIN();
  while (nRunning > 0) {

    EpsilonPreCycle();
    if (epsilonParams.usePmintCPUCycle) {
      PMINT_CPU_CYCLE();
    } else {
      EpsilonCPUCycle();
    }
    EpsilonPostCycle();
  }
  /* unhalt the CPUs */
  for (cpu = 0; (curPE = PE[cpu]) != NULL; cpu++) {
    AlphaState *pe = curPE;
    pe->cpuState = cpu_running;
  }
  PMINT_COPYOUT();
  for (cpu = 0; (curPE = PE[cpu]) != NULL; cpu++) {
    AlphaState *pe = curPE;
    EpsilonState *epe = &epeArray[cpu];
    EpsilonCopyStateOut(pe, epe);
  }
}

static void EpsilonDoneRunning(void)
{
  
  CPUWarning("Exiting the Epsilon simulator \n");
}
 


static regname_t lrmap[80] = {
   0,  1,  2,  3,  4,  5,  6,  7,
  8+64, 9+64, 10+64, 11+64, 12+64, 13+64, 14+64, 15,
  16, 17, 18, 19, 20, 21, 22, 23,
  24, 15+64, 26, 27, 28, 29, 30, 31,
  32, 33, 34, 35, 36, 37, 38, 39,
  40, 41, 42, 43, 44, 45, 46, 47, 
  48, 49, 50, 51, 52, 53, 54, 55,
  56, 57, 58, 59, 60, 61, 62, 63,
  64, 65, 66, 67, 68, 69, 70, 71, /* unused */
  /* now map shadow regs back to original regs */
   8,  9, 10, 11, 12, 13, 14, 25
};

/* registers 8, 9, 10, 11, 12, 13, 14 and 25 are shadowed on EV5 */
#define SHADOWED_REGISTER_MASK 02007F00
#define IS_SHADOWED_REGISTER(lr) ((SHADOWED_REGISTER_MASK >> lr) &1)
#define LOGICAL_SHADOW_REGISTER(lr) lrmap[lr]

Inum EpsilonFetchInstruction(int cpu, VA predictedPC)
{
  AlphaState *pe = PE[cpu];
  PA instrPAddr=0;
  MA instrMAddr;
  union alpha_instruction instr;
  int opcode;
  MMUStatus mmuStatus;
  EpsilonState *epe;
  Inum inum;
  Instid instid;
  InflightInstructionState *iis;

  curPE = pe;
  ASSERT(pe->cpuState == cpu_running);
  epe = EPE[cpu];

  inum = epe->nextInum++;
  instid = inumInstid(inum);
  iis = &epe->iqueue[instid];
  if (iis->inflight) {
    /* looks like we wrapped around */
    PrintAllInflightInstructions(epe);
  }
  ASSERT(!iis->inflight); /* wrapped around in iqueue */
  bzero((void*)iis, sizeof(InflightInstructionState));
  iis->inum = inum;
  { /* make the inorder links between this instruction and previous one */
    Inum prevInum = epe->lastInumUntrapped;
    InflightInstructionState *prevIIS = &epe->iqueue[inumInstid(prevInum)];
    ASSERT(prevInum <= epe->lastInumCommitted || !prevIIS->trapped);
    iis->inorderNext = inum+1;
    iis->inorderPrev = prevInum;
    prevIIS->inorderNext = inum;
  }
  epe->lastInumFetched = inum;
  epe->lastInumUntrapped = inum;
  iis->inflight = 1;
  iis->fetched = 1;
  iis->lrdest = ZERO_REGISTER; /* hasn't been renamed yet */
  epe->useShadowRegisters = pe->useShadowRegisters;

  pe->PC = predictedPC;
  pe->trapState = iis->trapState = epe->newestTrapState;
  iis->trapState->useShadowRegisters = epe->useShadowRegisters;
  ASSERT(iis->trapState != NULL);

  if (IS_PAL(pe)) {
    /*
     * PALCode: fetched from physical memory
     */
    instrPAddr =  pe->PC & ~0x3;
    instrMAddr = PHYS_TO_MEMADDR(pe->myNum,instrPAddr);
    STATS_INC(pe->myNum, iReads, 1);
    mmuStatus = MMU_SUCCESS;
  } else if ((instrMAddr = QCLookup(pe->curIQC,pe->PC,0)) != NULL) {
    /*
     * Fast access
     */
    STATS_INC(pe->myNum, iReads, 1);
    mmuStatus = MMU_SUCCESS;
  } else {
    /*
    * Fall-back case 
    */
    mmuStatus = EV5_ITranslateVirtual(pe,&instrPAddr,0);
    if (mmuStatus == MMU_SUCCESS) {
      instrMAddr = PHYS_TO_MEMADDR(pe->myNum,instrPAddr);     
      STATS_INC(pe->myNum, iReads, 1);
    } else {
      EpsilonTrapNotify(cpu, iis->inum, EpsilonTTIMMUException);
    }
  }

  if (mmuStatus==MMU_SUCCESS)
    instr = *(union alpha_instruction *)instrMAddr;
  else
    instr.word = 0;
  iis->immuStatus = mmuStatus;
  if (mmuStatus!=MMU_SUCCESS) {
    ASSERT( mmuStatus == MMU_EXCEPTION);
  }

  pe->instr = instr;
  gdbInstr = instr.word;
  gdbInstrCommon = instr.common.opcode;

  {
    /* fill in basic state */
    iis->PC = predictedPC;
    iis->instr = instr;
    iis->fetched = 1;
    /* do renaming */
    {
      regname_t lrdest = ZERO_REGISTER;
      regname_t lra = ZERO_REGISTER;
      regname_t lrb = ZERO_REGISTER;
      regname_t lrc = ZERO_REGISTER;

      opcode = instr.common.opcode;
      switch (opcode) {
      case op_call_pal:
	/* no registers mapped */
	/* @@@ we do the trap notify here */
	if (mmuStatus == MMU_SUCCESS)
	  EpsilonTrapNotify(cpu, iis->inum, EpsilonTTPalCall);
	break;
      case op_ldq_u:
#ifdef op_ldbu
      case op_ldbu:
      case op_ldwu:
#endif
      case op_ldl:
      case op_ldq:
      case op_ldl_l:
      case op_ldq_l:
	iis->isMemop = 1;
      case op_lda: 
      case op_ldah: 
	lrb = instr.m_format.rb;
	lrdest = instr.m_format.ra;
	break;

      case op_ldf:
      case op_ldg:
      case op_lds:
      case op_ldt:
	iis->isMemop = 1;
	lrb = instr.m_format.rb;
	lrdest = instr.m_format.ra+32;

	break;

#ifdef op_stb
      case op_stb:
      case op_stw:
#endif
      case op_stl:
      case op_stq:
      case op_stq_u:
      case op_stl_c:
      case op_stq_c:
	iis->isMemop = 1;
	iis->lastStoreInum = epe->lastStoreInum;
	epe->lastStoreInum = inum;
	lra = instr.m_format.ra;
	lrb = instr.m_format.rb;
	break;

      case op_stf:
      case op_stg:
      case op_sts:
      case op_stt:
	iis->isMemop = 1;
	iis->lastStoreInum = epe->lastStoreInum;
	epe->lastStoreInum = inum;
	lra = instr.m_format.ra+32;
	lrb = instr.m_format.rb;
	break;

      case op_br:
      case op_bsr:
	lrdest = instr.b_format.ra;
	break;
      case op_fbeq:
      case op_fblt:
      case op_fble:
      case op_fbne:
      case op_fbge:
      case op_fbgt:
	lra = instr.b_format.ra+32;
	break;
      case op_blbc:
      case op_beq:
      case op_blt:
      case op_ble:
      case op_blbs:
      case op_bne:
      case op_bge:
      case op_bgt:
	lra = instr.b_format.ra;
	break;

      case op_inta:
      case op_intl:
      case op_ints:
      case op_intm:
	lra = instr.o_format.ra;
	lrb = instr.o_format.rb;
	lrc = instr.o_format.rc;
	lrdest = instr.o_format.rc;
	break;

      case op_fltv:
      case op_flti:
      case op_fltl:
	lra = instr.f_format.fa+32;
	lrb = instr.f_format.fb+32;
	lrc = instr.f_format.fc+32;
	lrdest = instr.f_format.fc+32;
	break;

#ifdef op_fltm
      case op_fltm: {
	int func = instr.f_format.function;
	if (func == fltm_itofs || func == fltm_itoff || func == fltm_itoft) {
	  lra = instr.f_format.fa;
	  lrdest = instr.f_format.fc+32;
	} else {
	  lra = instr.f_format.fa+32;
	  lrdest = instr.f_format.fc+32;
	}
      } break;
#endif /* op_fltm */

#ifdef op_intmisc
      case op_intmisc: {
#ifdef intmisc_ftoit
	int func = instr.o_format.function;
	if (func < intmisc_ftoit) {
	  lra = instr.o_format.ra;
	  lrb = instr.o_format.rb;
	  lrc = instr.o_format.rc;
	  lrdest = instr.o_format.rc;
	} else if (func == intmisc_ftoit || func == intmisc_ftois) {
	  lra = instr.o_format.ra+32;
	  lrdest = instr.o_format.rc;
	} else {
	  ASSERT(0);
	}
#else
	lra = instr.o_format.ra;
	lrb = instr.o_format.rb;
	lrc = instr.o_format.rc;
	lrdest = instr.o_format.rc;
#endif /* intmisc_ftoit */
      } break;
#endif /* op_intmisc */

      case op_jsr:
	lrb = instr.j_format.rb;
	lrdest = instr.j_format.ra;
	break;
      case op_misc: {
	int func = instr.m_format.memory_displacement & 0xFFFF;
	switch (func) {
	case misc_rc:
	case misc_rs:
	case misc_rpcc:
	  lrdest = instr.m_format.ra;
	  break;
	default:
	  /* no regs */
	  break;
	}
      } break;

#ifndef SOLO
      case PRIV_OP_HW_LD: 
	iis->isMemop = 1;
	/* fall through to next case */
      case PRIV_OP_MFPR:
	lra = instr.m_format.ra;
	lrb = instr.m_format.rb;
	lrdest = instr.m_format.ra;
	break;
      case PRIV_OP_HW_ST: 
	iis->isMemop = 1;
	iis->lastStoreInum = epe->lastStoreInum;
	epe->lastStoreInum = inum;
	/* fall through to next case */
      case PRIV_OP_MTPR:
	lra = instr.m_format.ra;
	lrb = instr.m_format.rb;
	break;
      case PRIV_OP_REI:
	/* uses no registers */
	/* @@@ we do the trap notify here */
	if (mmuStatus == MMU_SUCCESS)
	  EpsilonTrapNotify(cpu, iis->inum, EpsilonTTHwRei);
	break;
#endif

      default:
	/* might speculate into non-code area */
	CPUWarning("EpsilonFetchInstruction: cpu=%d pc=%lx inum=%ld instr=%x opcode=%x (illop)\n",
		   cpu, predictedPC, iis->inum, instr.word, instr.common.opcode);
	EpsilonTrapNotify(cpu, iis->inum, EpsilonTTIllegalOpcode);
      } /* switch(opcode) */

      if (EV5_ShadowRegistersUsed(pe, pe->PC)) {
	lra = LOGICAL_SHADOW_REGISTER(lra);
	lrb = LOGICAL_SHADOW_REGISTER(lrb);
	lrc = LOGICAL_SHADOW_REGISTER(lrc);
	lrdest = LOGICAL_SHADOW_REGISTER(lrdest);
      }

      iis->pra = epe->regmap[lra];
      iis->prb = epe->regmap[lrb];
      iis->prc = epe->regmap[lrc];
      if (lrdest == (ZERO_REGISTER+32))
	lrdest = ZERO_REGISTER;
      iis->lrdest = lrdest;
      if (lrdest == ZERO_REGISTER || lrdest == (ZERO_REGISTER+32)) {
	iis->lrdest = ZERO_REGISTER;
	iis->oldprdest = iis->prdest = ZERO_REGISTER_DEST_PHYSREG;
      } else if (lrdest >= 0 && lrdest < N_LOGICAL_REGISTERS){
	regname_t oldprdest = epe->regmap[lrdest];
	regname_t prdest = epsilonPhysregAlloc(epe);
	iis->oldprdest = oldprdest;
	iis->prdest = prdest;
	epe->lastInumMapped = iis->inum;
	epe->regmap[lrdest] = prdest;
	EpsilonPhysicalRegisterClearValid(epe, prdest);
      } else {
	iis->lrdest = ZERO_REGISTER;
	iis->prdest = iis->oldprdest = ZERO_REGISTER_DEST_PHYSREG;
      }
    }
    ASSERT(iis->prdest >= 0); ASSERT(iis->prdest < N_PHYSICAL_REGS);
    ASSERT(iis->pra >= 0); ASSERT(iis->pra < N_PHYSICAL_REGS);
    ASSERT(iis->prb >= 0); ASSERT(iis->prb < N_PHYSICAL_REGS);
    ASSERT(iis->prc >= 0); ASSERT(iis->prc < N_PHYSICAL_REGS);

    /* print useful information */
    if (epsilonParams.debugScheduler)
      CPUWarning("EpsilonFetch: cpu=%d pc=%lx inum=%ld instr=%x opcode=%x mmuStatus=%d prdest=%d oldprdest=%d\n",
		 cpu, predictedPC, iis->inum, instr.word, instr.common.opcode,
		 mmuStatus, iis->prdest, iis->oldprdest);

    return inum;
  }
}


void EpsilonIssueInstruction (int cpu, Inum inum)
{
  AlphaState *pe = PE[cpu];
  EpsilonState *epe = EPE[cpu];
  Instid instid = inumInstid(inum);
  InflightInstructionState *iis = &epe->iqueue[instid];
  union alpha_instruction instr = iis->instr;
  Reg rav, rbv, *raPtr, *rcPtr;
  int opcode, ret, disp;
  VA nextpc = -1;
  int valid = 0;
  int trapfree = 1; /* most instructions are trapfree after they issue */

  if (epsilonParams.debugScheduler)
    CPUWarning("EpsilonIssue: cpu=%d pc=%lx inum=%ld instr=%x opcode=0x%x\n",
	       cpu, iis->PC, iis->inum, iis->instr.word, iis->instr.common.opcode);

  ASSERT(iis->fetched);
  iis->issued = 1;
  pe->PC = iis->PC;
  pe->instr = iis->instr;
  pe->trapState = iis->trapState;

  ASSERT(EpsilonPhysicalRegisterIsValid(epe, iis->pra));
  ASSERT(EpsilonPhysicalRegisterIsValid(epe, iis->prb));
  ASSERT(EpsilonPhysicalRegisterIsValid(epe, iis->prc));


  /* @@@@
   * if we are in PAL mode, then we need to verify that the previous instruction has committed, 
   * because the IPR's are not renamed.
   * In general, the scheduler could use a scoreboard mechanism, 
   * but ev5.c does not implement this yet. 
   */

  pe->instr = instr;
  gdbInstr = instr.word;
  gdbInstrCommon = instr.common.opcode;
    
  opcode = instr.common.opcode;
  switch (opcode) {
  case op_call_pal:
    /* set at fetch time */
    nextpc = iis->nextpc;
    valid = 1;
    break;
  case op_lda: 
  case op_ldah: 
    rbv = epe->regs[iis->prb];
    disp = instr.m_format.memory_displacement;
    raPtr = &epe->regs[iis->prdest];
    if (opcode == op_lda) { 
      *raPtr = rbv + disp;
    } else {
      *raPtr = rbv + (disp<<16);
    }
    nextpc = iis->PC + 4;
    valid = 1;
    break;

  case op_ldf:
  case op_ldg:
  case op_lds:
  case op_ldt:
    if (!FPENABLED(pe)) {
      EpsilonTrapNotify(cpu, inum, EpsilonTTFEN);
      trapfree = 0;
      nextpc = iis->PC + 4;
    } else {
      goto issue_load;
    } break;
  case op_ldq_u:
#ifdef op_ldbu 
  case op_ldbu:
  case op_ldwu:
#endif
  case op_ldl:
  case op_ldq:
  case op_ldl_l:
  case op_ldq_l: 
  issue_load: {
    PA paddr;
    MA maddr;
    /* @@@@ check MMU_SUCCESS */
    rbv = epe->regs[iis->prb];
    disp = instr.m_format.memory_displacement;
    iis->eaddr = rbv+disp;
    /* @@@@ need to do something about this to make the storebuffer work @@@@ */
    if (0 && opcode == op_ldq_u) {
      iis->instr.common.opcode = op_ldq;
      iis->eaddr = iis->eaddr & ~0x7L;
    }
    iis->dmmuStatus = EpsilonResolveEffectiveAddress(pe, iis->inum, iis->eaddr, 0, 0, &paddr, &maddr);
    if (iis->dmmuStatus == MMU_EXCEPTION)
       trapfree = 0;
    iis->paddr = paddr;
    iis->maddr = maddr;
    nextpc = iis->PC + 4;
  } break;

  case op_stf:
  case op_stg:
  case op_sts:
  case op_stt:
    if (!FPENABLED(pe)) {
      EpsilonTrapNotify(cpu, inum, EpsilonTTFEN);
      nextpc = iis->PC + 4;
    } else {
      goto issue_store;
    } break;
  case op_stq_u:
#ifdef op_stb
  case op_stb:
  case op_stw:
#endif
  case op_stl:
  case op_stq:
  case op_stl_c:
  case op_stq_c: 
  issue_store: {
    PA paddr;
    MA maddr;
    /* @@@@ check MMU_SUCCESS */
    rbv = epe->regs[iis->prb];
    disp = instr.m_format.memory_displacement;
    iis->eaddr = rbv+disp;
    iis->value = epe->regs[iis->pra];
    /* @@@@ need to do something about this to make the storebuffer work @@@@ */
    if (0 && opcode == op_stq_u) {
      iis->instr.common.opcode = op_stq;
      iis->eaddr = iis->eaddr & ~0x7L;
    }
    iis->dmmuStatus = EpsilonResolveEffectiveAddress(pe, iis->inum, iis->eaddr, 1, 0, &paddr, &maddr);
    iis->paddr = paddr;
    iis->maddr = maddr;
    nextpc = iis->PC + 4; 
    if (iis->dmmuStatus == MMU_EXCEPTION)
       trapfree = 0;
    EpsilonStoreBufferWrite(pe, epe, iis, iis->value);
  } break;
  case op_fbeq:
  case op_fblt:
  case op_fble:
  case op_fbne:
  case op_fbge:
  case op_fbgt:
    if (!FPENABLED(pe)) {
      EpsilonTrapNotify(cpu, inum, EpsilonTTFEN);
      nextpc = iis->PC + 4;
      trapfree = 0;
    } else {
      goto issue_br;
    } break;
  case op_br:
  case op_bsr:
  case op_blbc:
  case op_beq:
  case op_blt:
  case op_ble:
  case op_blbs:
  case op_bne:
  case op_bge:
  case op_bgt:
  issue_br:
    rav = epe->regs[iis->pra];
    if (GammaBranchOpcode(pe,opcode,&epe->regs[iis->prdest],rav,*(double*)&rav)) {
      disp = instr.b_format.branch_displacement;
      nextpc = iis->PC + (disp <<2)+4;
      valid = 1;
    } else {
      nextpc = iis->PC + 4;
      valid = 1;
    }	 	   
    if (epsilonParams.debugScheduler) {
       CPUWarning("    rav=%lx npc=%lx ftpc=%lx\n", rav, nextpc, iis->PC+4);
    }
    break;
  case op_inta:
  case op_intl:
  case op_ints:
  case op_intm:
    rav = epe->regs[iis->pra];
    if (instr.o_format.form) {
      /* literal */
      rbv = instr.l_format.literal;
    } else {
      rbv = epe->regs[iis->prb];
    }
    rcPtr = &epe->regs[iis->prc];
    ret = GammaOperateOpcode(pe,
			     opcode*(1<<7)+instr.o_format.function,
			     rav,rbv,rcPtr,&epe->regs[iis->prdest]);
    if (GammaFPCRIndicatesTrap(pe)) {
      EpsilonTrapNotify(cpu, inum, EpsilonTTArith);
      trapfree = 0;
    } else {
      nextpc = iis->PC + 4;
      valid = 1;
    }
    break;
  case op_opc14:
  case op_fltv:
  case op_flti:
  case op_fltl:
    if (!FPENABLED(pe)) {
      EpsilonTrapNotify(cpu, inum, EpsilonTTFEN);
      trapfree = 0;
      nextpc = iis->PC + 4;
    } else {
      GammaFloatingOpcode(pe,opcode,instr.f_format.function,
			  (double*)&epe->regs[iis->pra],
			  (double*)&epe->regs[iis->prb],
			  (double*)&epe->regs[iis->prc],
			  (double*)&epe->regs[iis->prdest]);
      if (GammaFPCRIndicatesTrap(pe)) {
	EpsilonTrapNotify(cpu, inum, EpsilonTTArith);
        trapfree = 0;
      } else {
	nextpc = iis->PC + 4;
	valid = 1;
      }
    } break;

  case op_jsr:
    { 
      /*
       * XXX be really careful of case when ra==rb
       * XXX AXP manual page 4-21
       */
      VA savedPC = iis->PC+4;
      regname_t prb = iis->prb;
      VA targetPC = epe->regs[prb];
      if (((targetPC >> 43) != 0) && ((targetPC >> 60) != 0xf)) {
	targetPC &= ~3L;
	if ((long)targetPC >= 0L)
	  targetPC &= VA_MASK;
	CPUWarning("JSR TARGET unaligned: cpu=%d pc=%lx inum=%ld prb=%d target=%lx (now %lx)\n",
		   cpu, iis->PC, iis->inum, prb, epe->regs[prb], targetPC);
      }
      if IS_PAL(pe) {
	nextpc = epe->regs[prb] | 0x1;
      } else {
	nextpc = epe->regs[prb];
      }
      epe->regs[iis->prdest] = savedPC;
      pe->reg[REG_RA] = savedPC;
      valid = 1;
      break;
    }
  case op_misc:
    GammaMiscOpcode(pe,instr.word, &epe->regs[iis->prdest]);
    nextpc = iis->PC + 4;
    valid = 1;
    break;

#ifndef SOLO
  case PRIV_OP_MTPR:
    {
      /* happens at commit time, rather than having shadow state */
      nextpc = iis->PC + 4;
      valid = 1;
      break;
    }
  case PRIV_OP_MFPR:
    {
      /* @@@@ happens at commit time because some MFPR's have a side-effect */
      nextpc = iis->PC + 4;
      valid = 1;
      break;
    }
  case PRIV_OP_HW_LD: 
    {
      /* actual memory operation is performed by EpsilonCompleteMemop */
      ev5_hwldst_format hwldst;

      hwldst.word = instr.word;
      /* @@@@ check MMU_SUCCESS */
      /* EV5_HW_Load recomputes effective address, so we preserve rbv in iis->eaddr */
      iis->eaddr = epe->regs[iis->prb];
      if (hwldst.hwmem_format.phys) {
	iis->dmmuStatus = MMU_SUCCESS;
      } else {
	PA paddr;
	MA maddr;
	iis->dmmuStatus = EpsilonResolveEffectiveAddress(pe, inum, iis->eaddr, 0, 1, &paddr, &maddr);
      }
      nextpc = iis->PC + 4;
      valid = 0; /* @@@@ register not valid until commit time */
      if (iis->dmmuStatus == MMU_EXCEPTION)
         trapfree = 0;
      break;
    }
  case PRIV_OP_HW_ST: 
    {
      /* actual memory operation is performed by EpsilonCompleteMemop */
      /* @@@@ check MMU_SUCCESS -- looks like no exceptions for this instruction */
      /* EV5_HW_Store recomputes effective address, so we preserve rbv in iis->eaddr */
      iis->eaddr = epe->regs[iis->prb];
      iis->value = epe->regs[iis->pra];
      iis->dmmuStatus = MMU_SUCCESS;
      nextpc = iis->PC + 4;
      valid = 0;
      break;
    }
  case PRIV_OP_REI:
    /* @@@@ -- We compute nextpc at fetch time */
    nextpc = iis->nextpc;
    valid = 1;
    break;
#endif

  default:
    /* might speculate into non-code area */
    if (epsilonParams.debugScheduler)
      CPUWarning("EpsilonIssueInstruction: cpu=%d pc=%lx instr=%x opcode=%x -- bogus opcode\n",
		 cpu, iis->PC, instr.word, instr.common.opcode);
    trapfree = 0;

  } /* switch(opcode) */

  /* some book-keeping */
  ASSERT(nextpc != -1);
  iis->nextpc = nextpc;
  iis->issued = 1;
  iis->trapfree = trapfree;
  if (valid)
    EpsilonPhysicalRegisterSetValid(epe, iis->prdest);
  if (iis->dmmuStatus == MMU_EXCEPTION) {
    if (epsilonParams.debugScheduler)
      CPUWarning("EpsilonIssueInstruction: dmmuStatus=%d ea=%lx pa=%lx\n", 
		 iis->dmmuStatus, iis->eaddr, iis->paddr);
  }
}

/* only destination register may be used here, 
 * because other physical registers may have been freed by EpsilonCommitInstruction 
 */
void EpsilonRequestMemop(int cpu, Inum inum)
{
  AlphaState *pe = curPE;
  EpsilonState *epe = EPE[cpu];
  Instid instid = inumInstid(inum);
  InflightInstructionState *iis = &epe->iqueue[instid];
  union alpha_instruction instr = iis->instr;
  Reg *raPtr = &epe->regs[iis->pra];
  VA vAddr = iis->eaddr;
  PA pAddr = iis->paddr;
  MA mAddr = iis->maddr;
  int opcode = iis->instr.common.opcode;
  RefSize refSize = alphaRefSizeMap[opcode];
  RefFlavor refFlavor = alphaRefFlavorMap[opcode];

  if (epsilonParams.debugScheduler)
    CPUWarning("EpsilonRequestMemop: cpu=%d pc=%lx inum=%ld instr=%x opcode=0x%x ea=%lx pa=%lx ma=%lx\n",
	       cpu, iis->PC, iis->inum, iis->instr.word, opcode,
	       iis->eaddr, iis->paddr, mAddr);

  ASSERT(iis->issued);
  pe->PC = iis->PC;
  pe->instr = iis->instr;
  pe->trapState = iis->trapState;

  pe->instr = instr;
  gdbInstr = instr.word;
  gdbInstrCommon = instr.common.opcode;
    
  switch (opcode) {
  case op_ldq_u: vAddr &= ~7L; pAddr &= ~7L; refSize = DOUBLE_SZ; goto requestLoad;
#ifdef op_ldbu
  case op_ldbu:
  case op_ldwu:
#endif
  case op_ldf:
  case op_ldg:
  case op_lds:
  case op_ldt:
  case op_ldl:
  case op_ldq:
  case op_ldl_l:
  case op_ldq_l:
  requestLoad: {
    Reg mdata = 0;
    Result memrefStatus;
    if (iis->dmmuStatus == MMU_SUCCESS) {
      memrefStatus = MemRefReadData(pe->myNum, vAddr, pAddr, &mdata, refSize, refFlavor);
      iis->memrefStatus = memrefStatus;
      switch (memrefStatus) {
      case SUCCESS:
         iis->value = mdata; /* stash the value for later */
         iis->memopRequested = 1;
         iis->memopReadyToComplete = 1;
         break;
      case STALL:
	iis->memopRequested = 1;
	epe->stalledMemop = inum;
	break;
      case FAILURE:
	iis->memopRequested = 0; /* pretend we didn't request it */
	break;
      case BUSERROR:
	NOTREACHED();
	break;
      default:
	NOTREACHED();
	break;
      }
    } else {
      /* @@@@ UNCACHED -- call GammaMemoryReadOpcode at memop completion time */
      iis->memopRequested = 1;
    }
  } break;

  case op_stq_u: vAddr &= ~7L; pAddr &= ~7L; goto requestStore;
#ifdef op_stb
  case op_stb:
  case op_stw:
#endif
  case op_stf:
  case op_stg:
  case op_sts:
  case op_stt:
  case op_stl:
  case op_stq:
  case op_stl_c:
  case op_stq_c:
  requestStore: {
    /* @@@@ fix the storebuffer -- works with inorder commits */
    if (iis->lastStoreInum > epe->lastInumCommitted && !epe->iqueue[inumInstid(iis->lastStoreInum)].memopRequested) {
      InflightInstructionState *lsiis = &epe->iqueue[inumInstid(iis->lastStoreInum)];
      ASSERT(lsiis->inum == iis->lastStoreInum);
      CPUWarning("EpsilonRequestMemop: OutOfOrder -- cpu=%d pc=%lx inum=%d instr=%x opcode=0x%x lastStoreInum=%ld \n",
		 cpu, iis->PC, iis->inum, iis->instr.word, iis->instr.common.opcode, iis->lastStoreInum);
    }
    if (iis->dmmuStatus == MMU_SUCCESS) {
      Result memrefStatus;

      CHECK_ST_ANN(vAddr,pAddr);
      STATS_INC(pe->myNum, dWrites, 1);
      TRACE_ENTRY(TRACE_STORE,curPE->myNum,vAddr);

      iis->untrappable = 1; /* not allowed to trap once we issue the store request? */
      memrefStatus = MemRefWriteData(pe->myNum, vAddr, pAddr, *raPtr, refSize, refFlavor);
      iis->memrefStatus = memrefStatus;
      switch (memrefStatus) {
      case SUCCESS:
	iis->memopRequested = 1;
	iis->memopReadyToComplete = 1;
	break;
      case STALL:
	iis->memopRequested = 1;
	epe->stalledMemop = inum;
	break;
      case FAILURE:
	iis->memopRequested = 0; /* pretend we didn't request it */
	break;
      case BUSERROR:
	NOTREACHED();
	break;
      default:
	NOTREACHED();
      }
    } else {
      ASSERT(iis->dmmuStatus == MMU_UNCACHED);
      /* handler this at memop completion time for now */
      /* @@@@ should check the trapPoint and issue the request when we're past the trap point */
      iis->memopRequested = 1;
    }
  } break;

#ifndef SOLO
  case PRIV_OP_HW_LD: 
  case PRIV_OP_HW_ST: 
    /* handle these at memop completion time */
    /* @@@@ should check the trapPoint and issue the request when we're past the trap point */
    iis->memopRequested = 1;
    break;
#endif

  default:
    ASSERT (0);
  } /* switch(opcode) */
}

#ifdef EPSILON_USE_MEMREF
/* only destination register may be used here, 
 * because other physical registers may have been freed by EpsilonCommitInstruction 
 */
void EpsilonCompleteMemop(int cpu, Inum inum)
{
  AlphaState *pe = curPE;
  EpsilonState *epe = EPE[cpu];
  Instid instid = inumInstid(inum);
  InflightInstructionState *iis = &epe->iqueue[instid];
  union alpha_instruction instr = iis->instr;
  Reg *rdestp = &epe->regs[iis->prdest];
  VA vAddr = iis->eaddr;
  PA pAddr = iis->paddr;
  MA mAddr = iis->maddr;
  int opcode = iis->instr.common.opcode;
  RefSize refSize = alphaRefSizeMap[opcode];
  RefFlavor refFlavor = alphaRefFlavorMap[opcode];

  if (epsilonParams.debugScheduler)
    CPUWarning("EpsilonCompleteMemop: cpu=%d pc=%lx inum=%ld instr=%x opcode=0x%x va=%lx pa=%lx ma=%lx\n",
	       cpu, iis->PC, iis->inum, iis->instr.word, opcode,
	       vAddr, pAddr, mAddr);

  ASSERT(iis->issued);
  pe->PC = iis->PC;
  pe->instr = iis->instr;
  pe->trapState = iis->trapState;

  pe->instr = instr;
  gdbInstr = instr.word;
  gdbInstrCommon = instr.common.opcode;
    
  switch (opcode) {
  case op_ldq_u:
  case op_ldf:
  case op_ldg:
  case op_lds:
  case op_ldt:
  case op_ldl:
  case op_ldq:
  case op_ldl_l:
  case op_ldq_l: {
    Reg mdata = iis->value;       /* result should be in iis->value for cached reads */
    Reg *raPtr = &epe->regs[iis->prdest];
    if (iis->dmmuStatus == MMU_UNCACHED) {
       iis->untrappable = 1;
       iis->memrefStatus = GammaMemoryReadOpcode(opcode,&mdata,iis->eaddr);
       *raPtr = mdata;
       ASSERT(iis->memrefStatus == MMU_SUCCESS);
    } else {
       if (iis->memrefStatus != SUCCESS)
          CPUWarning("Assertion about to fail: expected SUCCESS, inum=%ld memrefStatus=%d\n",
                     inum, iis->memrefStatus);
       ASSERT(iis->memrefStatus == SUCCESS);

      /* cached read completion */
      /* pick up any bytes that haven't made it to memory yet */
      /* @@@@ fix the storebuffer -- works with inorder commits */
      /* @@@@ also make sure loads are pushed out beyond store-conditional commits */
#ifdef USE_STOREBUFFER
      if (iis->lastStoreInum > epe->lastInumCommitted && !epe->iqueue[inumInstid(iis->lastStoreInum)].memopCompleted) {
	InflightInstructionState *lsiis = &epe->iqueue[inumInstid(iis->lastStoreInum)];
	ASSERT(lsiis->inum == iis->lastStoreInum);
	CPUWarning("EpsilonCompleteMemop: OutOfOrder -- cpu=%d pc=%lx inum=%d instr=%x opcode=0x%x lastStoreInum=%ld \n",
		   cpu, iis->PC, iis->inum, iis->instr.word, iis->instr.common.opcode, iis->lastStoreInum);
      }
      if (0) /* @@@@ turned off for now */
	*raPtr = EpsilonStoreBufferRead(pe, epe, iis, mdata);
      else
	*raPtr = mdata;
#endif
      CHECK_LD_ANN(vAddr,pAddr);
      STATS_INC(cpu, dReads, 1);
      TRACE_ENTRY(TRACE_LOAD,curPE->myNum,vAddr);

      if (0) gammaMemoryTable[opcode](raPtr, NULL,(Reg*)mAddr, pAddr, vAddr);
      if (refSize == WORD_SZ) {
         if (opcode == op_lds || op_ldf) {
            /* these opcodes munge the data */
            gammaMemoryTable[opcode](&mdata, NULL,(Reg*)&mdata, pAddr, vAddr);
         } else {
            /* word loads sign extend the data */
            if (mdata & 0x80000000L)
               mdata |= 0xFFFFFFFF00000000L;
         }
      }
      if (refFlavor == LL_FLAVOR) {
         PA align = pAddr & ~(VA)(CACHE_LINE_SIZE-1);
         curPE->lock_flag = 1;
         curPE->locked_addr = align;
         numLLActive++;
      }

      *raPtr = iis->value;

      if (gammaParams.debugLoadStore || vAddr == alphaDebugParams.debugVATranslation)
	CPUWarning("EpsilonCompleteMemopRd: pc=%lx val=%016lx vaddr=%lx paddr=%lx status=%d cycle=%ld\n",
		   pe->PC, *raPtr, vAddr, pAddr, iis->memrefStatus, gammaCurrentTime[cpu]);
    }
    EpsilonPhysicalRegisterSetValid(epe, iis->prdest);
  } break;

  case op_stq_u:
  case op_stf:
  case op_stg:
  case op_sts:
  case op_stt:
  case op_stl:
  case op_stq:
  case op_stl_c:
  case op_stq_c: {
    /* @@@@ fix the storebuffer -- works with inorder commits */
    if (iis->lastStoreInum > epe->lastInumCommitted && !epe->iqueue[inumInstid(iis->lastStoreInum)].memopCompleted) {
      InflightInstructionState *lsiis = &epe->iqueue[inumInstid(iis->lastStoreInum)];
      ASSERT(lsiis->inum == iis->lastStoreInum);
      CPUWarning("EpsilonCompleteMemop: OutOfOrder -- cpu=%d pc=%lx inum=%d instr=%x opcode=0x%x lastStoreInum=%ld \n",
		 cpu, iis->PC, iis->inum, iis->instr.word, iis->instr.common.opcode, iis->lastStoreInum);
    }
    if (iis->dmmuStatus == MMU_UNCACHED) {
      /* uncached, handle it now */
       iis->untrappable = 1;
      iis->memrefStatus = GammaMemoryWriteOpcode(opcode, &epe->regs[iis->prdest], &iis->value,iis->eaddr);
      ASSERT(iis->memrefStatus == MMU_SUCCESS);
    } else if (iis->memrefStatus == SUCCESS) {
       iis->untrappable = 1;
       gammaMemoryTable[opcode](rdestp, (Reg*)&iis->value,(Reg*)mAddr,pAddr,vAddr);
    } else {
       NOTREACHED();
    }
  } break;

#ifndef SOLO
  case PRIV_OP_HW_LD: 
    {
      PA pAddr = 0;
      int size=0;
      MA mA;
      Reg *regPtr = &epe->regs[iis->prdest];
      ev5_hwldst_format hwldst;
      int rb = hwldst.hwmem_format.rb;
      Reg rbv = iis->eaddr; /* in this case, we have left the displacement off of iis->eaddr */
      int mmuStatus;
      hwldst.word = instr.word;      
      mmuStatus = EV5_HW_Load(pe,instr.word, rbv,&pAddr,&size);
      mA = PHYS_TO_MEMADDR(0, pAddr);
      ASSERT( size==4 || size == 8);
      iis->dmmuStatus = mmuStatus;
      switch (mmuStatus) { 
      case MMU_SUCCESS:
	ASSERT (pAddr < MEM_SIZE(0));
	/* @@@@ maybe I should remove this later, but it seems handy */
	if        (size==4) { *regPtr = *(int32*)mA;  
	} else if (size==8) { *regPtr = *(int64*)mA;
	} 
	CHECK_LD_ANN(rbv+hwldst.hwmem_format.disp, pAddr);
	break;
      case MMU_LL:
        NOTREACHED();
        break;
      case MMU_UNCACHED:
         iis->untrappable = 1;
	if (size==4) {  
	  mmuStatus = GammaUncachedOperation((pAddr&PA_MASK)|K1BASE,BDOOR_LOAD_WORD,regPtr);
	} else {
	  mmuStatus = GammaUncachedOperation((pAddr&PA_MASK)|K1BASE,BDOOR_LOAD_DOUBLE,regPtr);
	}
	iis->dmmuStatus = mmuStatus;
	if (mmuStatus==MMU_SUCCESS) { 
	  /* OK */
	} else {
	  ASSERT (mmuStatus==MMU_EXCEPTION);
	}
	break;
      case MMU_EXCEPTION:
	break;
      default: ASSERT(0);
      }
      EpsilonPhysicalRegisterSetValid(epe, iis->prdest);
      if (gammaParams.debugLoadStore || (rbv + hwldst.hwmem_format.disp == alphaDebugParams.debugVATranslation)) {
	CPUWarning("PRIV_OP_HW_LD: pc=%lx val=%lx r%d=%lx vaddr=%lx paddr=%lx maddr=%lx status=%d cycle=%ld\n",
		   iis->nextpc, *regPtr, rb, rbv, rbv + hwldst.hwmem_format.disp, pAddr, mA,
		   mmuStatus, gammaCurrentTime[cpu]);
      }
      break;
    }
  case PRIV_OP_HW_ST: 
    {
      PA pAddr=0;
      int size=0;
      Reg *regPtr=&iis->value;
      MA mA;
      ev5_hwldst_format hwldst;
      Reg rbv = iis->eaddr;
      int mmuStatus;

      hwldst.word = instr.word;      
      mmuStatus = EV5_HW_Store(pe,instr.word, rbv, &pAddr,&size);
      mA = PHYS_TO_MEMADDR(0,pAddr);
      ASSERT( size==4 || size == 8);
      iis->dmmuStatus = mmuStatus;
      if (gammaParams.debugLoadStore || (rbv + hwldst.hwmem_format.disp == alphaDebugParams.debugVATranslation))
	CPUWarning("PRIV_OP_HW_ST: pc=%lx val=%lx vaddr=%lx paddr=%lx maddr=%lx status=%d cycle=%ld\n",
		   iis->nextpc, *regPtr, rbv + hwldst.hwmem_format.disp, pAddr, mA, mmuStatus, gammaCurrentTime[cpu]);
      switch (mmuStatus) { 
      case MMU_SUCCESS:
         iis->untrappable = 1;
	if (pAddr>MEM_SIZE(0)) { 
	  CPUError("hw_st: pAddr=0x%lx memory ref too large \n",pAddr);
	}
	if        (size==4) { *(int32*)mA = *regPtr;
	} else if (size==8) { *(int64*)mA = *regPtr;
	} 
	 /* @@@@ maybe I should remove this later, but it seems handy */
	 CHECK_ST_ANN(rbv+hwldst.hwmem_format.disp, pAddr);
	break;
      case MMU_SC:
        NOTREACHED();
        break;
      case MMU_UNCACHED:
         iis->untrappable = 1;
	if (size==4) {  
	  mmuStatus = GammaUncachedOperation((pAddr&PA_MASK)|K1BASE,BDOOR_STORE_WORD,regPtr);
	} else {
	  mmuStatus = GammaUncachedOperation((pAddr&PA_MASK)|K1BASE,BDOOR_STORE_DOUBLE,regPtr);
	}
	iis->dmmuStatus = mmuStatus;
	if (mmuStatus==MMU_SUCCESS) { 
	  /* OK */
	} else {
	  ASSERT (mmuStatus==MMU_EXCEPTION);
	}
	break;
      case MMU_EXCEPTION:
	break;
      default: ASSERT(0);
      }
      break;
    }
#endif

  default:
    ASSERT (0);
  } /* switch(opcode) */

  /* some book-keeping */
  ASSERT(iis->nextpc != -1);
  iis->memopCompleted = 1;
  ASSERT(inum > epe->lastInumDone);
  EpsilonPhysicalRegisterSetValid(epe, iis->prdest);

  if (iis->committed && (!iis->isMemop || iis->memopCompleted)) {
    iis->done = 1;
    EpsilonCleanupInstruction(cpu, inum);
  }
}

#else /* not EPSILON_USE_MEMREF */

/* only destination register may be used here, 
 * because other physical registers may have been freed by EpsilonCommitInstruction 
 */
void EpsilonCompleteMemop(int cpu, Inum inum)
{
  AlphaState *pe = curPE;
  EpsilonState *epe = EPE[cpu];
  Instid instid = inumInstid(inum);
  InflightInstructionState *iis = &epe->iqueue[instid];
  union alpha_instruction instr = iis->instr;
  Reg *raPtr;
  int opcode;

  if (epsilonParams.debugScheduler)
    CPUWarning("EpsilonCompleteMemop: cpu=%d pc=%lx inum=%ld instr=%x opcode=0x%x ea=%lx pa=%lx ma=%lx\n",
	       cpu, iis->PC, iis->inum, iis->instr.word, iis->instr.common.opcode,
	       iis->eaddr, iis->paddr, iis->maddr);

  ASSERT(iis->issued);
  pe->PC = iis->PC;
  pe->instr = iis->instr;
  pe->trapState = iis->trapState;

  pe->instr = instr;
  gdbInstr = instr.word;
  gdbInstrCommon = instr.common.opcode;
    
  opcode = instr.common.opcode;
  switch (opcode) {
  case op_ldq_u:
  case op_ldf:
  case op_ldg:
  case op_lds:
  case op_ldt:
  case op_ldl:
  case op_ldq:
  case op_ldl_l:
  case op_ldq_l: {
    Reg mdata;
    raPtr = &epe->regs[iis->prdest];
    if (iis->dmmuStatus == MMU_UNCACHED)
       iis->untrappable = 1;
    iis->dmmuStatus = GammaMemoryReadOpcode(opcode,&mdata,iis->eaddr);
    ASSERT(iis->dmmuStatus == MMU_SUCCESS);
    /* pick up any bytes that haven't made it to memory yet */
    /* @@@@ fix the storebuffer -- works with inorder commits */
    /* @@@@ also make sure loads are pushed out beyond store-conditional commits */
    if (iis->lastStoreInum > epe->lastInumCommitted && !epe->iqueue[inumInstid(iis->lastStoreInum)].memopCompleted) {
      InflightInstructionState *lsiis = &epe->iqueue[inumInstid(iis->lastStoreInum)];
      ASSERT(lsiis->inum == iis->lastStoreInum);
      CPUWarning("EpsilonCompleteMemop: OutOfOrder -- cpu=%d pc=%lx inum=%d instr=%x opcode=0x%x lastStoreInum=%ld \n",
		 cpu, iis->PC, iis->inum, iis->instr.word, iis->instr.common.opcode, iis->lastStoreInum);
    }
    if (0) /* @@@@ turned off for now */
      *raPtr = EpsilonStoreBufferRead(pe, epe, iis, mdata);
    else
      *raPtr = mdata;
    EpsilonPhysicalRegisterSetValid(epe, iis->prdest);
  } break;

  case op_stq_u:
  case op_stf:
  case op_stg:
  case op_sts:
  case op_stt:
  case op_stl:
  case op_stq:
  case op_stl_c:
  case op_stq_c: {
    /* @@@@ fix the storebuffer -- works with inorder commits */
    if (iis->lastStoreInum > epe->lastInumCommitted && !epe->iqueue[inumInstid(iis->lastStoreInum)].memopCompleted) {
      InflightInstructionState *lsiis = &epe->iqueue[inumInstid(iis->lastStoreInum)];
      ASSERT(lsiis->inum == iis->lastStoreInum);
      CPUWarning("EpsilonCompleteMemop: OutOfOrder -- cpu=%d pc=%lx inum=%d instr=%x opcode=0x%x lastStoreInum=%ld \n",
		 cpu, iis->PC, iis->inum, iis->instr.word, iis->instr.common.opcode, iis->lastStoreInum);
    }
    iis->untrappable = 1;
    iis->dmmuStatus = GammaMemoryWriteOpcode(opcode, &epe->regs[iis->prdest], &iis->value,iis->eaddr);
    ASSERT(iis->dmmuStatus == MMU_SUCCESS);
  } break;

#ifndef SOLO
  case PRIV_OP_HW_LD: 
    {
      PA pAddr = 0;
      int size=0;
      MA mA;
      Reg *regPtr = &epe->regs[iis->prdest];
      ev5_hwldst_format hwldst;
      int rb = hwldst.hwmem_format.rb;
      Reg rbv = iis->eaddr; /* in this case, we have left the displacement off of iis->eaddr */
      int mmuStatus;
      
      hwldst.word = instr.word;      
      iis->untrappable = 1;
      mmuStatus = EV5_HW_Load(pe,instr.word, rbv,&pAddr,&size);
      mA = PHYS_TO_MEMADDR(0, pAddr);
      ASSERT( size==4 || size == 8);
      iis->dmmuStatus = mmuStatus;
      switch (mmuStatus) { 
      case MMU_SUCCESS:
	ASSERT (pAddr < MEM_SIZE(0));
	/* @@@@ maybe I should remove this later, but it seems handy */
	if        (size==4) { *regPtr = *(int32*)mA;  
	} else if (size==8) { *regPtr = *(int64*)mA;
	} 
	CHECK_LD_ANN(rbv+hwldst.hwmem_format.disp, pAddr);
	break;
      case MMU_LL:
        NOTREACHED();
        break;
      case MMU_UNCACHED:
	if (size==4) {  
	  mmuStatus = GammaUncachedOperation((pAddr&PA_MASK)|K1BASE,BDOOR_LOAD_WORD,regPtr);
	} else {
	  mmuStatus = GammaUncachedOperation((pAddr&PA_MASK)|K1BASE,BDOOR_LOAD_DOUBLE,regPtr);
	}
	iis->dmmuStatus = mmuStatus;
	if (mmuStatus==MMU_SUCCESS) { 
	  /* OK */
	} else {
	  ASSERT (mmuStatus==MMU_EXCEPTION);
	}
	break;
      case MMU_EXCEPTION:
	break;
      default: ASSERT(0);
      }
      EpsilonPhysicalRegisterSetValid(epe, iis->prdest);
      if (gammaParams.debugLoadStore || (rbv + hwldst.hwmem_format.disp == alphaDebugParams.debugVATranslation)) {
	CPUWarning("PRIV_OP_HW_LD: pc=%lx val=%lx r%d=%lx vaddr=%lx paddr=%lx maddr=%lx status=%d cycle=%ld\n",
		   iis->nextpc, *regPtr, rb, rbv, rbv + hwldst.hwmem_format.disp, pAddr, mA,
		   mmuStatus, gammaCurrentTime[cpu]);
      }
      break;
    }
  case PRIV_OP_HW_ST: 
    {
      PA pAddr=0;
      int size=0;
      Reg *regPtr=&iis->value;
      MA mA;
      ev5_hwldst_format hwldst;
      Reg rbv = iis->eaddr;
      int mmuStatus;

      hwldst.word = instr.word;      
      iis->untrappable = 1;
      mmuStatus = EV5_HW_Store(pe,instr.word, rbv, &pAddr,&size);
      mA = PHYS_TO_MEMADDR(0,pAddr);
      ASSERT( size==4 || size == 8);
      iis->dmmuStatus = mmuStatus;
      if (gammaParams.debugLoadStore || (rbv + hwldst.hwmem_format.disp == alphaDebugParams.debugVATranslation))
	CPUWarning("PRIV_OP_HW_ST: pc=%lx val=%lx vaddr=%lx paddr=%lx maddr=%lx status=%d cycle=%ld\n",
		   iis->nextpc, *regPtr, rbv + hwldst.hwmem_format.disp, pAddr, mA, mmuStatus, gammaCurrentTime[cpu]);
      switch (mmuStatus) { 
      case MMU_SUCCESS:
	if (pAddr>MEM_SIZE(0)) { 
	  CPUError("hw_st: pAddr=0x%lx memory ref too large \n",pAddr);
	}
	if        (size==4) { *(int32*)mA = *regPtr;
	} else if (size==8) { *(int64*)mA = *regPtr;
	} 
	 /* @@@@ maybe I should remove this later, but it seems handy */
	 CHECK_ST_ANN(rbv+hwldst.hwmem_format.disp, pAddr);
	break;
      case MMU_SC:
        NOTREACHED();
        break;
      case MMU_UNCACHED:
	if (size==4) {  
	  mmuStatus = GammaUncachedOperation((pAddr&PA_MASK)|K1BASE,BDOOR_STORE_WORD,regPtr);
	} else {
	  mmuStatus = GammaUncachedOperation((pAddr&PA_MASK)|K1BASE,BDOOR_STORE_DOUBLE,regPtr);
	}
	iis->dmmuStatus = mmuStatus;
	if (mmuStatus==MMU_SUCCESS) { 
	  /* OK */
	} else {
	  ASSERT (mmuStatus==MMU_EXCEPTION);
	}
	break;
      case MMU_EXCEPTION:
	break;
      default: ASSERT(0);
      }
      break;
    }
  case PRIV_OP_REI:
    if (epsilonParams.debugSchedulerDuringInterrupt) {
      savedDebugScheduler--;
      epsilonParams.debugScheduler = savedDebugScheduler;
    }
    break;
#endif

  default:
    ASSERT (0);
  } /* switch(opcode) */

  /* some book-keeping */
  ASSERT(iis->nextpc != -1);
  iis->memopCompleted = 1;
  EpsilonPhysicalRegisterSetValid(epe, iis->prdest);

  if (iis->committed && (!iis->isMemop || iis->memopCompleted)) {
    iis->done = 1;
    EpsilonCleanupInstruction(cpu, inum);
  }
}
#endif /* EPSILON_USE_MEMREF */

void EpsilonCleanupInstruction(int cpu, Inum inum)
{
  Instid instid = inumInstid(inum);
  EpsilonState *epe = EPE[cpu];
  InflightInstructionState *iis = &epe->iqueue[instid];
  ASSERT(iis->done);
  if (!iis->trapped && epe->lastInumDone < inum)
     epe->lastInumDone = inum;  
  iis->inflight = 0;
}

void EpsilonCommitInstruction(int cpu, Inum inum)
{
  AlphaState *pe = PE[cpu];
  EpsilonState *epe = EPE[cpu];
  Instid instid = inumInstid(inum);
  InflightInstructionState *iis = &epe->iqueue[instid];
  regname_t lrdest = iis->lrdest;
  pe->PC = iis->PC;
  pe->instr = iis->instr;
  pe->trapState = iis->trapState;

  if (epsilonParams.debugScheduler)
    CPUWarning("EpsilonCommit: cpu=%d pc=%lx inum=%ld instr=%x opcode=0x%x prdest=%d oldprdest=%d\n",
	       cpu, iis->PC, iis->inum, iis->instr.word, iis->instr.common.opcode, 
	       iis->prdest, iis->oldprdest);

  ASSERT(iis->fetched);
  ASSERT(iis->issued);
  ASSERT(!iis->trapped);
  STATS_INC(pe->myNum, numInstructions, 1);

  PRINT_INSTRUCTION(pe,iis->instr.word);

  /* some instructions have a commit-time action */
  switch (iis->instr.common.opcode) {
  case PRIV_OP_MFPR:
  case PRIV_OP_MTPR:
    {
      int priv = iis->instr.m_format.memory_displacement;
      regname_t pra = iis->pra;
      if (iis->instr.common.opcode == PRIV_OP_MFPR) {
	epe->regs[iis->prdest] = EV5_MoveFromPriv(pe,priv);
	EpsilonPhysicalRegisterSetValid(epe, iis->prdest);
      } else {
	/* happens at commit time, rather than having shadow state */
	EV5_MoveToPriv(pe,priv,epe->regs[pra]);
      }
      iis->nextpc = iis->PC + 4;
      break;
    }
  case PRIV_OP_REI:
    if (0 && epsilonParams.debugScheduler) {
      epsilonParams.debugScheduler = 0; /* @@@@ remove this */
      CPUWarning("@@@@ committing PRIV_OP_REI and reseting epsilonParams.debugScheduler to zero\n");
    }
  }

  if (lrdest >= 0 && lrdest != ZERO_REGISTER && lrdest < N_LOGICAL_REGISTERS) {
    /* updated committed regmap */
    epe->committedRegmap[lrdest] = iis->prdest;
    /* release the old physical register */
    epsilonPhysregFree(epe, iis->oldprdest);
  }

  iis->committed = 1;
  iis->trapfree = 1;
  epe->lastInumCommitted = iis->inum;
  if (epe->oldestTrapState != iis->trapState) {
    ASSERT(epsilonParams.newTrapStates);
    freeTrapState(epe->oldestTrapState);
    epe->oldestTrapState = iis->trapState;
  }

  pe->PC = iis->nextpc;
  AlphaCommitAnnExec();
  RunPCAnnotations(iis->PC, iis->nextpc);
  /* an annotation might modify the PC */
  iis->nextpc = pe->PC;
  if (iis->committed && (!iis->isMemop || iis->memopCompleted)) {
    iis->done = 1;
    EpsilonCleanupInstruction(cpu, inum);
  }
}


/* only valid on most recently fetched instruction */
void EpsilonTrapInstruction(int cpu, Inum inum)
{
#if 0
  AlphaState *pe = PE[cpu];
#endif
  EpsilonState *epe = EPE[cpu];
  Instid instid = inumInstid(inum);
  InflightInstructionState *iis = &epe->iqueue[instid];
  InflightInstructionState *nextiis = &epe->iqueue[inumInstid(iis->inorderNext)];
  regname_t lrdest = iis->lrdest;

  if (epsilonParams.debugScheduler)
    CPUWarning("EpsilonTrap: cpu=%d pc=%lx inum=%ld instr=%x opcode=%x prdest=%d oldprdest=%d\n",
	       cpu, iis->PC, iis->inum, iis->instr.word, iis->instr.common.opcode, 
	       iis->prdest, iis->oldprdest);

  ASSERT(!iis->trapped);
  ASSERT(!iis->untrappable);
  ASSERT(!iis->committed);
  ASSERT(iis->inorderNext > epe->lastInumFetched || !nextiis->inflight);
  epe->lastInumUntrapped = iis->inorderPrev;
  iis->trapped = 1;
  iis->trapfree = 1; /* can no longer cause a trap */

  if (epe->stallFetchUntilInum >= iis->inum)
    epe->stallFetchUntilInum = iis->inorderPrev;

  /* unwind the trapState */
  {
    AlphaTrapState *newestTrapState;
    if (iis->inorderPrev > epe->lastInumCommitted) {
      newestTrapState = epe->iqueue[inumInstid(iis->inorderPrev)].trapState;
    } else {
      newestTrapState = epe->oldestTrapState;
    }
    ASSERT(newestTrapState);
    if (newestTrapState != iis->trapState) {
      ASSERT(epsilonParams.newTrapStates);
      freeTrapState(iis->trapState);
      epe->newestTrapState = newestTrapState;
    }
  }

  EpsilonPhysicalRegisterClearValid(epe, iis->prdest);

  if (lrdest >= 0 && lrdest < N_LOGICAL_REGISTERS
      && iis->prdest != ZERO_REGISTER_DEST_PHYSREG) {
    /* unwind the register map */
    ASSERT(EpsilonPhysicalRegisterIsAllocated(epe, iis->oldprdest));
    epe->regmap[iis->lrdest] = iis->oldprdest;
    epsilonPhysregFree(epe, iis->prdest);
  }

  /* unwind storebuffer */
  if (epe->lastStoreInum == inum)
    epe->lastStoreInum = iis->lastStoreInum;

  iis->done = 1;
  EpsilonCleanupInstruction(cpu, inum);
}

/* ******************************************************* 
 * EpsilonParameters
 * *******************************************************/

int PrintAllInflightInstructionsCmd(void *cdata, Tcl_Interp *interp, int argc, char *argv)
{
   if (EPE != NULL && EPE[0] != NULL)
      PrintAllInflightInstructions(EPE[0]);
  return TCL_OK;
}

void EpsilonTclInit(Tcl_Interp *interp)
{
  ParamRegister("PARAM(CPU.Epsilon.Width)", (char *)&epsilonParams.width, PARAM_INT);
  ParamRegister("PARAM(CPU.Epsilon.Pipelined)", (char *)&epsilonParams.pipelined, PARAM_INT);
  epsilonParams.fetchAheadLimit = 128;
  ParamRegister("PARAM(CPU.Epsilon.FetchAheadLimit)", (char *)&epsilonParams.fetchAheadLimit, PARAM_INT);
  ParamRegister("PARAM(CPU.Epsilon.UseMemRef)", (char *)&epsilonParams.useMemRef, PARAM_INT);
  ParamRegister("PARAM(CPU.Epsilon.InterruptFrequencyMask)", (char *)&epsilonParams.interruptFrequencyMask, PARAM_INT);
  ParamRegister("PARAM(CPU.Epsilon.NewTrapStates)", (char *)&epsilonParams.newTrapStates, PARAM_INT);
  ParamRegister("PARAM(CPU.Epsilon.Interrupt.AtRei)", (char *)&epsilonParams.takeInterruptAtRei, PARAM_INT);


  ParamRegister("PARAM(CPU.Epsilon.Debug.Scheduler)", (char *)&epsilonParams.debugScheduler, PARAM_INT);
  ParamRegister("PARAM(CPU.Epsilon.Debug.SchedulerDuringInterrupt)", (char *)&epsilonParams.debugSchedulerDuringInterrupt, PARAM_INT);
  ParamRegister("PARAM(CPU.Epsilon.Debug.TrapMask)", (char *)&epsilonParams.debugTrapMask, PARAM_INT);
  ParamRegister("PARAM(CPU.Epsilon.Debug.Regmap)", (char *)&epsilonParams.checkRegmap, PARAM_INT);
  ParamRegister("PARAM(CPU.Epsilon.Trap.ClearRegisterLocks)",
		(char *)&epsilonParams.trapClearRegisterLocks, PARAM_INT);
  ParamRegister("PARAM(CPU.Epsilon.Trap.TrapNotifyInterrupts)",
		(char *)&epsilonParams.trapNotifyInterrupts, PARAM_INT);
  ParamRegister("PARAM(CPU.Epsilon.Interrupt.TrapUncommitted)",
		(char *)&epsilonParams.interruptTrapUncommitted, PARAM_INT);
  Tcl_TraceVar(TCLInterp, "PHYSREGS",
                TCL_TRACE_READS | TCL_TRACE_WRITES | TCL_GLOBAL_ONLY, 
                EpsilonPhysRegsAccess, NULL); 
  Tcl_CreateCommand(interp, "PrintAllInflightInstructions", (void*)PrintAllInflightInstructionsCmd, NULL, NULL);
}


/*****************************************************************
 * CPUVectorInit
 *****************************************************************/

Result EpsilonGetReg(int cpu, int regnum, Reg *val)
{
  AlphaState *pe = PE[cpu];
  EpsilonState *epe = EPE[cpu];
  Inum inum = epe->lastInumCommitted;
  InflightInstructionState *iis = &epe->iqueue[inumInstid(inum)];

  if (regnum >=0 && regnum < 64) {
    if (EV5_ShadowRegistersUsed(pe, iis->PC)) {
      regnum = LOGICAL_SHADOW_REGISTER(regnum);
    }
    *val = epe->regs[epe->committedRegmap[regnum]];
    return SUCCESS;
  }
  if (regnum >= SHADOW_REGNUM && regnum < (SHADOW_REGNUM+8)) {
    if (EV5_ShadowRegistersUsed(pe, iis->PC)) {
      /* return the saved register value */
      *val = epe->regs[epe->committedRegmap[lrmap[regnum-3+8]]];
    } else {
      /* return the value of the shadow register */
      *val = epe->regs[epe->committedRegmap[regnum-3+8]];
    }
    return SUCCESS;
  }
  {
    Inum inum = epe->lastInumCommitted;
    InflightInstructionState *iis = &epe->iqueue[inumInstid(inum)];
    if (0 && iis->nextpc == 0) 
      CPUWarning("EpsilongGetReg: null nextpc for cpu=%d pc=%lx inum=%ld\n",
		 cpu, iis->PC, inum);

    switch (regnum) {
    case INSTR_REGNUM: *val = pe->instr.word; break;
    case PC_REGNUM: *val = pe->PC; break;
    case FPCR_REGNUM: *val = pe->fpcr; break;
    default: 
      CPUError("EpsilonGetReg: gdb register %d not defined \n",regnum);
    }
  }
  return SUCCESS;
}


Result EpsilonPutReg(int cpu, int regnum, Reg val)
{
  AlphaState *pe = PE[cpu];
  EpsilonState *epe = EPE[cpu];
  Inum inum = epe->lastInumCommitted;
  InflightInstructionState *iis = &epe->iqueue[inumInstid(inum)];

  if (regnum >=0 && regnum < 64) {
    regname_t physreg;
    if (EV5_ShadowRegistersUsed(pe, iis->PC)) {
      regnum = LOGICAL_SHADOW_REGISTER(regnum);
    }
    physreg = epe->committedRegmap[regnum];
    epe->regs[physreg] = val;
    EpsilonTrapNotify(cpu, epe->lastInumCommitted, EpsilonTTRegisterChangedExternally);
    return SUCCESS;
  }
  switch (regnum) {
  case PC_REGNUM: 
    CPUWarning("@@@@ Modifying PC: oldpc=%lx newpc=%lx\n", pe->PC, val);
    pe->PC = val; 
    iis->nextpc = val;
    EpsilonTrapNotify(cpu, epe->lastInumCommitted, EpsilonTTRegisterChangedExternally);
    break;
  default: 
    CPUError("EpsilonPutReg: gdb register %d not defined \n",regnum);
  }
  return SUCCESS;
}

VA
EpsilonCurrentPC(int cpu)
{
  AlphaState *pe = PE[cpu];
#if 0
  EpsilonState *epe = EPE[cpu];
  Inum inum = epe->lastInumCommitted;
  InflightInstructionState *iis = &epe->iqueue[inumInstid(inum)];
  return iis->nextpc;
#else
  return pe->PC;
#endif
}

Inst EpsilonCurrentInstruction(int cpu)
{
#if 0
  EpsilonState *epe = EPE[cpu];
  Inum inum = epe->lastInumCommitted;
  InflightInstructionState *iis = &epe->iqueue[inumInstid(inum)];
  return iis->instr.word;
#else
  AlphaState *pe = PE[cpu];
  return pe->instr.word;
#endif
}

void EpsilonStall(AlphaState *P, CPUStatus state)
{
  NOTREACHED();
}

void EpsilonUnstall(int cpu)
{
  EpsilonState *epe = EPE[cpu];
  Inum inum = epe->stalledMemop;
  if (inum != 0) {
    InflightInstructionState *iis = &epe->iqueue[inumInstid(inum)];
    if (epsilonParams.debugScheduler)
       CPUWarning("EPSILON: unstalling cpu=%d inum=%ld pc=%lx opcode=%x dmmuStatus=%d memrefStatus=%d\n",
                  cpu, inum, iis->PC, iis->instr.common.opcode,
                  iis->dmmuStatus, iis->memrefStatus);
    if (iis->trapped) {
      if (EpsilonInstructionIsStore(cpu, inumInstid(inum))) {
	CPUError("Unstalled store instruction that was trapped cpu=%d inum=%ld pc=%lx at cycle=%ld\n",
		 cpu, inum, iis->PC, gammaCurrentTime[cpu]);
      }
    }
    epe->pendingMemops--;
    iis->memopReadyToComplete = 1;
    iis->memrefStatus = SUCCESS;
    epe->stalledMemop = 0;
  }
}

void EpsilonExitSimulator( CPUType toSim )
{
   notifyExit = 1;
   toCPUType = toSim;
}

void 
EpsilonCPUVectorInit(void)
{
   CPUVec.GetRegister          = EpsilonGetReg;
   CPUVec.PutRegister          = EpsilonPutReg;
   CPUVec.Unstall              = EpsilonUnstall;
#if 1
   CPUVec.useMemRef            = epsilonParams.useMemRef;
#else
   CPUVec.useMemRef            = TRUE;
#endif
   CPUVec.CurrentPC            = EpsilonCurrentPC;
   CPUVec.CurrentInstruction   = EpsilonCurrentInstruction;
   CPUVec.ExitSimulator        = EpsilonExitSimulator;
}
