/*
 * Copyright (C) 1998 by the Board of Trustees
 *    of Leland Stanford Junior University.
 * Copyright (C) 1998 Digital Equipment Corporation
 *
 * This file is part of the SimOS distribution.
 * See LICENSE file for terms of the license.
 *
 */



/*
 * ntosloader.c
 * 
 * This file actually loads the ntosloader stored in its proprietary file format. 
 *
 */

#include <stdio.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#include <filehdr.h>
#include <stddef.h>

#include "simtypes.h"
#include "sim_error.h"
#include "alpha.h"
#include "arch_specifics.h"
#include "alpha_access.h"


#define KSEG32 0xFFFFFFFF80000000L

#define PAGE_SHIFT PAGE_BITS

/*
 * NT defines. !!! LONG is 32-bit (seems like that at least)
 */


/*
 * from NT/public/sdk/inc/ntimage.h
 */

typedef uint16 USHORT;
typedef int32  LONG;
typedef uint32 ULONG;
typedef char   BOOLEAN;
typedef uint16 WCHAR;
typedef char   UCHAR;
typedef char   CHAR;
typedef uint16 SHORT;

/*
 * 32bit pointers???
 */
typedef uint32 PVOID;
typedef uint32 PULONG;
typedef uint32 PUCHAR;
typedef uint32 PUSHORT;



#define UNALIGNED 


#include "ntimage.h"


#define NT_SECTOR_SIZE 512


//
// s_flags values
//

#define STYP_REG      0x00000000
#define STYP_TEXT     0x00000020
#define STYP_INIT     0x80000000
#define STYP_RDATA    0x00000100
#define STYP_DATA     0x00000040
#define STYP_LIT8     0x08000000
#define STYP_LIT4     0x10000000
#define STYP_SDATA    0x00000200
#define STYP_SBSS     0x00000080
#define STYP_BSS      0x00000400
#define STYP_LIB      0x40000000
#define STYP_UCODE    0x00000800
#define S_NRELOC_OVFL 0x20000000

//
// Section numbers for local relocation entries
//

#define R_SN_TEXT   1
#define R_SN_INIT   7
#define R_SN_RDATA  2
#define R_SN_DATA   3
#define R_SN_SDATA  4
#define R_SN_SBSS   5
#define R_SN_BSS    6
#define R_SN_LIT8   8
#define R_SN_LIT4   9
#define R_SN_MAX    10



int Load_NTOSLoader(char *filename)
{
   UCHAR LocalBuffer[NT_SECTOR_SIZE+32];
   
   int i;
   IMAGE_FILE_HEADER *FileHeader    = (IMAGE_FILE_HEADER *)LocalBuffer;
   IMAGE_OPTIONAL_HEADER *OptionalHeader = (IMAGE_OPTIONAL_HEADER *)(LocalBuffer+sizeof(IMAGE_FILE_HEADER));
   IMAGE_SECTION_HEADER  *SectionHeader;
    ULONG ActualBase;
    ULONG SectionBase;
    ULONG SectionOffset;
    ULONG SectionIndex;
    ULONG Index;
    ULONG NumberOfSections;
    ULONG SectionFlags;
    ULONG FwPageCount;
    ULONG FwActualBasePage;

    VA entryPoint;
    VA kernStart=0,kernEnd=0;

    ULONG TopAddress = 0;

   int fd = open(filename,0);

   if (fd<0) return -1;
   
   i = read(fd,LocalBuffer,NT_SECTOR_SIZE);
   ASSERT (i==NT_SECTOR_SIZE);

   if (FileHeader->Machine != IMAGE_FILE_MACHINE_ALPHA) { 
      CPUError("Load_NTOSLoader:: fileH->Machine = %x != %x \n",
               FileHeader->Machine, IMAGE_FILE_MACHINE_ALPHA);
   }

   SectionHeader =
      (PIMAGE_SECTION_HEADER)(LocalBuffer + sizeof(IMAGE_FILE_HEADER) +
                              FileHeader->SizeOfOptionalHeader);
   
   NumberOfSections = FileHeader->NumberOfSections;
   CPUWarning("Number of sections %d offset %lx\n",
              NumberOfSections,
              sizeof(IMAGE_FILE_HEADER) +
              FileHeader->SizeOfOptionalHeader);
   
   if ((FileHeader->Characteristics & IMAGE_FILE_RELOCS_STRIPPED) != 0) {
      ActualBase = OptionalHeader->BaseOfCode;
      FwPageCount = (OptionalHeader->BaseOfData + OptionalHeader->SizeOfInitializedData) -
         ActualBase;
      CPUWarning("Characteristics 1 base=0x%lx pagecount=%d \n",ActualBase,FwPageCount);
                 
   } else {
        FwPageCount = OptionalHeader->SizeOfCode +
           OptionalHeader->SizeOfInitializedData +
                      OptionalHeader->SizeOfUninitializedData;
        
        ActualBase = (TopAddress - FwPageCount) & ~(PAGE_SIZE - 1);
      CPUWarning("Characteristics 0 base=0x%lx pagecount=%d \n",ActualBase,FwPageCount);
   }

   FwActualBasePage = (ActualBase & 0x1fffffff) >> PAGE_SHIFT;
   CPUWarning("Actual base 0x%lx pageBase 0x%lx \n",ActualBase,FwActualBasePage);

   FwPageCount = (FwPageCount + PAGE_SIZE - 1) >> PAGE_SHIFT;
   
   entryPoint = ((ActualBase | KSEG32) +
                 (OptionalHeader->AddressOfEntryPoint - OptionalHeader->BaseOfCode)
                 );
   CPUWarning("entryPoint 0x%lx \n",entryPoint);
   SectionOffset = 0;
   for (Index = 0; Index < NumberOfSections; Index += 1) {
      

        //
        // Compute the destination address for the current section.
        //
        if ((FileHeader->Characteristics & IMAGE_FILE_RELOCS_STRIPPED) != 0) {
            SectionBase = SectionHeader->VirtualAddress | KSEG32;
        } else {
           SectionBase = ActualBase + SectionOffset;
           /*
            * //
            * //  Store the section relocation information in the table.
            * //
            */
           SectionFlags = SectionHeader->Characteristics;
           if  (SectionFlags & STYP_TEXT) {
              SectionIndex = R_SN_TEXT;
           } else if (SectionFlags & STYP_INIT) {
              SectionIndex = R_SN_INIT;
           } else if (SectionFlags & STYP_RDATA) {
              SectionIndex = R_SN_RDATA;
           } else if (SectionFlags & STYP_DATA) {
              SectionIndex = R_SN_DATA;
           } else if (SectionFlags & STYP_SDATA) {
              SectionIndex = R_SN_SDATA;
           } else if (SectionFlags & STYP_SBSS) {
              SectionIndex = R_SN_SBSS;
           } else if (SectionFlags & STYP_BSS) {
              SectionIndex = R_SN_BSS;
           } else {
              CPUError("FW_UNKNOWN_SECTION_TYPE_MSG \n");
           }
#if 0
           RelocationTable[SectionIndex].PointerToRelocations = SectionHeader->PointerToRelocations;
           RelocationTable[SectionIndex].NumberOfRelocations = SectionHeader->NumberOfRelocations;
           RelocationTable[SectionIndex].FixupValue = SectionBase - SectionHeader->VirtualAddress;
#endif
        }
        if ((SectionHeader->Characteristics &
             (STYP_TEXT | STYP_INIT | STYP_RDATA | STYP_DATA | STYP_SDATA)) != 0) {
           CPUWarning("Section  type 0x%4lx (%10s) , seeking to %10ld, reading into 0x%lx size=0x%lx\n",
                      (uint64)SectionHeader->Characteristics,
                      SectionHeader->Name,
                      SectionHeader->PointerToRawData,
                      SectionBase,
                      SectionHeader->SizeOfRawData);
           
	   if (!kernStart || (kernStart >SectionBase)) {
	      kernStart = SectionBase; 
	   }
	   if (!kernEnd || (kernEnd < SectionBase+SectionHeader->SizeOfRawData)) {
		kernEnd = SectionBase+SectionHeader->SizeOfRawData;
	   }

           i = lseek(fd, SectionHeader->PointerToRawData,SEEK_SET);
           i = read(fd,PHYS_TO_MEMADDR(0,SectionBase & BITMASK(31)),
		    SectionHeader->SizeOfRawData);
           if (i!=SectionHeader->SizeOfRawData){ 
              CPUWarning("i=0x%x != SectionHeader->SizeOfRawData=0x%x \n",
                       i,SectionHeader->SizeOfRawData);
              perror("NTOSloader");
           }
	   
           SectionOffset += SectionHeader->SizeOfRawData;

        } else if ((SectionHeader->Characteristics & (STYP_BSS | STYP_SBSS)) != 0) {
           CPUWarning("Section .bss base=0x%lx size=0x%lx \n",SectionBase,SectionHeader->SizeOfRawData);
	   if (!kernStart || (kernStart >SectionBase)) {
	      kernStart = SectionBase; 
	   }
	   if (!kernEnd || (kernEnd < SectionBase+SectionHeader->SizeOfRawData)) {
	      kernEnd = SectionBase+SectionHeader->SizeOfRawData;
	   }

#if 0
           RtlZeroMemory((PVOID)(SectionBase), SectionHeader->SizeOfRawData);
#endif
           SectionOffset += SectionHeader->SizeOfRawData;
           
        }
        
        SectionHeader += 1;
   }
   close(fd);
   
   alphaAccess.kernStart = KSEG32 | (kernStart & BITMASK(31));
   alphaAccess.kernEnd = KSEG32 | (kernEnd & BITMASK(31));
   alphaAccess.entryPoint = KSEG32 | (entryPoint&BITMASK(31));
   CPUWarning("Loading NT OSLoader start=0x%lx end=0x%lx entry=0x%lx\n",
	      alphaAccess.kernStart,alphaAccess.kernEnd,alphaAccess.entryPoint);
   

   return 1;
}

