/*
 * Copyright (C) 1998 by the Board of Trustees
 *    of Leland Stanford Junior University.
 * Copyright (C) 1998 Digital Equipment Corporation
 *
 * This file is part of the SimOS distribution.
 * See LICENSE file for terms of the license.
 *
 */



#include "simtypes.h"
#include "sim_error.h"
#include "machine_params.h"
#include "machine_defs.h"
#include "registry.h"
#include "console.h"
#include "hd.h"
#include "alpha_access.h"
#include "cpu_interface.h"
#include "gamma.h"
#include "ev5.h"
#include "tlaser.h"
#include "tlaser_clock.h"
#include "tlaser_uart.h"

#include <stddef.h>
#include <stdlib.h>

extern char *bdoorType[];


#define IO_SHIFT 5
#define WIDTH_SHIFT ( IO_SHIFT -2)

/*
 * Common PCI offsets
 */
#define PCI_VENDOR_ID    	0x00   /* ro */
#define PCI_DEVICE_ID		0x02   /* ro */
#define PCI_COMMAND		0x04   /* device control */
#define PCI_STATUS		0x06
#define PCI_REVISION_ID 	0x08   /* ro */
#define PCI_PROG_IF_CODE	0x09   /* ro */
#define PCI_SUB_CLASS_CODE	0x0A   /* ro */
#define PCI_BASE_CLASS_CODE	0x0B   /* ro */
#define PCI_CACHE_LINE_SIZE	0x0C   /* ro+ */
#define PCI_LATENCY_TIMER	0x0D   /* ro+ */
#define PCI_HEADER_TYPE		0x0E   /* ro */
#define PCI_BIST		0x0F   /* Built in self test */
/*
 * Type 0 PCI offsets 
 */
#define PCI0_BASE_ADDR          0x10
#define PCI0_CIS                0x28
#define PCI0_SUB_VENDOR_ID      0x2C
#define PCI0_SUB_SYSTEM_ID      0x2E
#define PCI0_ROM_BASE_ADDR      0x30
#define PCI0_INTERRUPT_LINE     0x3C
#define PCI0_INTERRUPT_PIN      0x3D   /* ro */
#define PCI0_MINIMUM_GRANT      0x3E   /* ro */
#define PCI0_MAXIMUM_LATENCY    0x3F   /* ro */
/*
 * Device specific offsets
 */
#define PCI_DEVICE_SPECIFIC     0x40   /* 192 bytes */


int TLaserPCIA_access(int cpuNum, VA vA, int type, Reg *buff)
{  
   VA base = vA & BITMASK(31); /* clear top bits */
   uint64 width = (base >> WIDTH_SHIFT) & 0x3;
   uint64 pciAddr = (base>>IO_SHIFT);
   uint64 offset  = pciAddr & 0xff;
   VA ra;
   CPUVec.GetRegister(cpuNum, REG_RA, &ra);

   if (type==BDOOR_LOAD_WORD) {
      uint64 res = 0;
      switch (offset) { 
      case PCI_VENDOR_ID: 
	 res = 0x1077;  /* QLOGIC_ID */ /* too complicated */
	 res = 0x101A;  /* NCR_ID    */ /* does not work on tlaser (handler_add fails) */
	 res = 0x1011;  /* DEC_ID    */
         res = 0x1291 | ((0x1291) << 16);  /* simos: August 1st */
	 break;
      case PCI_DEVICE_ID: 
	 res = 0x1020;  /* QLOGIC: ISP1020 SCSI Adapter */ 
	 res = 0x0001;  /* NCR: NCR_810 */
	 res = 0x0008;  /* DEC: PZA SCSI */
         res = 0x1291;  /* simos:  August 1st */
	 break;
      case PCI_HEADER_TYPE: 
         res = 0;       /* type 0 header */ 
         break;
      case PCI0_INTERRUPT_LINE:
         res = 0x9 | (0x1 << 8);  /* Line and pin (latency and grant to 0) */
         break;
      default: 
         res = 0;
         CPUWarning("TLaser_PCI OFF=0x%lx PC=0x%lx RA=0x%lx vA=%lx pci=0x%lx %s\n",
                    offset, PE[cpuNum]->PC,ra,vA,pciAddr,bdoorType[type]);  
      }
      
      /*
       * PCI->Tlaser data conversion. See io/dec/pci/pcia.c 
       */
      *buff = res<<((pciAddr&3)<<3);

      return 0;
   }

   CPUPrint("TLaser_PCI OFF=0x%lx PC=0x%lx RA=0x%lx vA=%lx pci=0x%lx value=0x%lx %s\n",
	      offset, PE[cpuNum]->PC,ra,vA,pciAddr,*buff,bdoorType[type]);  

   return 0;
}




int TLaserISP_access(int cpuNum, VA vA, int type, Reg *buff)
{  
   VA off = vA  & 0xfffff;
   VA ra;
   CPUVec.GetRegister(cpuNum, REG_RA, &ra);

   if (type==BDOOR_LOAD_WORD) {
      switch (off) { 
      case 0x800:     /* Timer 1 register (2 reads - LSB, then MSB) */
         *buff = 0;
         return 0;
      }
   }
   
   if (type==BDOOR_STORE_WORD) {
      switch (off) {
      case 0x860:      /* Command 1 register */
         if(*buff == 0) /* SELECT_COUNTER_0 | SB_COUNTER_LATCH */
            return 0;
         else {
            CPUPrint("TLaser_ISP: Command1 0x%lx PC 0x%lx RA 0x%lx va=0x%lx type=%s \n",
                       *buff, 
                       PE[cpuNum]->PC,
                       ra,
                       vA,bdoorType[type]);
            return 0;
         }
      }
   }

   CPUPrint("TLaser_ISP: OFF 0x%lx PC 0x%lx RA 0x%lx va=0x%lx type=%s \n",
              off, 
              PE[cpuNum]->PC,
              ra,
              vA,bdoorType[type]);
   
   if (type==BDOOR_LOAD_WORD) {
      *buff = 0;
   }
   return 0;
}
