/*
 * Copyright (C) 1996-1998 by the Board of Trustees
 *    of Leland Stanford Junior University.
 * 
 * This file is part of the SimOS distribution. 
 * See LICENSE file for terms of the license. 
 *
 */


	/*
	 *  ms  -  The simulator component of MXS.
	 *
	 *	Simulates a variety of architectures, from a statically
	 *	scheduled R3000-like processor, to a dynamically
	 *	scheduled processor with speculative execution.
	 *
	 *	Jim Bennett
	 *	1993, 1994, 1995
	 */

#ifdef MIPSY_MXS
#include <sys/types.h>
#endif
#include "ms.h"

#ifdef MIPSY_MXS
#include "cpu_state.h"
#include "pcache.h"
#include "cp0.h"
#else
extern void ms_cache_callbacks (struct s_cpu_state *st);
#endif

int ms_writeback (struct s_cpu_state *st);
void ms_issue (struct s_cpu_state *st);
void ms_execute (struct s_cpu_state *st);
void ms_ldst_dep (struct s_cpu_state *st);
void ms_memory (struct s_cpu_state *st);


void free_inst (struct s_cpu_state *st, int inum);


#ifdef MIPSY_MXS
	/*
	 *  ms_cycle_once  -  Step the given processor forward by one
	 *			cycle.
	 */

void ms_cycle_once (struct s_cpu_state *st)
	{
	WorkList *w;

/* Main loop.  Each time around the loop we emulate a single CPU cycle,	*/
/* so this should be coded as tightly as possible.			*/

		IncStat(ST_RUN_CYCLES);
		st->work_cycle++;
		if (st->work_cycle >= TICK_COUNT)
			{		/* Every 100,000 cycles, reset	*/
					/* the work clock and put out	*/
					/* a tick mark.			*/
			st->work_cycle -= TICK_COUNT;
			st->work_ticks++;
#if 0
			if (st->icount == 
			    ((CPUState *)(st->mipsyPtr))->numInstructions) {
			    /* No isntruction graded since last time - trouble */        
			    ms_break (st, NULL, "TIMOUT");
			}
			st->icount =  ((CPUState *)(st->mipsyPtr))->numInstructions;
#endif
#ifdef PRINT_DOTS
			printf (".0x%x",st->grad.pc); fflush(stdout);
			if ((st->work_ticks & 0x3f) == 0) 
				printf ("\n\r0x%x", st->grad.pc);
#endif
			for (w=st->work_head; (w); w = w->next)
				w->cycle -= TICK_COUNT;
			}
#ifdef BREAKPOINT
		if ((st->work_ticks == (cycbrk/TICK_COUNT)) &&
		    (st->work_cycle == (cycbrk % TICK_COUNT)))
			ms_break (st, NULL, "CYCBRK");
#endif

	/* Perform the per cycle operations: Writeback, Fetch, Issue,	*/
	/* Execute, and Memory access.					*/

		ms_writeback (st);
		ms_fetch (st);
		ms_issue (st);
		ms_execute (st);
		ms_ldst_dep (st);     /* Update load/store dependencies	*/
		ms_memory (st);
		ms_graduate (st);
		ms_except (st);


	}

#else
	/*
	 *  ms_cycle_loop  -  All set up, let the machine run
	 */

void ms_cycle_loop (struct s_cpu_state *st)
	{
	int	watchdog;
	WorkList *w;

/* Main loop.  Each time around the loop we emulate a single CPU cycle,	*/
/* so this should be coded as tightly as possible.			*/

	watchdog = 0;
	while (1)
		{
#ifdef DEBUG_CHECKS
		watchdog++;
		if (watchdog >= 1000)
			ms_break (st, NULL, "TIMOUT");
#endif
		st->work_cycle++;
		if (st->work_cycle >= TICK_COUNT)
			{		/* Every 100,000 cycles, reset	*/
					/* the work clock and put out	*/
					/* a tick mark.			*/
			st->work_cycle -= TICK_COUNT;
			st->work_ticks++;
#ifdef PRINT_DOTS
			printf (".0x%x",st->grad.pc); fflush(stdout);
			if ((st->work_ticks & 0x3f) == 0) 
				printf ("\n\r0x%x", st->grad.pc);
#endif
			for (w=st->work_head; (w); w = w->next)
				w->cycle -= TICK_COUNT;
			}
#ifdef BREAKPOINT
		if ((st->work_ticks == (cycbrk/TICK_COUNT)) &&
		    (st->work_cycle == (cycbrk % TICK_COUNT)))
			ms_break (st, NULL, "CYCBRK");
#endif

	/* Perform the per cycle operations: Writeback, Fetch, Issue,	*/
	/* Execute, and Memory access.					*/

		if (ms_writeback (st))
			watchdog = 0;		/* Reset watchdog timer */
						/* every time some work */
						/* is done.             */
		ms_fetch (st);
		ms_issue (st);
		ms_execute (st);

		ms_ldst_dep (st);     /* Update load/store dependencies	*/
		ms_memory (st);
		ms_graduate (st);
		ms_cache_callbacks (st);

skip_to_end:	;

		}
	}

#endif /* MIPSY_MXS */

	/*
	 *  ms_writeback  -  WRITEBACK the results of previous operations,
	 *		and unlock registers.  Handled by the work list
	 *		mechanism (see the routine reg_writeback, below,
	 *		for the actual operations).
	 *
	 *	As an indicator for the watchdog timer, ms_writeback
	 *	returns the amount of work performed.
	 */

int ms_writeback (struct s_cpu_state *st)
	{
	int	loop_counter;
	WorkList *w;

	w=st->work_head;
	st->wbacks = 0;
	loop_counter = 0;
	while ((w) && (st->work_cycle >= w->cycle))
		{
		loop_counter++;
				/* Perform work list item if	*/
				/* it is due this cycle.	*/
		(*w->f)((void *)st, w->arg2);
		st->free_tail->next = w;
		st->free_tail = w;
		w = w->next;
		st->free_tail->next = NULL;
		if ((st->work_head = w) == NULL)
			st->work_tail = NULL;
		}
	return (loop_counter);
	}


	/*
	 *  update_excuse  -  Change the reason why a given register
	 *			is not available.
	 */

void update_excuse (void *st, void *ix)
	{
	int	reg_ix = (int)ix >> 1;

	STX->reg_excuse[reg_ix] = STX->new_excuse[reg_ix];
	STX->new_excuse[reg_ix] = ST_NO_EXCUSE;
	}


	/*
	 *  unstall_fetch  -  Clear a branch stall condition for a thread
	 *			
	 */

void unstall_fetch (void *st, void *ix)
	{
	THREAD	*th;

	th = &STX->threads [(int)ix];
	th->stall_branch = 0;
	UpdateStallFetch (th);
	}


	/*
	 *  finish_call  -  Complete call processing by unstalling the
	 *		    fetch unit and writing back the link pointer
	 */

void finish_call (void *st, void *ix)
	{
	INST	*ip;
	BrTREE	*br;
	THREAD	*th;
	int	inum, branch_node;

	ip = (INST *)ix;
	inum = ip - STX->iwin;
	branch_node = STX->iwin_br_node[inum];
	if (branch_node >= 0)
		{
		br = &STX->branch_tree [branch_node];
		th = &STX->threads [br->thread];
		th->stall_branch = 0;
		UpdateStallFetch (th);
		prune_branch (STX, branch_node);
		}
	if (ip->r1 == 0) 
	    free_inst(st,inum);
	else
	    reg_writeback (STX, (void *)ip->r1);
	}


	/*
	 *  finish_branch  -  Complete branch processing by unstalling the
	 *		      fetch unit and freeing the branch instruction
	 */

void finish_branch (void *st, void *ix)
	{
	INST	*ip;
	BrTREE	*br;
	THREAD	*th;
	int	inum, branch_node;

	ip = (INST *)ix;
	inum = ip - STX->iwin;
	branch_node = STX->iwin_br_node[inum];
	if (branch_node >= 0)
		{
		br = &STX->branch_tree [branch_node];
		th = &STX->threads [br->thread];
		th->stall_branch = 0;
		UpdateStallFetch (th);
		prune_branch (STX, branch_node);
		}
	free_inst (STX, inum);
	}


	/*
	 *  free_spec_inst  -  Free up instruction slot belonging to a
	 *			speculated instruction.
	 *
	 *	Just like free_inst, except that this instruction might
	 *	be on someone's dependents list, so we have to fix that.
	 *	Also, the register owned by this instruction needs to
	 *	be freed.
	 */

void free_spec_inst (struct s_cpu_state *st, int this_inst)
	{
	INST	*ip;
	int	i, j, inum, reg_ix;

		/* Update the dependent lists.				*/
	ip = &st->iwin [this_inst];
	if (ip->r2 > 0)
		{
		reg_ix = ip->r2 >> 1;
		if (this_inst == st->reg_otherhalf2 [reg_ix])
			st->reg_otherhalf2 [reg_ix] = -1;

		inum = st->reg_owner [reg_ix << 1];
		for (i=0, j=0; i<st->iwin_index2 [inum]; j++)
			{
			st->iwin_dep2 [inum][i] = st->iwin_dep2 [inum][j];
			if (st->iwin_dep2 [inum][j] == this_inst)
				st->iwin_index2 [inum]--;
			else
				i++;
			}

		if (inum != st->reg_owner [(reg_ix << 1) + 1])
		    {
		    inum = st->reg_owner [(reg_ix << 1) + 1];
		    for (i=0, j=0; i<st->iwin_index2 [inum]; j++)
			{
			st->iwin_dep2 [inum][i] = st->iwin_dep2 [inum][j];
			if (st->iwin_dep2 [inum][j] == this_inst)
				st->iwin_index2 [inum]--;
			else
				i++;
			}
		    }
		}

	if (ip->r3 > 0)
		{
		reg_ix = ip->r3 >> 1;
		if (this_inst == st->reg_otherhalf3 [reg_ix])
			st->reg_otherhalf3 [reg_ix] = -1;

		inum = st->reg_owner [reg_ix << 1];
		for (i=0, j=0; i<st->iwin_index3 [inum]; j++)
			{
			st->iwin_dep3 [inum][i] = st->iwin_dep3 [inum][j];
			if (st->iwin_dep3 [inum][j] == this_inst)
				st->iwin_index3 [inum]--;
			else
				i++;
			}

		if (inum != st->reg_owner [(reg_ix << 1) + 1])
		    {
		    inum = st->reg_owner [(reg_ix << 1) + 1];
		    for (i=0, j=0; i<st->iwin_index3 [inum]; j++)
			{
			st->iwin_dep3 [inum][i] = st->iwin_dep3 [inum][j];
			if (st->iwin_dep3 [inum][j] == this_inst)
				st->iwin_index3 [inum]--;
			else
				i++;
			}
		    }
		}

		/* Take care of the target (destination) register	*/
	if (ip->r1 > 0)
		{
			/* Special registers get the full treatment,	*/
			/* since they might have dependents that aren't	*/
			/* descended from this thread.			*/
		if (st->iwin_flags[this_inst] & IWIN_CTL)
			{
			reg_writeback ((void *)st, (void *)ip->r1);
			return;
			}
		else
			{
			REGSTAT	*rs;

			reg_ix = ip->r1 >> 1;
			rs = &st->reg_rstat [reg_ix];
			if (rs->reg_status & REG_DMAP)
				rs->reg_status &= ~REG_DMAP;
			else
				{
				rs->reg_status &= ~REG_IN_WIN;
				st->reg_excuse[reg_ix] = ST_NO_EXCUSE;
				st->new_excuse[reg_ix] = ST_NO_EXCUSE;
				CheckRegFree (st, rs, reg_ix);
				}
			}
		}

		/* Then free like a regular instruction.		*/
	free_inst (st, this_inst);
	}


	/*
	 *  reg_writeback  -  Perform WRITEBACK actions:
	 *
	 *	1.  Free slot of completing instruction
	 *	2.  Decrement ref counts of source registers for
	 *		completing instruction
	 *	3.  Clear dependent flags of instructions waiting for
	 *		this result
	 *	4.  Update register status (no longer "in window")
	 *	5.  Free load/store buffer entry of loads
	 */

void reg_writeback (void *st, void *ix)
	{
	int	inum, inum2, dwi;
	int	reg_ix = (int)ix >> 1;
	REGSTAT	*rs;
	WorkDecls;

#ifdef DEBUG_CHECKS
	if ((int)ix == 0) {
	    fprintf(stderr, "r0 passed to reg_writeback\n");
	    ms_break(st, NULL, "r0write");
	}
#endif
		/* If we haven't exceeded allowed write backs, unlock	*/
		/* the register.					*/
	if (STX->wbacks < WBACK_WIDTH)
		{
#ifdef BREAKPOINT
		if ((int)ix == trace_preg)
			printf (
		  "Register %d value = 0x%8.8x  %12.6f  %12.6f\r\n", (int)ix,
		    STX->regs[(int)ix], *(float *)&STX->regs[(int)ix],
					*(double *)&STX->regs[(int)ix] );
#endif
		STX->wbacks++;
		inum = STX->reg_owner [(int)ix];
		rs = &STX->reg_rstat [reg_ix];
		if (rs->reg_status & REG_DMAP)
			{
			rs->reg_status &= ~REG_DMAP;

				/* If this is the first half of a two	*/
				/* instruction definition, turn off the	*/
				/* dependent flags of the instruction	*/
				/* that writes the other half of the	*/
				/* register.				*/
			inum2 = STX->reg_otherhalf2 [reg_ix];
			if (inum2 >= 0)
				{
				STX->iwin_flags [inum2] &= ~IWIN_DEP2;
				CheckInstAvail (STX, inum2);
				}
			inum2 = STX->reg_otherhalf3 [reg_ix];
			if (inum2 >= 0)
				{
				STX->iwin_flags [inum2] &= ~IWIN_DEP3;
				CheckInstAvail (STX, inum2);
				}

				/* Transfer dependent list to the other	*/
				/* instruction that defines this register */
			inum2 = STX->reg_owner [((int)ix) ^ 0x01];
			STX->reg_owner [(int)ix] = inum2;

			for (dwi = 0; dwi < STX->iwin_index2[inum]; dwi++)
			  STX->iwin_dep2 [inum2] [STX->iwin_index2[inum2]++] =
				STX->iwin_dep2 [inum] [dwi];

			for (dwi = 0; dwi < STX->iwin_index3[inum]; dwi++)
			  STX->iwin_dep3 [inum2] [STX->iwin_index3[inum2]++] =
				STX->iwin_dep3 [inum] [dwi];
			}
		else
			{
				/* Turn off REG_IN_WIN status of this	*/
				/* reg and see if it is now free.	*/
			rs->reg_status &= ~REG_IN_WIN;
			STX->reg_excuse[reg_ix] = ST_NO_EXCUSE;
			STX->new_excuse[reg_ix] = ST_NO_EXCUSE;
			CheckRegFree (STX, rs, reg_ix);

				/* Turn off dependent flags of inst's	*/
				/* waiting for this result.		*/
			for (dwi = 0; dwi < STX->iwin_index2[inum]; dwi++)
				{
				inum2 = STX->iwin_dep2 [inum] [dwi];
				STX->iwin_flags [inum2] &= ~IWIN_DEP2;
				CheckInstAvail (STX, inum2);
				}
			for (dwi = 0; dwi < STX->iwin_index3[inum]; dwi++)
				{
				inum2 = STX->iwin_dep3 [inum] [dwi];
				STX->iwin_flags [inum2] &= ~IWIN_DEP3;
				CheckInstAvail (STX, inum2);
				}
			}

			/* Then release this instruction slot.		*/
		free_inst (STX, inum);
		}
	else
		/* Otherwise, put it back on the queue for next cycle	*/
		{
		Add_to_worklist (STX, 1, reg_writeback, ix);
		}
	}

void reg0_writeback (void *st, void *ix)
       {
       STX->regs[0] = 0;  /* Keep R0 zero */
       free_inst (STX, (int) ix);
       }


	/*
	 *  free_inst  -  Free up instruction slot in instruction window
	 *
	 *	For the precise interrupts model, just set a flag to
	 *	indicate that the instruction can be freed.
	 */

void free_inst (struct s_cpu_state *st, int inum)
	{
	int	reg_ix;
	REGSTAT	*rs;
	INST	*ip;

#ifdef DEBUG_CHECKS
	if (!(st->iwin_flags [inum] & IWIN_BUSY))
		{
		fprintf (stderr, "Error: instruction slot already free\r\n");
		ms_break (st, NULL, "ERR");
		}
#ifdef PRECISE
	if (st->iwin_flags [inum] & IWIN_FREED)
		{
		fprintf (stderr, "Error: instruction slot already free\r\n");
		ms_break (st, NULL, "ERR"); 
		}
#endif /* PRECISE */
/* These checks are somewhat expensive.  Comment out for now:
	{
	int	i, pri;

	for (i=0, pri=st->iwin_head_ldst; (i<10) && (pri >= 0);
					i++, pri = st->iwin_ldst [pri])
		{
		if (pri == inum)
			{
			fprintf (stderr,
				"Error: freed instruction still in use\r\n");
			ms_break (st, NULL, "ERR");
			}
		}

	for (i=0, pri=st->iwin_headpri; (i<10) && (pri >= 0);
					i++, pri = st->iwin_pri [pri])
		{
		if (pri == inum)
			{
			fprintf (stderr,
				"Error: freed instruction still in use\r\n");
			ms_break (st, NULL, "ERR");
			}
		}

	for (i=0, pri=st->iwin_next_avail; (i<10) && (pri >= 0);
					i++, pri = st->iwin_avail_list [pri])
		{
		if (pri == inum)
			{
			fprintf (stderr,
				"Error: freed instruction still in use\r\n");
			ms_break (st, NULL, "ERR");
			}
		}

	for (i=0, pri=st->iwin_nextfree; (i<10) && (pri >= 0);
					i++, pri = st->iwin_freelist [pri])
		{
		if (pri == inum)
			{
			fprintf (stderr,
			    "Error: Error: instruction slot already free\r\n");
			ms_break (st, NULL, "ERR");
			}
		}
	}
*/
#endif /* DEBUG_CHECKS */

#ifdef PRECISE
	st->iwin_flags [inum] |= IWIN_FREED;
#else
	UnthreadInst (st, inum);
#endif


		/* Decrement reference counts of source		*/
		/* registers for this instruction.		*/
	ip = &st->iwin [inum];
	if (ip->r2 > 0)
		{
		reg_ix = ip->r2 >> 1;
		rs = &st->reg_rstat [reg_ix];
		rs->reg_ref--;
		CheckRegFree (st, rs, reg_ix);
		}

	if (ip->r3 > 0)
		{
		reg_ix = ip->r3 >> 1;
		rs = &st->reg_rstat [reg_ix];
		rs->reg_ref--;
		CheckRegFree (st, rs, reg_ix);
		}
	}


#ifndef INLINE

	/*
	 *  UnthreadInst  -  Remove instruction from its thread and
	 *	put it back on the free list.
	 */

void UnthreadInst (struct s_cpu_state *st, int inum)
	{
	int	br_node;

		/* Remove from the instruction chain of its thread	*/
	br_node = st->iwin_br_node [inum];
	if (br_node >= 0)
		{
		int	tmpi;
		BrTREE	*br;

		br = &st->branch_tree [br_node];
		if ((tmpi = st->iwin_bthread [inum]) >= 0)
			st->iwin_thread [tmpi] = st->iwin_thread [inum];
		else
			br->iwin_head_th = st->iwin_thread [inum];

		if ((tmpi = st->iwin_thread [inum]) >= 0)
			st->iwin_bthread [tmpi] = st->iwin_bthread [inum];
		else
			br->iwin_tail_th = st->iwin_bthread [inum];
		}

		/* Put instruction back on free list	*/
	st->iwin_freelist [inum] = st->iwin_nextfree;
	st->iwin_nextfree = inum;
	st->iwin_flags [inum] = 0;
	st->iwin_index2 [inum] = 0;
	st->iwin_index3 [inum] = 0;
	}


	/*
	 *  AcquireRegMap  -  Mark physical register mapped, and track
	 *			number of times it is mapped.
	 */

void AcquireRegMap (REGSTAT *rs, THREAD *th, int lreg)
	{
	rs->reg_nmap++;
	rs->reg_status |= REG_MAPPED;
	if ((rs->reg_status & REG_CLAIMED) &&
	    (th->half_def[lreg >> 1] & REGNAME_CLAIM))
		rs->reg_nclaims++;
	}


	/*
	 *  ReleaseRegMap  -  Decrement map count for register and mark
	 *			it unmapped when the count reaches 0.
	 */

void ReleaseRegMap (struct s_cpu_state *st, THREAD *th, int reg_ix, int lreg)
	{
	int	lreg_ix;
	REGSTAT	*rs;

	rs = &st->reg_rstat[reg_ix];
	rs->reg_nmap--;
	if (rs->reg_nmap == 0)
		{
		rs->reg_status &= ~REG_MAPPED;
		CheckRegFree (st, rs, reg_ix);
		}

	lreg_ix = lreg>>1;
	if ((rs->reg_status & REG_CLAIMED) &&
	    (th->half_def[lreg_ix] & REGNAME_CLAIM))
		{
		rs->reg_nclaims--;
		CheckRegFree (st, rs, reg_ix);
		if ((rs->reg_status & REG_CLAIMED) == 0)
			th->half_def[lreg_ix] &= ~REGNAME_CLAIM;
		}
	}


	/*
	 *  CheckRegFree  -  Update register state
	 *
	 *  For normal registers, they become free when the number of
	 *  references drops to zero, and there is no other reason to
	 *  stay busy (i.e. they aren't still mapped and aren't in the
	 *  instruction window).
	 *
	 *  In the PRECISE case, when the writer of a register graduates,
	 *  the prior name for that register is available.  This is
	 *  indicated by setting the REG_FREED flag.
	 *
	 *  For special registers, they stay mapped all the time, so
	 *  just check if they are still in the instruction window.
	 *
	 *  For claimed registers (the other half of double precision
	 *  registers), unclaim them if writeback has happened, the
	 *  number of claims is zero, and the number of references
	 *  is zero.  This guarantees that a subsequent claimant can
	 *  write to the register without interfering with any thread
	 *  relying on a prior value in that register.
	 *
	 *  The reg number for this routine has already been divided by 2
	 */

void CheckRegFree (struct s_cpu_state *st, REGSTAT *rs, int reg_ix)
	{
	if ((rs->reg_status & REG_CLAIMED) && (rs->reg_nclaims <= 0) &&
            (rs->reg_ref <= 0) && ((rs->reg_status & REG_IN_WIN) == 0) )
		rs->reg_status &= ~REG_CLAIMED;

#ifdef PRECISE
	if (rs->reg_status & REG_FREED)
		{
		if (rs->reg_status & REG_MAPPED)
			rs->reg_status = REG_MAPPED;
		else
			{
			rs->reg_status = 0;
			st->reg_freelist [reg_ix] = st->reg_nextfree;
			st->reg_nextfree = reg_ix;
			}
		}
	if (preg_is_special (reg_ix) && ((rs->reg_status &
			  (REG_BUSY | REG_IN_WIN | REG_MAPPED | REG_DMAP))
						== (REG_BUSY + REG_MAPPED)) )
			rs->reg_status = REG_MAPPED;
#else
	if ((rs->reg_nclaims <= 0) && (rs->reg_ref <= 0) &&
	    ((rs->reg_status & REG_IN_WIN) == 0) )
		{
		if ((rs->reg_status &
		    (REG_BUSY | REG_IN_WIN | REG_MAPPED | REG_DMAP))
						== REG_BUSY)
			{
			rs->reg_status = 0;
			st->reg_freelist [reg_ix] = st->reg_nextfree;
			st->reg_nextfree = reg_ix;
			}
		else
		if (preg_is_special (reg_ix) && ((rs->reg_status &
			  (REG_BUSY | REG_IN_WIN | REG_MAPPED | REG_DMAP))
						== (REG_BUSY + REG_MAPPED)) )
			rs->reg_status = REG_MAPPED;
		}
#endif /* PRECISE */
	}


	/*
	 *  CheckInstAvail  -  Check if the instruction is now available for
	 *			execution
	 *
	 *	Inum is the index of the instruction to check
	 */

void CheckInstAvail (struct s_cpu_state *st, int inum)
	{

#ifdef DEBUG_CHECKS
	if (!(st->iwin_flags [inum] & IWIN_BUSY))
		{
		fprintf (stderr, "Instruction slot not in use\r\n");
		ms_break (st, NULL, "ERR");
		}
#endif

		/* Check if the instruction is unencumbered	*/
	if (st->iwin_flags [inum] &
(IWIN_DEP2 | IWIN_DEP3 | IWIN_FLUSH | IWIN_AVAIL | IWIN_LDST_DEP | IWIN_BRDLY))
		return;

		/* Don't allow speculative writes to special registers	*/
	if ((st->iwin_flags [inum] & (IWIN_SPEC | IWIN_CTL)) ==
						(IWIN_SPEC | IWIN_CTL))
		return;

		/* Then mark it available and add it to the available	*/
		/* list.						*/
	st->iwin_flags [inum] |= IWIN_AVAIL;

#ifdef PRIORITIZE_AVAIL_LIST
		/* Insert the instruction into the available list in	*/
		/* priority order.					*/
	pri = st->iwin_headpri;
	avail = st->iwin_next_avail;
	last_avail = -1;
	while (pri != inum)
		{
#ifdef DEBUG_CHECKS
		if (pri < 0)
			{
			fprintf (stderr,
		  "Error - Available instruction not in priority list\r\n");
			ms_break (st, NULL, "ERR");
			}
#endif
		if (pri == avail)
			{
			last_avail = avail;
			avail = st->iwin_avail_list [avail];
			}
		pri = st->iwin_pri [pri];
		}

	st->iwin_avail_list [inum] = avail;
	if (last_avail < 0)
		st->iwin_next_avail = inum;
	else
		st->iwin_avail_list [last_avail] = inum;

#else
		/* Otherwise just add newly available instructions	*/
		/* to the tail of the available list.			*/
	if (st->iwin_next_avail >= 0)
		{
		st->iwin_avail_list [st->iwin_last_avail] = inum;
		st->iwin_last_avail = inum;
		}
	else
		{
		st->iwin_next_avail = inum;
		st->iwin_last_avail = inum;
		}
	st->iwin_avail_list [inum] = -1;
#endif
	}

#endif /* !INLINE */


	/*
	 *  ms_issue  -  ISSUE instructions that are ready to go.
	 *		 Chase down the priority chain, issuing the first
	 *		 available instructions.
	 */

void ms_issue (struct s_cpu_state *st)
	{
	int	pri, prev_pri, reg_ix, ex_count, issues;

		/* If issue is stalled, skip it and retry execution.	*/
		/* Pending interrupts stall issue.			*/
	if (st->exception_pending)
		{
		AddToStat (ST_ISSUE_STALLED, ISSUE_WIDTH);
		AddToStat (ST_EXCEPT, ISSUE_WIDTH);
		return;
		}

	if (st->stall_issue)
		{
		AddToStat (ST_ISSUE_STALLED, ISSUE_WIDTH);
		AddToStat (st->stall_type, ISSUE_WIDTH);
#ifdef DEBUG_CHECKS
		if (st->stall_type == ST_NO_EXCUSE)
			{
			fprintf (stderr,  "No excuse stall type\r\n");
			ms_break (st, NULL, "ERR");
			}
#endif
		return;
		}

	AddToStat (ST_ISSUE_TRIED, ISSUE_WIDTH);
	for (issues=0, ex_count=0;
	     (issues<ISSUE_WIDTH) && (ex_count<(2*ISSUE_WIDTH)); )
		{
			/* Get next available instruction	*/
		pri = st->iwin_next_avail;
		prev_pri = -1;
		while (pri >= 0)
		    {
		    if ((st->iwin_flags [pri] & IWIN_LDST) && (!PhaseA (pri)))
			{
				/* If it's a memory operation, make	*/
				/* sure that there is room for it.	*/
			if (ldst_buffer_reserve (st) >= 0) break;

				/* No room, proceed to next instruction	*/
			prev_pri = pri;
			pri = st->iwin_avail_list [pri];
			continue;
			}
		    break;
		    }
		if (pri < 0) break;

		if (pri == st->iwin_next_avail)
		    st->iwin_next_avail = st->iwin_avail_list [pri];
		else
		    st->iwin_avail_list [prev_pri] = st->iwin_avail_list [pri];

		if (pri == st->iwin_last_avail)
		    st->iwin_last_avail = prev_pri;

#ifdef DEBUG_CHECKS
		if (!(st->iwin_flags [pri] & IWIN_BUSY))
			{
			fprintf (stderr,
			  "Inconsistent instruction window state\r\n");
			ms_break (st, NULL, "ERR");
			}
#endif

			/* Then issue to execution unit		*/
		st->ex [ex_count++] = pri;
		st->iwin_flags [pri] |= IWIN_ISSUED;

			/* and remove from priority chain,	*/
			/* except for Phase A of loads/stores.	*/
		if (st->iwin_flags [pri] & IWIN_LDST)
		    {
		    if (!PhaseA (pri))
			{
#ifdef PREF_DEBUG
               {
                  INST *ip = &st->iwin [pri];
                  if (is_store(ip->op))
                     CPUPrint("CPU %d ISSUE Store addr = %08x cycle = %d\n", 
                              CPUNUM(st), Ireg(ip->r2) + ip->imm, st->work_cycle);
                  else if (is_load(ip->op))
                     CPUPrint("CPU %d ISSUE Load addr = %08x cycle = %d\n", 
                              CPUNUM(st), Ireg(ip->r2) + ip->imm, st->work_cycle);
                  else if (is_prefetch(ip->op))
                     CPUPrint("CPU %d ISSUE Prefetch addr = %08x cycle = %d\n", 
                              CPUNUM(st), Ireg(ip->r2) + ip->imm, st->work_cycle);
               }
#endif
               
			ms_ldst_dequeue (st, pri);
			ms_pri_dequeue (st, pri);
			IncStat (ST_EXECUTED);
			issues++;

#ifdef ONE_PHASE_LS
				/* Update store count and set flag to	*/
				/* force update of load/store status.	*/
			if (st->iwin_flags [pri] & IWIN_STORE)
				st->iwin_nstores--;
#endif
			}
		    else
			{
				/* Don't count Phase A as issue	*/
			IncStat (ST_PHASE_A);
			}
		    }
		else
		    {
		    ms_pri_dequeue (st, pri);
		    IncStat (ST_EXECUTED);
		    issues++;
		    }
		}

		/* Special check for flush instructions (not issued	*/
		/* until all preceeding instructions have issued).	*/
	if (ex_count == 0)
		{
			/* If there is exactly one instruction in the	*/
			/* buffer, and it is waiting for the window to	*/
			/* flush, and it belongs to branch node 0 (=>	*/
			/* it is no longer speculative), and it has no	*/
			/* unresolved dependencies, then issue it.	*/
		pri = st->iwin_headpri;
		if ((pri >= 0) && (st->iwin_flags [pri] & IWIN_FLUSH) &&
		    (pri == st->iwin_tailpri) &&
#ifdef PRECISE
		    (pri == st->iwin_headgrad) &&
#endif
		    (st->iwin_br_node [pri] == 0) &&
		    (!(st->iwin_flags [pri] & (IWIN_DEP2 | IWIN_DEP3))) )
			{
			st->ex [ex_count++] = pri;
			st->iwin_flags [pri] |= IWIN_ISSUED;
			if (st->iwin_flags [pri] & IWIN_LDST)
				ms_ldst_dequeue (st, pri);
			ms_pri_dequeue (st, pri);
			IncStat (ST_EXECUTED);
			issues++;
			}
		}

		/* Update statistics of successful and unsuccessful	*/
		/* instruction issue tries.				*/
	for (pri = st->iwin_headpri; issues<ISSUE_WIDTH;
					pri = st->iwin_pri [pri] )
		{
		if (pri < 0)
			{
			AddToStat (ST_IWIN_EMPTY, ISSUE_WIDTH-issues);
			break;
			}

		if (!(st->iwin_flags [pri] & IWIN_ISSUED) ||
		     (st->iwin_flags [pri] & IWIN_LDST_DEP)) 
			{
				/* If it isn't issued because it	*/
				/* depends on an inst still in the	*/
				/* window, skip it.			*/
			if (st->iwin_flags [pri] & IWIN_DEP2)
				{
				reg_ix = st->iwin [pri] .r2 >> 1;
				if (st->reg_excuse [reg_ix] == ST_NOT_ISSUED)
					continue;
				}
			if (st->iwin_flags [pri] & IWIN_DEP3)
				{
				reg_ix = st->iwin [pri] .r3 >> 1;
				if (st->reg_excuse [reg_ix] == ST_NOT_ISSUED)
					continue;
				}

			issues++;
			IncStat (ST_INST_STALLED);
			if (st->iwin_flags [pri] & IWIN_LDST_DEP)
				{
				IncStat (ST_LDST_DEP);
				}
			else
			if (st->iwin_flags [pri] & IWIN_DEP2)
				{
				IncStat (ST_REG_DEP);
				reg_ix = st->iwin [pri] .r2 >> 1;
				IncStat (st->reg_excuse [reg_ix]);
#ifdef DEBUG_CHECKS
				if (st->reg_excuse [reg_ix] == ST_NO_EXCUSE)
					{
			fprintf (stderr,  "Register absent without excuse\r\n");
					ms_break (st, NULL, "ERR");
					}
#endif
				}
			else
			if (st->iwin_flags [pri] & IWIN_DEP3)
				{
				IncStat (ST_REG_DEP);
				reg_ix = st->iwin [pri] .r3 >> 1;
				IncStat (st->reg_excuse [reg_ix]);
#ifdef DEBUG_CHECKS
				if (st->reg_excuse [reg_ix] == ST_NO_EXCUSE)
					{
			fprintf (stderr,  "Register absent without excuse\r\n");
					ms_break (st, NULL, "ERR");
					}
#endif
				}
			else
			if (st->iwin_flags [pri] & IWIN_FLUSH)
				{
				IncStat (ST_IWIN_FLUSH);
				}
			else
			if (st->iwin_flags [pri] & IWIN_BRDLY)
				{
				IncStat (ST_BR_DLY);
				}
			else
			if ((st->iwin_flags [pri] & (IWIN_SPEC | IWIN_CTL)) ==
						(IWIN_SPEC | IWIN_CTL))
				{
				IncStat (ST_SPEC_CTL);
				}
			else
			if (st->iwin_flags [pri] & IWIN_LDST)
				{
				if (ldst_buffer_reserve (st) >= 0)
					{
					ldst_buffer_release (st);
					IncStat (ST_NO_EXCUSE);
					}
				else
					{
					IncStat (ST_LDST_FULL);
					}
				}
			else
				{
				IncStat (ST_NO_EXCUSE);
#ifdef DEBUG_CHECKS
				if ((ex_count<(2*ISSUE_WIDTH)) &&
				    (st->stall_type == ST_NO_EXCUSE))
					{
		fprintf (stderr,  "No excuse for unissued instruction\r\n");
					ms_break (st, NULL, "ERR");
					}
#endif
				}
			}
		}

	st->ex_count = ex_count;
	}



	/*
	 *  ms_pri_dequeue  -  Remove instruction from priority queue
	 */

void ms_pri_dequeue (struct s_cpu_state *st, int pri)
	{
	int	inum;

	st->iwin_ninst--;
	if ((inum = st->iwin_bpri [pri]) >= 0)
		st->iwin_pri [inum] = st->iwin_pri [pri];
	else
		st->iwin_headpri = st->iwin_pri [pri];

	if ((inum = st->iwin_pri [pri]) >= 0)
		st->iwin_bpri [inum] = st->iwin_bpri [pri];
	else
		st->iwin_tailpri = st->iwin_bpri [pri];
	}


	/*
	 *  ms_pri_enqueue  -  Add instruction to priority queue
	 */

void ms_pri_enqueue (struct s_cpu_state *st, int pri)
	{
	st->iwin_ninst++;
	if (st->iwin_tailpri >= 0)
		{
		st->iwin_pri [st->iwin_tailpri] = pri;
		st->iwin_bpri [pri] = st->iwin_tailpri;
		st->iwin_pri [pri] = -1;
		st->iwin_tailpri = pri;
		}
	else
		{
		st->iwin_headpri = pri;
		st->iwin_tailpri = pri;
		st->iwin_bpri [pri] = -1;
		st->iwin_pri [pri] = -1;
		}
	}



	/*
	 *  ms_ldst_dequeue  -  Remove instruction from load/store queue
	 */

void ms_ldst_dequeue (struct s_cpu_state *st, int inum)
	{
	int	tmpi;

#ifdef DEBUG_CHECKS
	st->iwin_nldst--;
	if (st->iwin_nldst < 0)
		{
		fprintf (stderr, "Underflow in load/store queue\r\n");
		ms_break (st, NULL, "ERR");
		}
#endif
	if ((tmpi = st->iwin_bldst [inum]) >= 0)
		st->iwin_ldst [tmpi] = st->iwin_ldst [inum];
	else
		st->iwin_head_ldst = st->iwin_ldst [inum];

	if ((tmpi = st->iwin_ldst [inum]) >= 0)
		st->iwin_bldst [tmpi] = st->iwin_bldst [inum];
	else
		st->iwin_tail_ldst = st->iwin_bldst [inum];
	}


	/*
	 *  ms_ldst_enqueue  -  Add instruction to load/store queue
	 */

void ms_ldst_enqueue (struct s_cpu_state *st, int inum)
	{
#ifdef DEBUG_CHECKS
	st->iwin_nldst++;
	if (st->iwin_nldst > IWIN_SIZE)
		{
		fprintf (stderr, "Overflow in load/store queue\r\n");
		ms_break (st, NULL, "ERR");
		}
#endif
	st->iwin_flags [inum] |= IWIN_LDST;
	if (st->iwin_tail_ldst >= 0)
		{
		st->iwin_ldst [st->iwin_tail_ldst] = inum;
		st->iwin_bldst [inum] = st->iwin_tail_ldst;
		st->iwin_ldst [inum] = -1;
		st->iwin_tail_ldst = inum;
		}
	else
		{
		st->iwin_head_ldst = inum;
		st->iwin_tail_ldst = inum;
		st->iwin_bldst [inum] = -1;
		st->iwin_ldst [inum] = -1;
		}
	}



	/*
	 *  ms_execute  -  EXECUTE the instructions in the execution unit.
	 */
 
void ms_execute (struct s_cpu_state *st)
	{
	INST	*ip;
	BrTREE	*br;
	THREAD	*th;
	int	exi, ex_count, inum;

	ex_count = st->ex_count;
	exi = (st->stall_issue ? st->stall_issue-1 : 0);
	for (; exi < ex_count; exi++)
		{
		st->stall_issue = 0;
		st->stall_type = ST_NO_EXCUSE;
		inum = st->ex [exi];
		ip = &st->iwin [inum];
		if (st->iwin_flags [inum] & IWIN_SQUASH)
			{
			if (st->iwin_flags [inum] & IWIN_LDST)
			    {
			    if (PhaseA(inum))
				{
				ms_ldst_dequeue (st, inum);
				ms_pri_dequeue (st, inum);
				IncStat (ST_PRUNED);
				}
			    else
				ldst_buffer_release (st);
			    }
			free_spec_inst (st, inum);
			continue;
			}

		br = &st->branch_tree [st->iwin_br_node[inum]];
		th = &st->threads [br->thread];
#ifdef BREAKPOINT
		st->iwin_curi = inum;
#endif
		if (PhaseA (inum))
			{	/* Special case for Phase A issue of	*/
				/* load/store instructions.		*/
				/* Do the address computation		*/
#ifndef ONE_PHASE_LS
			st->iwin_addr [inum] = Ireg(ip->r2) + ip->imm;
			if (!is_prefetch(ip->op) && (ip->r1 >= 0))
				st->reg_excuse[ip->r1>>1] = ST_PC_DLY;
#ifdef MIPSY_MXS
		        {
			    PA paddr;
			    uint tlbFlavor;
			    Result ret;

             if (is_store(ip->op))
                tlbFlavor = TLB_WRITING;
             else
                tlbFlavor = TLB_READING;

             if (is_prefetch(ip->op)) {
                ret = TranslateVirtualNoSideeffect((CPUState *)st->mipsyPtr,
                                                   (VA) st->iwin_addr[inum],
                                                   &paddr);
                if (ret != SUCCESS)
                   st->iwin_paddr[inum] = (PA) -1;
                else
                   st->iwin_paddr[inum] = paddr;
                st->iwin_flags [inum] |=
                   (IWIN_ADDR_VALID | IWIN_LDST_DEP);
                st->iwin_flags [inum] &= ~IWIN_AVAIL;
             }
             else {
                   void *bdoorAddr = 0;

			    ret = TranslateVirtual((CPUState *)st->mipsyPtr,
						   (VA) st->iwin_addr[inum],
						   &paddr, &tlbFlavor, &bdoorAddr);
			    if (ret != SUCCESS) {
				st->iwin_except [inum] = GetLastException(st);
				st->iwin_paddr[inum] = (PA) -1;
				st->iwin_flags [inum] |=
			(IWIN_TLBFAULT | IWIN_ADDR_VALID | IWIN_LDST_DEP);
				st->iwin_flags [inum] &= ~IWIN_AVAIL;
			    } else {
				st->iwin_paddr[inum] = paddr;
				st->iwin_flags [inum] |=
					(IWIN_ADDR_VALID | IWIN_LDST_DEP);
				st->iwin_flags [inum] &= ~IWIN_AVAIL;

				if (tlbFlavor) {
				   st->iwin_flags[inum] |= IWIN_UNCACHED; 
				}
			    }
             }
			}
#else /* MIPSY_MXS */
			st->iwin_paddr [inum] = st->iwin_addr [inum];
			st->iwin_flags [inum] |=
				(IWIN_ADDR_VALID | IWIN_LDST_DEP);
			st->iwin_flags [inum] &= ~(IWIN_AVAIL | IWIN_ISSUED);
#endif /* MIPSY_MXS */
#endif /* !ONE_PHASE_LS */
			}
		else
			{
			(*optab[ip->op])(st, ip, th);
			Ireg(0) = 0;  /* Keep r0 == 0 */

				/* If stalled, remember where.		*/
			if (st->stall_issue)
				{
				st->stall_issue = exi+1;
				break;
				}

				/* Writeback reg if single cycle inst	*/
				/* and free up instruction slot.	*/
			        /* No need or want to writeback faulted inst or
				 * load/store instructions . */
			if ((lat_tab [ip->op] <= 1) && 
			    !(st->iwin_flags[inum] & (IWIN_FAULT+IWIN_LDST)))
				{
				if (ip->r1 > 0)
				    reg_writeback ((void *)st, (void *)ip->r1);
				else
				    free_inst (st, inum);
				}
			}
		}
	if (exi == st->ex_count)
	    st->ex_count = 0;
	}



#ifdef BREAKPOINT
void ms_break (struct s_cpu_state *st, INST *ip, char *typ)
	{
	int  t, flagp, inum;
	int	branch_node;
	BrTREE	*br;
	THREAD	*th;

		/* If no instruction specified for the breakpoint, then	*/
		/* pick a recent instruction of some current thread.	*/
	if (ip == NULL)
		{
		br = &st->branch_tree[0];
		while (br->lchild >= 0)
			{
			if (st->branch_tree[br->lchild].iwin_tail_th >= 0)
				br = &st->branch_tree[br->lchild];
			else
				break;
			}
		inum = br->iwin_tail_th;
		ip = &st->iwin[inum];
		}
	else
		{
		inum = ip - st->iwin;
		branch_node = st->iwin_br_node [inum];
		br = &st->branch_tree[branch_node];
		}

	th = &st->threads[br->thread];

	printf ("%s:", typ);
	printf ("% 7d%0.5d @0x%8.8x: ",
		st->work_ticks, st->work_cycle, th->debugpc);
	print_inst (stdout, ip);

	flagp = 0;
	if ((ip->r2 > 0) && (ip->r2  < MAX_VAR))
		{
		t = th->regnames[ip->r2];
		printf (" [");
		print_reg_contents (stdout, ip->r2, &st->regs[t]);
		flagp = 1;
		}
	if ((ip->r3 > 0) && (ip->r3  < MAX_VAR))
		{
		t = th->regnames[ip->r3];
		if (flagp)
			printf (", ");
		else
			printf (" [");
		print_reg_contents (stdout, ip->r3, &st->regs[t]);
		flagp = 1;
		}

	if (flagp)
		printf ("]\r\n");
	else
		printf ("\r\n");
	}
#endif
