/*
 * Copyright (C) 1996-1998 by the Board of Trustees
 *    of Leland Stanford Junior University.
 * 
 * This file is part of the SimOS distribution. 
 * See LICENSE file for terms of the license. 
 *
 */


	/*
	 *  ms_compile  -  Compile a single MIPS instruction into an
	 *		   instruction that the MXS simulator can deal with.
	 *
	 *	Jim Bennett
	 *	1993, 1994, 1995
	 */

#include <stdlib.h>
#include "ms.h"
#include "mc.h"

int mc_spec (uint inst, uint addr, INST *ip);
int mc_regimm (uint inst, uint addr, INST *ip);
int mc_cop1 (uint inst, uint addr, INST *ip);
void mc_cop0 (uint inst, uint addr, INST *ip);
void mc_fp_s (uint inst, uint addr, INST *ip);
void mc_fp_d (uint inst, uint addr, INST *ip);


	/*
	 *  compile_inst  -  Fetch instruction at the given address and
	 *			compile it.
	 *
	 *	Returns indicators if the instruction was a no-op (squashed)
	 *	or a branch (branch_shadow)
	 */

void compile_inst (uint inst, uint addr, INST *ip,
			int *squashed, int *branch_shadow)
	{
	int	opc1;

	*squashed = 0;
	*branch_shadow = 0;
	opc1 = (inst >> 26) & 0x3f;
	switch (opc1)
		{
		case 0:				/* SPECIAL	*/
			if (inst == 0)
				*squashed = 1;
			*branch_shadow = mc_spec (inst, addr, ip);
			break;
		case 1:				/* REGIMM	*/
			*branch_shadow = mc_regimm (inst, addr, ip);
			break;

		case 2:
			ip->op = OPJ;
			ip->r1 = -1;
			ip->r2 = -1;
			ip->r3 = -1;
			ip->imm = ((addr+4) & 0xf0000000) |
					((inst & 0x3ffffff) << 2);
			*branch_shadow = 1;
			break;

		case 3:
			ip->op = OPJAL;
			ip->r1 = LP;
			ip->r2 = -1;
			ip->r3 = -1;
			ip->imm = ((addr+4) & 0xf0000000) |
					((inst & 0x3ffffff) << 2);
			*branch_shadow = 1;
			break;

		case 4:
			ip->op = OPBEQ;
			ip->r1 = -1;
			ip->r2 = RTRI(inst);
			ip->r3 = RSRI(inst);
			ip->imm = (inst & 0x00008000 ?
		    (addr+4) - ((0x10000 - (inst&0x0000ffff)) << 2) :
				(addr+4) + ((inst&0x0000ffff) << 2) );
			*branch_shadow = 1;
			break;

		case 5:
			ip->op = OPBNE;
			ip->r1 = -1;
			ip->r2 = RTRI(inst);
			ip->r3 = RSRI(inst);
			ip->imm = (inst & 0x00008000 ?
		    (addr+4) - ((0x10000 - (inst&0x0000ffff)) << 2) :
				(addr+4) + ((inst&0x0000ffff) << 2) );
			*branch_shadow = 1;
			break;

		case 6:				/* BLEZ		*/
			ip->op = OPBLEZ;
			ip->r1 = -1;
			ip->r2 = RSRI(inst);
			ip->r3 = -1;
			ip->imm = (inst & 0x00008000 ?
		    (addr+4) - ((0x10000 - (inst&0x0000ffff)) << 2) :
				(addr+4) + ((inst&0x0000ffff) << 2) );
			*branch_shadow = 1;
			break;

		case 7:				/* BGTZ		*/
			ip->op = OPBGTZ;
			ip->r1 = -1;
			ip->r2 = RSRI(inst);
			ip->r3 = -1;
			ip->imm = (inst & 0x00008000 ?
		    (addr+4) - ((0x10000 - (inst&0x0000ffff)) << 2) :
				(addr+4) + ((inst&0x0000ffff) << 2) );
			*branch_shadow = 1;
			break;

		case 8:
			ip->op = OPADDI;
			ip->r1 = RTRI(inst);
			ip->r2 = RSRI(inst);
			ip->r3 = -1;
			ip->imm = IMMED(inst);
			break;

		case 9:
			ip->op = OPADDIU;
			ip->r1 = RTRI(inst);
			ip->r2 = RSRI(inst);
			ip->r3 = -1;
			ip->imm = IMMED(inst);
			break;

		case 10:
			ip->op = OPSLTI;
			ip->r1 = RTRI(inst);
			ip->r2 = RSRI(inst);
			ip->r3 = -1;
			ip->imm = IMMED(inst);
			break;

		case 11:
			ip->op = OPSLTIU;
			ip->r1 = RTRI(inst);
			ip->r2 = RSRI(inst);
			ip->r3 = -1;
			ip->imm = IMMED(inst);
			break;

		case 12:
			ip->op = OPANDI;
			ip->r1 = RTRI(inst);
			ip->r2 = RSRI(inst);
			ip->r3 = -1;
			ip->imm = inst & 0x0000ffff;
			break;

		case 13:
			ip->op = OPORI;
			ip->r1 = RTRI(inst);
			ip->r2 = RSRI(inst);
			ip->r3 = -1;
			ip->imm = inst & 0x0000ffff;
			break;

		case 14:
			ip->op = OPXORI;
			ip->r1 = RTRI(inst);
			ip->r2 = RSRI(inst);
			ip->r3 = -1;
			ip->imm = inst & 0x0000ffff;
			break;

		case 15:			/* LUI -> ORI	*/
			ip->op = OPORI;
			ip->r1 = RTRI(inst);
			ip->r2 = 0;
			ip->r3 = -1;
			ip->imm = inst << 16;
			break;

		case 16:			/* COP0		*/
			mc_cop0(inst, addr, ip);
			break;

		case 17:			/* COP1		*/
			*branch_shadow = mc_cop1 (inst, addr, ip);
			break;

		case 18:			/* COP2		*/
#ifdef PRINT_WARN
			fprintf (stderr,
				"Illegal instruction (COP2)\n");
#endif
			ip->r1 = ip->r2 = ip->r3 = -1;
			ip->op = OPILL;
			break;

		case 19:			/* COP3		*/
#ifdef PRINT_WARN
			fprintf (stderr,
				"Illegal instruction (COP3)\n");
#endif
			ip->r1 = ip->r2 = ip->r3 = -1;
			ip->op = OPILL;
			break;

		case 20:			/* BEQL		*/
			ip->op = OPBEQL;
			ip->r1 = -1;
			ip->r2 = RTRI(inst);
			ip->r3 = RSRI(inst);
			ip->imm = (inst & 0x00008000 ?
		    (addr+4) - ((0x10000 - (inst&0x0000ffff)) << 2) :
				(addr+4) + ((inst&0x0000ffff) << 2) );
			*branch_shadow = 1;
			break;

		case 21:			/* BNEL		*/
			ip->op = OPBNEL;
			ip->r1 = -1;
			ip->r2 = RTRI(inst);
			ip->r3 = RSRI(inst);
			ip->imm = (inst & 0x00008000 ?
		    (addr+4) - ((0x10000 - (inst&0x0000ffff)) << 2) :
				(addr+4) + ((inst&0x0000ffff) << 2) );
			*branch_shadow = 1;
			break;

		case 22:			/* BLEZL	*/
			ip->op = OPBLEZL;
			ip->r1 = -1;
			ip->r2 = RSRI(inst);
			ip->r3 = -1;
			ip->imm = (inst & 0x00008000 ?
		    (addr+4) - ((0x10000 - (inst&0x0000ffff)) << 2) :
				(addr+4) + ((inst&0x0000ffff) << 2) );
			*branch_shadow = 1;
			break;

		case 23:			/* BGTZL	*/
			ip->op = OPBGTZL;
			ip->r1 = -1;
			ip->r2 = RSRI(inst);
			ip->r3 = -1;
			ip->imm = (inst & 0x00008000 ?
		    (addr+4) - ((0x10000 - (inst&0x0000ffff)) << 2) :
				(addr+4) + ((inst&0x0000ffff) << 2) );
			*branch_shadow = 1;
			break;

		/* 24-31 unused */

		case 32:
			ip->op = OPLB;
			ip->r1 = RTRI(inst);
			ip->r2 = RSRI(inst);
			ip->r3 = -1;
			ip->imm = IMMED(inst);
            break;

		case 33:
			ip->op = OPLH;
			ip->r1 = RTRI(inst);
			ip->r2 = RSRI(inst);
			ip->r3 = -1;
			ip->imm = IMMED(inst);
			break;

		case 34:
			ip->op = OPLWL;
			ip->r1 = RTRI(inst);
			ip->r2 = RSRI(inst);
			ip->r3 = RTRI(inst);
			ip->imm = IMMED(inst);
			break;

		case 35:
			ip->op = OPLW;
			ip->r1 = RTRI(inst);
			ip->r2 = RSRI(inst);
			ip->r3 = -1;
			ip->imm = IMMED(inst);
			if ((ip->r2 == 0) && (ip->r1 == 2) && (ip->imm == 3)) {
			   /* in SimOS direct exeception mode - lw v0,3($0) is a 
			    * system call */
			   ip->op = OPSYSCALL;
			   ip->r1 = -1;
			   ip->r2 = -1;
			   ip->r3 = -1;
			   ip->imm = 0;
			}
			break;

		case 36:
			ip->op = OPLBU;
			ip->r1 = RTRI(inst);
			ip->r2 = RSRI(inst);
			ip->r3 = -1;
			ip->imm = IMMED(inst);
			break;

		case 37:
			ip->op = OPLHU;
			ip->r1 = RTRI(inst);
			ip->r2 = RSRI(inst);
			ip->r3 = -1;
			ip->imm = IMMED(inst);
			break;

		case 38:
			ip->op = OPLWR;
			ip->r1 = RTRI(inst);
			ip->r2 = RSRI(inst);
			ip->r3 = RTRI(inst);
			ip->imm = IMMED(inst);
			break;

		/* 39 unused */

		case 40:
			ip->op = OPSB;
			ip->r1 = -1;
			ip->r2 = RSRI(inst);
			ip->r3 = RTRI(inst);
			ip->imm = IMMED(inst);
			break;

		case 41:
			ip->op = OPSH;
			ip->r1 = -1;
			ip->r2 = RSRI(inst);
			ip->r3 = RTRI(inst);
			ip->imm = IMMED(inst);
			break;

		case 42:
			ip->op = OPSWL;
			ip->r1 = -1;
			ip->r2 = RSRI(inst);
			ip->r3 = RTRI(inst);
			ip->imm = IMMED(inst);
			break;

		case 43:
			ip->op = OPSW;
			ip->r1 = -1;
			ip->r2 = RSRI(inst);
			ip->r3 = RTRI(inst);
			ip->imm = IMMED(inst);
			break;

		/* 44, 45 unused */

		case 46:
			ip->op = OPSWR;
			ip->r1 = -1;
			ip->r2 = RSRI(inst);
			ip->r3 = RTRI(inst);
			ip->imm = IMMED(inst);
			break;

		case 47:			/* CACHE	*/
			ip->op = OPCP0;
                        ip->r1 = -1;
                        ip->r2 = RSRI(inst);
                        ip->r3 = -1;
                        ip->imm = inst;
			break;

		case 48:			/* LL		*/
			ip->op = OPLL;
			ip->r1 = RTRI(inst);
			ip->r2 = RSRI(inst);
			ip->r3 = -1;
			ip->imm = IMMED(inst);
			break;

		case 49:
			ip->op = OPLWC1;
			ip->r1 = FTRI(inst);
			ip->r2 = RSRI(inst);
			ip->r3 = -1;
			ip->imm = IMMED(inst);
			break;

		case 50:			/* LWC2		*/
#ifdef PRINT_WARN
			fprintf (stderr,
				"LWC2 instruction not supported\n");
			fprintf (stderr, "Location: 0x%8.8x\n", addr);
#endif
			ip->r1 = ip->r2 = ip->r3 = -1;
			ip->op = OPILL;
			break;


		case 51:			/* PREF		*/
            ip->op = OPPREF;
            /* RPB hack -- make hint field negative so it isn't treated as a register */
            ip->r1 = -RTRI(inst);
            ip->r2 = RSRI(inst);
            ip->r3 = -1;
            ip->imm = IMMED(inst);
			break;

		/* 52 unused */

		case 53:
			ip->op = OPLDC1;
			ip->r1 = DTRI(inst);
			ip->r2 = RSRI(inst);
			ip->r3 = -1;
			ip->imm = IMMED(inst);
			break;
		case 55:			/* LD		*/
#ifdef PRINT_WARN
                        fprintf (stderr, "Hacked LD instruction found\n");
#endif
		case 54:			/* LDC2		*/
                   /* This is not a real LD is the special hacked LD
                    * that allows us to commication with flashlite from
                    * the 32bit CPU model we are currently using 
                    */
			ip->r1 = ip->r2 = ip->r3 = -1;
			ip->op = OPLDHACK;
                        ip->imm = inst;
			break;

		case 56:			/* SC		*/
			ip->op = OPSC;
			ip->r1 = RTRI(inst);
			ip->r2 = RSRI(inst);
			ip->r3 = RTRI(inst);
			ip->imm = IMMED(inst);
			break;

		case 57:
			ip->op = OPSWC1;
			ip->r1 = -1;
			ip->r2 = RSRI(inst);
			ip->r3 = FTRI(inst);
			ip->imm = IMMED(inst);
			break;

		case 58:			/* SWC2		*/
#ifdef PRINT_WARN
			fprintf (stderr,
				"SWC2 instruction not supported\n");
			fprintf (stderr, "Location: 0x%8.8x\n", addr);
#endif
			ip->r1 = ip->r2 = ip->r3 = -1;
			ip->op = OPILL;
			break;

		case 59:			/* SWC3		*/
#ifdef PRINT_WARN
			fprintf (stderr,
				"SWC3 instruction not supported\n");
			fprintf (stderr, "Location: 0x%8.8x\n", addr);
#endif
			ip->r1 = ip->r2 = ip->r3 = -1;
			ip->op = OPILL;
			break;

		/* 60 unused */

		case 61:
			ip->op = OPSDC1;
			ip->r1 = -1;
			ip->r2 = RSRI(inst);
			ip->r3 = DTRI(inst);
			ip->imm = IMMED(inst);
			break;
		case 63:			/* SD		*/
#ifdef PRINT_WARN
                        fprintf (stderr, "Hacked LD instruction found\n");
#endif
		case 62:			/* SDC2		*/
                   /* This is not a real SD is the special hacked D
                    * that allows us to commication with flashlite from
                    * the 32bit CPU model we are currently using 
                    */
			ip->r1 = ip->r2 = ip->r3 = -1;
			ip->op = OPSDHACK;
                        ip->imm = inst;
			break;

		default:
#ifdef PRINT_WARN
			fprintf (stderr,
				"Illegal instruction (undefined)\n");
			fprintf (stderr, "Location: 0x%8.8x\n", addr);
#endif
			ip->r1 = ip->r2 = ip->r3 = -1;
			ip->op = OPILL;
			break;
		}
	}


	/*
	 *  mc_spec  -  Compile the SPECIAL opcodes
	 */

int mc_spec (uint inst, uint addr, INST *ip)
	{
	int branch_shadow = 0;

	switch (inst & 0x03f)
		{
		case 0:
			ip->op = OPSLL;
			ip->r1 = RDRI(inst);
			ip->r2 = RTRI(inst);
			ip->r3 = -1;
			ip->imm = SHAMT(inst);
			break;

		/* 1 unused */

		case 2:
			ip->op = OPSRL;
			ip->r1 = RDRI(inst);
			ip->r2 = RTRI(inst);
			ip->r3 = -1;
			ip->imm = SHAMT(inst);
			break;

		case 3:
			ip->op = OPSRA;
			ip->r1 = RDRI(inst);
			ip->r2 = RTRI(inst);
			ip->r3 = -1;
			ip->imm = SHAMT(inst);
			break;

		case 4:
			ip->op = OPSLLV;
			ip->r1 = RDRI(inst);
			ip->r2 = RTRI(inst);
			ip->r3 = RSRI(inst);
			ip->imm = 0;
			break;

		/* 5 unused */

		case 6:
			ip->op = OPSRLV;
			ip->r1 = RDRI(inst);
			ip->r2 = RTRI(inst);
			ip->r3 = RSRI(inst);
			ip->imm = 0;
			break;

		case 7:
			ip->op = OPSRAV;
			ip->r1 = RDRI(inst);
			ip->r2 = RTRI(inst);
			ip->r3 = RSRI(inst);
			ip->imm = 0;
			break;

		case 8:				/* Use JRET opcode to	*/
						/* identify subroutine	*/
						/* returns.		*/
			ip->r1 = -1;
			ip->r2 = RSRI(inst);
			ip->r3 = -1;
			ip->imm = 0;
			ip->op = (ip->r2 == LP ? OPJRET : OPJR);
			branch_shadow = 1;
			break;

		case 9:
			ip->op = OPJALR;
			ip->r1 = RDRI(inst);
			ip->r2 = RSRI(inst);
			ip->r3 = -1;
			ip->imm = 0;
			branch_shadow = 1;
			break;

		/* 10, 11 unused */

		case 12:
			ip->op = OPSYSCALL;
			ip->r1 = -1;
			ip->r2 = -1;
			ip->r3 = -1;
			ip->imm = 0;
			break;

			
		case 13:
			ip->op = OPBREAK;
			ip->r1 = -1;
			ip->r2 = -1;
			ip->r3 = -1;
			ip->imm = 0;
			break;

		/* 14 unused */

		case 15:				/* SYNC		*/
#ifdef PRINT_WARN
			fprintf (stderr, "SYNC instruction not supported\n");
			fprintf (stderr, "Location: 0x%8.8x\n", addr);
#endif
			ip->r1 = ip->r2 = ip->r3 = -1;
			ip->op = OPILL;
			break;

		case 16:
			ip->op = OPCPY;
			ip->r1 = RDRI(inst);
			ip->r2 = HIREG;
			ip->r3 = -1;
			ip->imm = 0;
			break;

		case 17:
			ip->op = OPHCPY;
			ip->r1 = HIREG;
			ip->r2 = RSRI(inst);
			ip->r3 = HILOREG;
			ip->imm = 0;
			break;

		case 18:
			ip->op = OPCPY;
			ip->r1 = RDRI(inst);
			ip->r2 = LOREG;
			ip->r3 = -1;
			ip->imm = 0;
			break;

		case 19:
			ip->op = OPHCPY;
			ip->r1 = LOREG;
			ip->r2 = RSRI(inst);
			ip->r3 = HILOREG;
			ip->imm = 0;
			break;

		/* 20 - 23 unused */

		case 24:
			ip->op = OPMULT;
			ip->r1 = HILOREG;
			ip->r2 = RTRI(inst);
			ip->r3 = RSRI(inst);
			ip->imm = 0;
			break;

		case 25:
			ip->op = OPMULTU;
			ip->r1 = HILOREG;
			ip->r2 = RTRI(inst);
			ip->r3 = RSRI(inst);
			ip->imm = 0;
			break;

		case 26:
			ip->op = OPDIV;
			ip->r1 = HILOREG;
			ip->r2 = RSRI(inst);
			ip->r3 = RTRI(inst);
			ip->imm = 0;
			break;

		case 27:
			ip->op = OPDIVU;
			ip->r1 = HILOREG;
			ip->r2 = RSRI(inst);
			ip->r3 = RTRI(inst);
			ip->imm = 0;
			break;

		/* 28 - 31 unused */

		case 32:
			ip->op = OPADD;
			ip->r1 = RDRI(inst);
			ip->r2 = RSRI(inst);
			ip->r3 = RTRI(inst);
			ip->imm = 0;
			break;

		case 33:
			ip->op = OPADDU;
			ip->r1 = RDRI(inst);
			ip->r2 = RSRI(inst);
			ip->r3 = RTRI(inst);
			ip->imm = 0;
			break;

		case 34:
			ip->op = OPSUB;
			ip->r1 = RDRI(inst);
			ip->r2 = RSRI(inst);
			ip->r3 = RTRI(inst);
			ip->imm = 0;
			break;

		case 35:
			ip->op = OPSUBU;
			ip->r1 = RDRI(inst);
			ip->r2 = RSRI(inst);
			ip->r3 = RTRI(inst);
			ip->imm = 0;
			break;

		case 36:
			ip->op = OPAND;
			ip->r1 = RDRI(inst);
			ip->r2 = RSRI(inst);
			ip->r3 = RTRI(inst);
			ip->imm = 0;
			break;

		case 37:
			ip->op = OPOR;
			ip->r1 = RDRI(inst);
			ip->r2 = RSRI(inst);
			ip->r3 = RTRI(inst);
			ip->imm = 0;
			break;

		case 38:
			ip->op = OPXOR;
			ip->r1 = RDRI(inst);
			ip->r2 = RSRI(inst);
			ip->r3 = RTRI(inst);
			ip->imm = 0;
			break;

		case 39:
			ip->op = OPNOR;
			ip->r1 = RDRI(inst);
			ip->r2 = RSRI(inst);
			ip->r3 = RTRI(inst);
			ip->imm = 0;
			break;

		/* 40, 41 unused */

		case 42:
			ip->op = OPSLT;
			ip->r1 = RDRI(inst);
			ip->r2 = RSRI(inst);
			ip->r3 = RTRI(inst);
			ip->imm = 0;
			break;

		case 43:
			ip->op = OPSLTU;
			ip->r1 = RDRI(inst);
			ip->r2 = RSRI(inst);
			ip->r3 = RTRI(inst);
			ip->imm = 0;
			break;

		/* 44 - 47 unused */

		case 48:				/* TGE		*/
		case 49:				/* TGEU		*/
		case 50:				/* TLT		*/
		case 51:				/* TLTU		*/
		case 52:				/* TEQ		*/
		case 54:				/* TNE		*/
#ifdef PRINT_WARN
			fprintf (stderr, "Trap instructions not supported\n");
			fprintf (stderr, "Location: 0x%8.8x\n", addr);
#endif
			ip->r1 = ip->r2 = ip->r3 = -1;
			ip->op = OPILL;
			break;

		/* 53, 55 - 63 unused */

		default:
#ifdef PRINT_WARN
			fprintf (stderr, "Illegal instruction (undefined)\n");
			fprintf (stderr, "Location: 0x%8.8x\n", addr);
#endif
			ip->r1 = ip->r2 = ip->r3 = -1;
			ip->op = OPILL;
			break;
		}

	return (branch_shadow);
	}


	/*
	 *  mc_regimm  -  Handle the register immediate opcodes
	 */

int mc_regimm (uint inst, uint addr, INST *ip)
	{
	int branch_shadow = 0;

	switch ((inst>>16) & 0x1f)
		{
		case 0:					/* BLTZ		*/
			ip->op = OPBLTZ;
			ip->r1 = -1;
			ip->r2 = RSRI(inst);
			ip->r3 = -1;
			ip->imm = (inst & 0x00008000 ?
			    (addr+4) - ((0x10000 - (inst&0x0000ffff)) << 2) :
					(addr+4) + ((inst&0x0000ffff) << 2) );
			branch_shadow = 1;
			break;

		case 1:					/* BGEZ		*/
			ip->op = OPBGEZ;
			ip->r1 = -1;
			ip->r2 = RSRI(inst);
			ip->r3 = -1;
			ip->imm = (inst & 0x00008000 ?
			    (addr+4) - ((0x10000 - (inst&0x0000ffff)) << 2) :
					(addr+4) + ((inst&0x0000ffff) << 2) );
			branch_shadow = 1;
			break;

		case 2:					/* BLTZL	*/
			ip->op = OPBLTZL;
			ip->r1 = -1;
			ip->r2 = RSRI(inst);
			ip->r3 = -1;
			ip->imm = (inst & 0x00008000 ?
			    (addr+4) - ((0x10000 - (inst&0x0000ffff)) << 2) :
					(addr+4) + ((inst&0x0000ffff) << 2) );
			branch_shadow = 1;
			break;

		case 3:					/* BGEZL	*/
			ip->op = OPBGEZL;
			ip->r1 = -1;
			ip->r2 = RSRI(inst);
			ip->r3 = -1;
			ip->imm = (inst & 0x00008000 ?
			    (addr+4) - ((0x10000 - (inst&0x0000ffff)) << 2) :
					(addr+4) + ((inst&0x0000ffff) << 2) );
			branch_shadow = 1;
			break;

		/* 4-7 unused */

		case  8:				/* TGEI		*/
		case  9:				/* TGEIU	*/
		case 10:				/* TLTI		*/
		case 11:				/* TLTIU	*/
		case 12:				/* TEQI		*/
		case 14:				/* TNEI		*/
#ifdef PRINT_WARN
			fprintf (stderr, "Trap instructions not supported\n");
			fprintf (stderr, "Location: 0x%8.8x\n", addr);
#endif
			ip->r1 = ip->r2 = ip->r3 = -1;
			ip->op = OPILL;
			break;

		/* 13, 15 unused */

		case 16:				/* BLTZAL	*/
			ip->op = OPBLTZAL;
			ip->r1 = LP;
			ip->r2 = RSRI(inst);
			ip->r3 = -1;
			ip->imm = (inst & 0x00008000 ?
			    (addr+4) - ((0x10000 - (inst&0x0000ffff)) << 2) :
				(addr+4) + ((inst&0x0000ffff) << 2) );
			branch_shadow = 1;
			break;

		case 17:				/* BGEZAL	*/
			ip->op = OPBGEZAL;
			ip->r1 = LP;
			ip->r2 = RSRI(inst);
			ip->r3 = -1;
			ip->imm = (inst & 0x00008000 ?
			    (addr+4) - ((0x10000 - (inst&0x0000ffff)) << 2) :
				(addr+4) + ((inst&0x0000ffff) << 2) );
			branch_shadow = 1;
			break;

		case 18:				/* BLTZALL	*/
			ip->op = OPBLTZALL;
			ip->r1 = LP;
			ip->r2 = RSRI(inst);
			ip->r3 = -1;
			ip->imm = (inst & 0x00008000 ?
			    (addr+4) - ((0x10000 - (inst&0x0000ffff)) << 2) :
				(addr+4) + ((inst&0x0000ffff) << 2) );
			branch_shadow = 1;
			break;

		case 19:				/* BGEZALL	*/
			ip->op = OPBGEZALL;
			ip->r1 = LP;
			ip->r2 = RSRI(inst);
			ip->r3 = -1;
			ip->imm = (inst & 0x00008000 ?
			    (addr+4) - ((0x10000 - (inst&0x0000ffff)) << 2) :
				(addr+4) + ((inst&0x0000ffff) << 2) );
			branch_shadow = 1;
			break;

		/* 20 - 31 unused */

		default:
#ifdef PRINT_WARN
			fprintf (stderr, "Illegal instruction (undefined)\n");
			fprintf (stderr, "Location: 0x%8.8x\n", addr);
#endif
			ip->r1 = ip->r2 = ip->r3 = -1;
			ip->op = OPILL;
			break;
		}

	return (branch_shadow);
	}


	/*
	 *  mc_cop1  -  Handle the coprocessor 1 (floating point) opcodes
	 */

int mc_cop1 (uint inst, uint addr, INST *ip)
	{
	int branch_shadow = 0;

	switch ((inst >> 21) & 0x1f)		/* Decode the fmt field	*/
		{
		case 0:					/* MFC1		*/
			ip->op = OPCPYC1;
			ip->r1 = RTRI(inst);
			ip->r2 = FSRI(inst);
			ip->r3 = -1;
			ip->imm = 0;
			break;

		/* 1 unused */

		case 2:					/* CFC1		*/
			ip->op = OPCPYC1;
			ip->r1 = RTRI(inst);
			ip->r2 = ((inst>>11)&0x01) + FPCTL;
			ip->r3 = -1;
			ip->imm = 0;
			if ((inst>>11)&0x01)
				{
				ip->op = CTL_CFC;
				ip->r3 = CNDREG;
				}
			break;

		/* 3 unused */

		case 4:					/* MTC1		*/
			ip->op = OPHCPYC1;
			ip->r1 = FSRI(inst);
			ip->r2 = RTRI(inst);
			ip->r3 = DSRI(inst);
			ip->imm = 0;
			break;

		/* 5 unused */

		case 6:					/* CTC1		*/
			ip->r1 = ((inst>>11)&0x01) + FPCTL;
			ip->r2 = RTRI(inst);
			ip->r3 = FPCTL;
			ip->imm = 0;

			ip->op = (ip->r1 == FPCTL + 1 ? CTL_FPC : OPHCPYC1);
			break;


		/* 7 unused */

		case 8:				/* Conditional branches	*/
			switch ((inst>>16) & 0x1f)
				{
				case 0:
					ip->op = OPBC1F;
					ip->r1 = -1;
					ip->r2 = CNDREG;
					ip->r3 = -1;
					ip->imm = (inst & 0x00008000 ?
			    (addr+4) - ((0x10000 - (inst&0x0000ffff)) << 2) :
					(addr+4) + ((inst&0x0000ffff) << 2) );
					branch_shadow = 1;
					break;

				case 1:
					ip->op = OPBC1T;
					ip->r1 = -1;
					ip->r2 = CNDREG;
					ip->r3 = -1;
					ip->imm = (inst & 0x00008000 ?
			    (addr+4) - ((0x10000 - (inst&0x0000ffff)) << 2) :
					(addr+4) + ((inst&0x0000ffff) << 2) );
					branch_shadow = 1;
					break;

				case 2:			/* BC1FL	*/
					ip->op = OPBC1FL;
					ip->r1 = -1;
					ip->r2 = CNDREG;
					ip->r3 = -1;
					ip->imm = (inst & 0x00008000 ?
			    (addr+4) - ((0x10000 - (inst&0x0000ffff)) << 2) :
					(addr+4) + ((inst&0x0000ffff) << 2) );
					branch_shadow = 1;
					break;

				case 3:			/* BC1TL	*/
					ip->op = OPBC1TL;
					ip->r1 = -1;
					ip->r2 = CNDREG;
					ip->r3 = -1;
					ip->imm = (inst & 0x00008000 ?
			    (addr+4) - ((0x10000 - (inst&0x0000ffff)) << 2) :
					(addr+4) + ((inst&0x0000ffff) << 2) );
					branch_shadow = 1;
					break;

				/* 4 - 31 unused */

				default:
#ifdef PRINT_WARN
					fprintf (stderr,
					"Illegal instruction (undefined)\n");
					fprintf (stderr,
						"Location: 0x%8.8x\n", addr);
#endif
					ip->op = OPILL;
					ip->r1 = ip->r2 = ip->r3 = -1;
					break;
				}
			break;

		/* 9-15 unused */

		case 16:			/* Single precision	*/
			mc_fp_s (inst, addr, ip);
			break;

		case 17:			/* Double precision	*/
			mc_fp_d (inst, addr, ip);
			break;

		/* 18, 19 unused */

		case 20:			/* Fixed point		*/
			switch (inst & 0x3f)
				{
				/* 0 - 31 unused */

				case 32:		/* CVT.S.W	*/
					ip->op = OPCVTSW;
					ip->r1 = FDRI(inst);
					ip->r2 = FSRI(inst);
					ip->r3 = DDRI(inst);
					ip->imm = 0;
					break;

				case 33:		/* CVT.D.W	*/
					ip->op = OPCVTDW;
					ip->r1 = DDRI(inst);
					ip->r2 = FSRI(inst);
					ip->r3 = -1;
					ip->imm = 0;
					break;

				/* 34 - 63 unused */

				default:
#ifdef PRINT_WARN
					fprintf (stderr,
					"Illegal instruction (undefined)\n");
					fprintf (stderr,
						"Location: 0x%8.8x\n", addr);
#endif
					ip->op = OPILL;
					ip->r1 = ip->r2 = ip->r3 = -1;
					break;
				}
			break;

		/* 21 - 31 unused */

		default:
#ifdef PRINT_WARN
			fprintf (stderr, "Illegal instruction (undefined)\n");
			fprintf (stderr, "Location: 0x%8.8x\n", addr);
#endif
			ip->op = OPILL;
			ip->r1 = ip->r2 = ip->r3 = -1;
			break;
		}

	return (branch_shadow);
	}


	/*
	 *  mc_fp_s  -  Single precision floating point operations
	 */

void mc_fp_s (uint inst, uint addr, INST *ip)
	{
	switch (inst & 0x3f)
		{
		case 0:
			ip->op = OPFADDS;
			ip->r1 = FDRI(inst);
			ip->r2 = FSRI(inst);
			ip->r3 = FTRI(inst);
			ip->imm = 0;
			break;

		case 1:
			ip->op = OPFSUBS;
			ip->r1 = FDRI(inst);
			ip->r2 = FSRI(inst);
			ip->r3 = FTRI(inst);
			ip->imm = 0;
			break;

		case 2:
			ip->op = OPFMULS;
			ip->r1 = FDRI(inst);
			ip->r2 = FSRI(inst);
			ip->r3 = FTRI(inst);
			ip->imm = 0;
			break;

		case 3:
			ip->op = OPFDIVS;
			ip->r1 = FDRI(inst);
			ip->r2 = FSRI(inst);
			ip->r3 = FTRI(inst);
			ip->imm = 0;
			break;

		case 4:
			ip->op = OPFSQRTS;
			ip->r1 = FDRI(inst);
			ip->r2 = FSRI(inst);
			ip->r3 = -1;
			ip->imm = 0;
			break;

		case 5:
			ip->op = OPFABSS;
			ip->r1 = FDRI(inst);
			ip->r2 = FSRI(inst);
			ip->r3 = -1;
			ip->imm = 0;
			break;

		case 6:
			ip->op = OPCPYC1;
			ip->r1 = FDRI(inst);
			ip->r2 = FSRI(inst);
			ip->r3 = -1;
			ip->imm = 0;
			break;

		case 7:
			ip->op = OPFNEGS;
			ip->r1 = FDRI(inst);
			ip->r2 = FSRI(inst);
			ip->r3 = -1;
			ip->imm = 0;
			break;

		/* 8 - 11 unused */

		case 12:
			ip->op = OPFROUNDS;
			ip->r1 = FDRI(inst);
			ip->r2 = FSRI(inst);
			ip->r3 = -1;
			ip->imm = 0;
			break;

		case 13:
			ip->op = OPFTRUNCS;
			ip->r1 = FDRI(inst);
			ip->r2 = FSRI(inst);
			ip->r3 = -1;
			ip->imm = 0;
			break;

		case 14:
			ip->op = OPFCEILS;
			ip->r1 = FDRI(inst);
			ip->r2 = FSRI(inst);
			ip->r3 = -1;
			ip->imm = 0;
			break;

		case 15:
			ip->op = OPFFLOORS;
			ip->r1 = FDRI(inst);
			ip->r2 = FSRI(inst);
			ip->r3 = -1;
			ip->imm = 0;
			break;

		/* 16 - 32 unused */

		case 33:		/* CVT.D.S	*/
			ip->op = OPCVTDS;
			ip->r1 = DDRI(inst);
			ip->r2 = FSRI(inst);
			ip->r3 = -1;
			ip->imm = 0;
			break;

		/* 34 - 35 unused */

		case 36:		/* CVT.W.S	*/
			ip->op = OPCVTWS;
			ip->r1 = FDRI(inst);
			ip->r2 = FSRI(inst);
			ip->r3 = DDRI(inst);
			ip->imm = 0;
			break;

		/* 37 - 47 unused */

		case 48:		/* C.F.S	*/
			ip->op = OPCFS;
			ip->r1 = CNDREG;
			ip->r2 = -1;
			ip->r3 = -1;
			ip->imm = 0;
			break;

		case 49:		/* C.UN.S	*/
			ip->op = OPCUNS;
			ip->r1 = CNDREG;
			ip->r2 = -1;
			ip->r3 = -1;
			ip->imm = 0;
			break;

		case 50:		/* C.EQ.S	*/
			ip->op = OPCEQS;
			ip->r1 = CNDREG;
			ip->r2 = FSRI(inst);
			ip->r3 = FTRI(inst);
			ip->imm = 0;
			break;

		case 51:		/* C.UEQ.S	*/
			ip->op = OPCUEQS;
			ip->r1 = CNDREG;
			ip->r2 = FSRI(inst);
			ip->r3 = FTRI(inst);
			ip->imm = 0;
			break;

		case 52:		/* C.OLT.S	*/
			ip->op = OPCOLTS;
			ip->r1 = CNDREG;
			ip->r2 = FSRI(inst);
			ip->r3 = FTRI(inst);
			ip->imm = 0;
			break;

		case 53:		/* C.ULT.S	*/
			ip->op = OPCULTS;
			ip->r1 = CNDREG;
			ip->r2 = FSRI(inst);
			ip->r3 = FTRI(inst);
			ip->imm = 0;
			break;

		case 54:		/* C.OLE.S	*/
			ip->op = OPCOLES;
			ip->r1 = CNDREG;
			ip->r2 = FSRI(inst);
			ip->r3 = FTRI(inst);
			ip->imm = 0;
			break;

		case 55:		/* C.ULE.S	*/
			ip->op = OPCULES;
			ip->r1 = CNDREG;
			ip->r2 = FSRI(inst);
			ip->r3 = FTRI(inst);
			ip->imm = 0;
			break;

		case 56:		/* C.SF.S	*/
			ip->op = OPCSFS;
			ip->r1 = CNDREG;
			ip->r2 = -1;
			ip->r3 = -1;
			ip->imm = 0;
			break;

		case 57:		/* C.NGLE.S	*/
			ip->op = OPCNGLES;
			ip->r1 = CNDREG;
			ip->r2 = -1;
			ip->r3 = -1;
			ip->imm = 0;
			break;

		case 58:		/* C.SEQ.S	*/
			ip->op = OPCSEQS;
			ip->r1 = CNDREG;
			ip->r2 = FSRI(inst);
			ip->r3 = FTRI(inst);
			ip->imm = 0;
			break;

		case 59:		/* C.NGL.S	*/
			ip->op = OPCNGLS;
			ip->r1 = CNDREG;
			ip->r2 = FSRI(inst);
			ip->r3 = FTRI(inst);
			ip->imm = 0;
			break;

		case 60:		/* C.LT.S	*/
			ip->op = OPCLTS;
			ip->r1 = CNDREG;
			ip->r2 = FSRI(inst);
			ip->r3 = FTRI(inst);
			ip->imm = 0;
			break;

		case 61:		/* C.NGE.S	*/
			ip->op = OPCNGES;
			ip->r1 = CNDREG;
			ip->r2 = FSRI(inst);
			ip->r3 = FTRI(inst);
			ip->imm = 0;
			break;

		case 62:		/* C.LE.S	*/
			ip->op = OPCLES;
			ip->r1 = CNDREG;
			ip->r2 = FSRI(inst);
			ip->r3 = FTRI(inst);
			ip->imm = 0;
			break;

		case 63:		/* C.NGT.S	*/
			ip->op = OPCNGTS;
			ip->r1 = CNDREG;
			ip->r2 = FSRI(inst);
			ip->r3 = FTRI(inst);
			ip->imm = 0;
			break;

		default:
#ifdef PRINT_WARN
			fprintf (stderr, "Illegal instruction (undefined)\n");
			fprintf (stderr, "Location: 0x%8.8x\n", addr);
#endif
			ip->r1 = ip->r2 = ip->r3 = -1;
			ip->op = OPILL;
			break;
		}
	}


	/*
	 *  mc_fp_d  -  Double precision floating point operations
	 */

void mc_fp_d (uint inst, uint addr, INST *ip)
	{
	switch (inst & 0x3f)
		{
		case 0:
			ip->op = OPFADDD;
			ip->r1 = DDRI(inst);
			ip->r2 = DSRI(inst);
			ip->r3 = DTRI(inst);
			ip->imm = 0;
			break;

		case 1:
			ip->op = OPFSUBD;
			ip->r1 = DDRI(inst);
			ip->r2 = DSRI(inst);
			ip->r3 = DTRI(inst);
			ip->imm = 0;
			break;

		case 2:
			ip->op = OPFMULD;
			ip->r1 = DDRI(inst);
			ip->r2 = DSRI(inst);
			ip->r3 = DTRI(inst);
			ip->imm = 0;
			break;

		case 3:
			ip->op = OPFDIVD;
			ip->r1 = DDRI(inst);
			ip->r2 = DSRI(inst);
			ip->r3 = DTRI(inst);
			ip->imm = 0;
			break;

		case 4:
			ip->op = OPFSQRTD;
			ip->r1 = DDRI(inst);
			ip->r2 = DSRI(inst);
			ip->r3 = -1;
			ip->imm = 0;
			break;

		case 5:
			ip->op = OPFABSD;
			ip->r1 = DDRI(inst);
			ip->r2 = DSRI(inst);
			ip->r3 = -1;
			ip->imm = 0;
			break;

		case 6:
			ip->op = OPFMOVD;
			ip->r1 = DDRI(inst);
			ip->r2 = DSRI(inst);
			ip->r3 = -1;
			ip->imm = 0;
			break;

		case 7:
			ip->op = OPFNEGD;
			ip->r1 = DDRI(inst);
			ip->r2 = DSRI(inst);
			ip->r3 = -1;
			ip->imm = 0;
			break;

		/* 8 - 11 unused */

		case 12:
			ip->op = OPFROUNDD;
			ip->r1 = DDRI(inst);
			ip->r2 = DSRI(inst);
			ip->r3 = -1;
			ip->imm = 0;
			break;

		case 13:
			ip->op = OPFTRUNCD;
			ip->r1 = DDRI(inst);
			ip->r2 = DSRI(inst);
			ip->r3 = -1;
			ip->imm = 0;
			break;

		case 14:
			ip->op = OPFCEILD;
			ip->r1 = DDRI(inst);
			ip->r2 = DSRI(inst);
			ip->r3 = -1;
			ip->imm = 0;
			break;

		case 15:
			ip->op = OPFFLOORD;
			ip->r1 = DDRI(inst);
			ip->r2 = DSRI(inst);
			ip->r3 = -1;
			ip->imm = 0;
			break;

		/* 16 - 31 unused */

		case 32:		/* CVT.S.D	*/
			ip->op = OPCVTSD;
			ip->r1 = FDRI(inst);
			ip->r2 = DSRI(inst);
			ip->r3 = DDRI(inst);
			ip->imm = 0;
			break;

		/* 33 - 35 unused */

		case 36:		/* CVT.W.D	*/
			ip->op = OPCVTWD;
			ip->r1 = FDRI(inst);
			ip->r2 = DSRI(inst);
			ip->r3 = DDRI(inst);
			ip->imm = 0;
			break;

		/* 37 - 47 unused */

		case 48:		/* C.F.D	*/
			ip->op = OPCFS;
			ip->r1 = CNDREG;
			ip->r2 = -1;
			ip->r3 = -1;
			ip->imm = 0;
			break;

		case 49:		/* C.UN.D	*/
			ip->op = OPCUNS;
			ip->r1 = CNDREG;
			ip->r2 = -1;
			ip->r3 = -1;
			ip->imm = 0;
			break;

		case 50:		/* C.EQ.D	*/
			ip->op = OPCEQD;
			ip->r1 = CNDREG;
			ip->r2 = DSRI(inst);
			ip->r3 = DTRI(inst);
			ip->imm = 0;
			break;

		case 51:		/* C.UEQ.D	*/
			ip->op = OPCUEQD;
			ip->r1 = CNDREG;
			ip->r2 = DSRI(inst);
			ip->r3 = DTRI(inst);
			ip->imm = 0;
			break;

		case 52:		/* C.OLT.D	*/
			ip->op = OPCOLTD;
			ip->r1 = CNDREG;
			ip->r2 = DSRI(inst);
			ip->r3 = DTRI(inst);
			ip->imm = 0;
			break;

		case 53:		/* C.ULT.D	*/
			ip->op = OPCULTD;
			ip->r1 = CNDREG;
			ip->r2 = DSRI(inst);
			ip->r3 = DTRI(inst);
			ip->imm = 0;
			break;

		case 54:		/* C.OLE.D	*/
			ip->op = OPCOLED;
			ip->r1 = CNDREG;
			ip->r2 = DSRI(inst);
			ip->r3 = DTRI(inst);
			ip->imm = 0;
			break;

		case 55:		/* C.ULE.D	*/
			ip->op = OPCULED;
			ip->r1 = CNDREG;
			ip->r2 = DSRI(inst);
			ip->r3 = DTRI(inst);
			ip->imm = 0;
			break;

		case 56:		/* C.SF.D	*/
			ip->op = OPCSFS;
			ip->r1 = CNDREG;
			ip->r2 = -1;
			ip->r3 = -1;
			ip->imm = 0;
			break;

		case 57:		/* C.NGLE.D	*/
			ip->op = OPCNGLES;
			ip->r1 = CNDREG;
			ip->r2 = -1;
			ip->r3 = -1;
			ip->imm = 0;
			break;

		case 58:		/* C.SEQ.D	*/
			ip->op = OPCSEQD;
			ip->r1 = CNDREG;
			ip->r2 = DSRI(inst);
			ip->r3 = DTRI(inst);
			ip->imm = 0;
			break;

		case 59:		/* C.NGL.D	*/
			ip->op = OPCNGLD;
			ip->r1 = CNDREG;
			ip->r2 = DSRI(inst);
			ip->r3 = DTRI(inst);
			ip->imm = 0;
			break;

		case 60:		/* C.LT.D	*/
			ip->op = OPCLTD;
			ip->r1 = CNDREG;
			ip->r2 = DSRI(inst);
			ip->r3 = DTRI(inst);
			ip->imm = 0;
			break;

		case 61:		/* C.NGE.D	*/
			ip->op = OPCNGED;
			ip->r1 = CNDREG;
			ip->r2 = DSRI(inst);
			ip->r3 = DTRI(inst);
			ip->imm = 0;
			break;

		case 62:		/* C.LE.D	*/
			ip->op = OPCLED;
			ip->r1 = CNDREG;
			ip->r2 = DSRI(inst);
			ip->r3 = DTRI(inst);
			ip->imm = 0;
			break;

		case 63:		/* C.NGT.D	*/
			ip->op = OPCNGTD;
			ip->r1 = CNDREG;
			ip->r2 = DSRI(inst);
			ip->r3 = DTRI(inst);
			ip->imm = 0;
			break;

		default:
#ifdef PRINT_WARN
			fprintf (stderr, "Illegal instruction (undefined)\n");
			fprintf (stderr, "Location: 0x%8.8x\n", addr);
#endif
			ip->op = OPILL;
			ip->r1 = ip->r2 = ip->r3 = -1;
			break;
		}
	}



	/*
	 *  mc_cop0  -  Handle the coprocessor 0 
	 */

void mc_cop0 (uint inst, uint addr, INST *ip)
{
       ip->op = OPCP0;
        ip->r1 = -1;
        ip->r2 = -1;
        ip->r3 = -1;
        ip->imm = inst;
	switch ((inst >> 21) & 0x1f) {		
    	case 4:  /* move to C0processor */
	    ip->r2 = RTRI(inst);
            break;
        case 0:   /* move from C0 */
	    ip->r1 = RTRI(inst);
	    ip->op = OPCP0_SPEC;  /* Can speculative execute this one. */
            break;
         default:
           break;
	}
}	      

