/*
 * Copyright (C) 1996-1998 by the Board of Trustees
 *    of Leland Stanford Junior University.
 * 
 * This file is part of the SimOS distribution. 
 * See LICENSE file for terms of the license. 
 *
 */


/**************************************************************************
 * $File$
 * 
 * $Author: nikos $
 * $Date: 1998/02/16 23:20:26 $
 *************************************************************************/


#include <fcntl.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <sys/uio.h>
#include <string.h>
#ifdef sgi
#include <sys/sysmp.h>
#include <netinet/if_ether.h>
#endif
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#ifndef i386
#include <netinet/in.h>
#endif
#include <sys/types.h>
#include <unistd.h>

#include "cpu_interface.h" 
#include "eventcallback.h"
#include "cluster.h"
#include "ethernet.h"
#include "sim_error.h"
#include "simrecord.h"
#include "simutil.h"
#include "arch_specifics.h"
#if 0 
#if defined(SIM_MIPS32) || defined(SIM_MIPS64)
# include "../../cpus/mipsy/mipsy_simos.h"
# include "../../cpus/simos/startup.h"
# include "../../cpus/simos/simmisc.h"   
#endif
#endif
#if defined(SIM_X86)
# include "../../cpus-x86/simos/startup.h"
# include "../../cpus-x86/simos/simmisc.h"   
#endif
#include "params.h"
#include "machine_params.h"


#define CLUSTER_MAX_MSGS 128
typedef int MessageId; 
typedef struct Message {
   MessageId msgId; 
    MessageId next;
    int len;
    int source, dest,if_num;
    int id;
    SimTime sendTime, deliverTime;
    char data[SIMETHER_MAX_TRANSFER_SIZE];
} Message;


/* *********************************************************
 * ClusterArea
 * shared between the different hosts in the cluster simulation
 * !!!! can be mapped at different addresses in the different
 * processes. Do not have pointers in the structure. That
 * is why we use indices throughout the code (unfortunately).
 * *********************************************************/

typedef struct ClusterArea {
   char    barrierArea[256];
   char    commonLock[256];
   int     numInit;
   int     numParticipants;
   MessageId freeList;
   int     numMsgsUsed;
   int     msgCounter;
   int   barrierSync;
   struct HostArea{
      struct { 
         char etheraddr[6];
         char padding[2];
      } etherAddr[ETHER_MAX_CONTROLLERS];
      char privateLock[128];
      SimTime currentBarrier;
      SimTime nextBarrier;
      SimTime close;
      SimTime cycleCount;
      SimTime barrierInterval;
      int cpuClock; 
      int pid;
      int hostCPU;
      MessageId head;
      int     doCheckpoint;
   } simul[32];
   Message msgPool[CLUSTER_MAX_MSGS];
} ClusterArea;


ClusterArea *clusterArea = 0;
struct HostArea *myArea;
static EventCallbackHdr barrierCallback;
static EventCallbackHdr receiveMsgCallback;
static int barrierMyNum;

static char *filename = NULL;
static int barrierP = 1;
static int barrierNumSimul = 0;
static int barrierInterval = 40000;
static int clusterLatency = 100;

static int slaveCheckpoint = 0;

#ifdef notdef
#define LOCK_CLUSTER()   (Simcp0Lock  ((int*)clusterArea->commonLock))
#define UNLOCK_CLUSTER() (Simcp0Unlock((int*)clusterArea->commonLock))
#define LOCK_HOST(_id)   (Simcp0Lock  ((int*)clusterArea->simul[_id].privateLock))
#define UNLOCK_HOST(_id) (Simcp0Unlock((int*)clusterArea->simul[_id].privateLock))
#else
#define LOCK_CLUSTER()   
#define UNLOCK_CLUSTER() 
#define LOCK_HOST(_id)   
#define UNLOCK_HOST(_id) 
#endif

static void     ClusterFreeMsg(Message *msg);
static Message* GetMsg(SimTime now) ;
#ifdef sgi
static void     SendMsg(int cpuNum,Message *msg, int receiver);
static Message *ClusterAllocMsg(void);
#endif




void DeliverMsgCallback(int cpuNum,EventCallbackHdr *hdr,void *arg)
{
   Message *msg = (Message *) arg;
   
   LogEntry("CLUSTER-RCV",cpuNum,"sender=%i id=%i latency=%10lld len=%i\n",
            msg->source,msg->id,(uint64)CPUVec.CycleCount(cpuNum)-msg->sendTime,
            msg->len);
   ASSERT( PTR_TO_UINT64(msg) < 
           PTR_TO_UINT64(&clusterArea->msgPool[clusterArea->numMsgsUsed]));
   
   SimetherReceivePacket(msg->if_num, msg->data,msg->len);
   ClusterFreeMsg(msg);
   msg = GetMsg( CPUVec.CycleCount(cpuNum));
   if (msg) { 
      if (msg->deliverTime <  CPUVec.CycleCount(cpuNum) 
          && (clusterLatency  > myArea->barrierInterval)) { 
         CPUWarning("Cluster: delivering msg at %lld when received at %lld  lat=%lld\n",
                    (uint64)CPUVec.CycleCount(cpuNum), msg->deliverTime,
                    (uint64) CPUVec.CycleCount(cpuNum) - msg->deliverTime);
         EventDoCallback(cpuNum,DeliverMsgCallback,&receiveMsgCallback,
                         (void*)msg, 0);
      } else { 
         EventDoCallback(cpuNum,DeliverMsgCallback,&receiveMsgCallback,
                         (void*)msg,
                         msg->deliverTime -  CPUVec.CycleCount(cpuNum));
      }
   }
}


    
/* This is called at regular intervalls. The simulation completes when
 * one simulator sets syncArea->close to its nextBarrier field.
 * At this time, the requesting process is already draining its remaining
 * events. All other ones are still in their normal processing. 
 */

void BarrierCallback(int cpuNum,EventCallbackHdr *hdr,void *arg)
{
   ASSERT( hdr ==  &barrierCallback );
   ASSERT( cpuNum == 0 );
   ASSERT( myArea->nextBarrier <= CPUVec.CycleCount(cpuNum) );
   if( myArea->close && myArea->close < myArea->nextBarrier ) {
      /* someone else has already requested the last barrier. Time
       * to leave
       */
      CPUWarning("%lld Passed final barrier already! close=%lld\n",
               (uint64)CPUVec.CycleCount(cpuNum),
               (uint64)myArea->close);
      return; 
   }

   if( myArea->close ) 
      CPUPrint("Barrier synching at %lld close=%lld\n",
               (uint64)CPUVec.CycleCount(cpuNum),
               (uint64)myArea->close);

#ifdef notdef
   /*
    * Barrier 
    */

   if( clusterArea->barrierSync ) { 
      Simcp0Asm_Barrier((int *)&clusterArea->barrierArea[0],barrierNumSimul);
   }
#endif
   myArea->currentBarrier = myArea->nextBarrier; 
   myArea->nextBarrier  +=  myArea->barrierInterval;
   myArea->cycleCount =  CPUVec.CycleCount(cpuNum);

#ifdef notdef
   /* 2nd barrier: needed or danger of a race condition */
   if( clusterArea->barrierSync ) { 
      Simcp0Asm_Barrier((int *)&clusterArea->barrierArea[0],barrierNumSimul);
   }
#endif
   if( myArea->close  ) {
      CPUPrint("Cluster: Final barrier passed at %lld close=%lld\n",
               (uint64)CPUVec.CycleCount(cpuNum),
               (uint64)myArea->close);  
      
     /* 
      * go back to base mode, simulate a /bdoor/cpuEnter b 
      */
      CPUVec.ExitSimulator(sim_misc.enterThisCPU);
      return;
   }
    
    /*
     * debug cycleCount differences 
     */
    if (barrierMyNum && clusterArea->barrierSync ) { 
      int diff = myArea->cycleCount - clusterArea->simul[0].cycleCount;
      int diff2 = CPUVec.CycleCount(cpuNum) - clusterArea->simul[0].cycleCount;
      if (diff < 0)  diff = -diff;
      if (diff2 < 0) diff2 = -diff2;
      if (diff  > 2*myArea->barrierInterval || 
          diff2 > 2*myArea->barrierInterval ) {
        CPUWarning("%10lld Cycle count diff=%i diff2=%i  when interval is %i \n",
                   (uint64)myArea->cycleCount,diff,diff2,
                   myArea->barrierInterval);
      }
    }
      

   /* 
    * Get received messages 
    */
   if( !receiveMsgCallback.active ) {
       Message *msg = GetMsg( CPUVec.CycleCount(cpuNum));
       if( msg ) { 
           if( msg->deliverTime <  CPUVec.CycleCount(cpuNum) && 
               clusterLatency > myArea->barrierInterval ) { 
               CPUWarning("Cluster: delivering msg at %lld when received at %lld\n",
                          (uint64)CPUVec.CycleCount(cpuNum), 
                          msg->deliverTime);
               EventDoCallback(cpuNum,DeliverMsgCallback,
                               &receiveMsgCallback,(void*)msg, 0);
           } else { 
               EventDoCallback(cpuNum,DeliverMsgCallback,
                               &receiveMsgCallback,(void*)msg,
                                msg->deliverTime -  
                               CPUVec.CycleCount(cpuNum));
           }
       }
   }
   /* 
    * checkpoint ? 
    */
   if (myArea->doCheckpoint ) {
     slaveCheckpoint = 1;
     if( sim_misc.myCPUType == EMBRA_PAGE || 
         sim_misc.myCPUType == EMBRA_CACHE ) { 
        ASSERT( CPUVec.DoCheckpoint );
        CPUVec.DoCheckpoint(0);
        myArea->doCheckpoint = 0;
     } 
     slaveCheckpoint = 0;
   } 

   /* 
    * reinstall event 
    */
   if( myArea->cycleCount < myArea->nextBarrier ) { 
     EventDoCallback(cpuNum,BarrierCallback,&barrierCallback,(void*)0,
                     myArea->nextBarrier - myArea->cycleCount );
   } else { 
     EventDoCallback(cpuNum,BarrierCallback,&barrierCallback,(void*)0,0);
   }
}

void ClusterTclInit(Tcl_Interp *interp)
{
   filename = NULL;
   ParamRegister("PARAM(CLUSTER.File)", (char *)&filename, PARAM_STRING);
   
   ParamRegister("PARAM(CLUSTER.Interval)",
               (char *)&barrierInterval, PARAM_INT);
   
   ParamRegister("PARAM(CLUSTER.NumSimul)",
               (char *)&barrierNumSimul, PARAM_INT);
   
   ParamRegister("PARAM(CLUSTER.Latency)",
               (char *)&clusterLatency, PARAM_INT);
   
   ParamRegister("PARAM(CLUSTER.Barrier)",
               (char *)&barrierP, PARAM_BOOLEAN);
}

void ClusterInit(void)
{
   int fd,i;
   int pid = getpid();
   int pagesize = getpagesize();
   int size = (sizeof(ClusterArea) + pagesize-1) & ~(pagesize-1);
   int myHostCPU;
   unsigned long addr;

   if( clusterArea ) {
       CPUWarning("CLUSTER: already initialized\n");
       return;
   }

   if( !strcmp(filename,"") )
      return;
   
   fd = open(filename,O_RDWR,0);
   if( fd < 0 )
      CPUError("Could not open file '%s' read-write\n",filename);
   lseek(fd,size,SEEK_CUR);  /* grow the file */
   write(fd,"xxxx",4);

   /* Reserve memory in low memory */
   addr = (unsigned long)ZMALLOC(size + pagesize,"Cluster");
   addr = (addr + pagesize-1) & ~(pagesize-1);
   clusterArea = (ClusterArea *) 
       mmap((void*)addr,size,PROT_READ|PROT_WRITE,MAP_SHARED|MAP_FIXED,fd,0);
      
   if( (unsigned long)clusterArea != addr )
       CPUError("CLusterInit: mmap of file %s failed \n",filename);

   ASSERT( CPUVec.CycleCount && CPUVec.CycleCount(0) == 0 );

   CPUPrint("ClusterInit: initialization of PID %i\n",pid);
   LOCK_CLUSTER();
   if( !clusterArea->numInit ) {
       /* 
        * first one here, init common data structures 
        */
       clusterArea->numParticipants = barrierNumSimul;
       clusterArea->freeList = -1;
       clusterArea->numMsgsUsed = 0;
       clusterArea->msgCounter = 1000;
       bzero(clusterArea->barrierArea,sizeof(clusterArea->barrierArea));
       for(i=0;i<barrierNumSimul;i++) { 
         clusterArea->simul[i].hostCPU = -1;
       }
       clusterArea->barrierSync = barrierP;
   } else {
       ASSERT( barrierNumSimul == clusterArea->numParticipants) ;
   }
   barrierMyNum = clusterArea->numInit++;
   CPUWarning("Cluster, PID %i initialized with ID %i\n",pid,barrierMyNum);
   if( clusterArea->numInit > clusterArea->numParticipants ) 
       CPUError("Too many participants in this cluster simulation. \n");

   /* 
    * Detect if 2 processors are on the same real processor and 
    * move to the next best processor.
    */
   /* XXX This is totally bogus and should be fixed.
    */
   myHostCPU = -1;;
   for(i=0;i<barrierNumSimul;i++) { 
     if( myHostCPU == clusterArea->simul[i].hostCPU ) 
       break;
   }

   if (i!=barrierNumSimul) { 
#ifdef notdef
      extern int simcp0MustRun[];
#endif
#ifdef sgi      
      int nProcs = sysmp(MP_NPROCS,0);
#else
      int nProcs = 1;
#endif

      if (nProcs >=   barrierNumSimul) { 
         int max_idle=(myHostCPU ? 0 : 1);
         int j;
         for(j = 0; j < nProcs; j++ ) {
            if( j == myHostCPU ) continue;
#ifdef gone
            if( hostCPUIdleStats[j] > hostCPUIdleStats[max_idle] ) {
               max_idle = j;
            }
#endif
         }
#ifdef gone
         hostCPUIdleStats[max_idle] = 0; /* no reuse */
#endif
#ifdef notdef
         simcp0MustRun[0] = max_idle;
#endif
#ifdef sgi
         (void) sysmp(MP_MUSTRUN, max_idle);
#endif
         myHostCPU = max_idle;
         CPUWarning("Cluster: moved to real cpu %i\n",max_idle);
      } else {
         CPUWarning("Cluster: too many hosts for this machine. Expect poor performance\n");
      }
   }

   /* 
    * Initialize local state in shared area
    */

   myArea = &clusterArea->simul[barrierMyNum];
   myArea->hostCPU = myHostCPU;
   myArea->pid = pid;
   myArea->currentBarrier = 0;
   myArea->nextBarrier = 0;
   myArea->head = -1;
   myArea->cpuClock = CPU_CLOCK;
   myArea->barrierInterval = NanoSecsToCycles(barrierInterval);

   UNLOCK_CLUSTER();

   bzero(myArea->privateLock,sizeof(myArea->privateLock));
   SimetherInitCluster();
   BarrierCallback(0,&barrierCallback,(void*)0);
}


void ClusterClose(void)
{
   int i;
   if( !clusterArea )
      return;
   CPUPrint("CloseMipsyBarrier at %lld close=%lld\n",
            (uint64)CPUVec.CycleCount(0),
            (uint64)myArea->close);

   if( myArea->close )
      return;
   for (i=0;i<barrierNumSimul;i++) { 
      clusterArea->simul[i].close =  clusterArea->simul[i].nextBarrier;
   }
}

/* ***********************************************************
 * ClusterCheckpoint: called from Embra/Driver.c
 * Called when one of the hosts checkpoints. Triggers a checkpoint
 * on all other hosts.
 * Delay the checkpoint so that all participants take it at the
 * next barrier.
 * return 1 to delay it.
 * **********************************************************/

int  ClusterCheckpoint(void) 
{
  int i;
  if( !clusterArea ) return 0 ;    /* do the checkpoint now */
  if( slaveCheckpoint ) return 0; /* do the checkpoint now */
  
  /* set it up */
  for(i=0;i<barrierNumSimul;i++) { 
    clusterArea->simul[i].doCheckpoint = 1;
  } 
  return 1;
} 

/* ************************************************************************
 * ClusterSetEtherAddr
 * Exported to simether. 
 * ************************************************************************/

void ClusterSetEtherAddr( char * addr , int if_num)
{
    int i;
    if( !myArea )   return;
    ASSERT(if_num < ETHER_MAX_CONTROLLERS );
    for(i=0;i<6;i++) { 
        myArea->etherAddr[if_num].etheraddr[i] = addr[i];
    }
}

/* **************************************************************************
 * ClusterSendPacket
 * Exported to simether. Returns 1 if the packet was sent (the etheraddr
 * is known and 0 if it could not be sent (unknown receiver)
 * **************************************************************************/

int ClusterSendPacket( int cpuNum, struct   iovec     *iov, int io_len) 
{
#ifdef sgi
    struct ether_header* etherhdr =  (struct ether_header*)iov[0].iov_base;
    char *destaddr = (char *)&etherhdr->ether_dhost;
    Message *msg;
    int rcv,i,j,k,len=0;
    

    if( !clusterArea ) return 0 ;
    CPUPrint("CLUSTER: looking for dest etheraddr = %i:%i:%i:%i:%i:%i \n",
                 destaddr[0],destaddr[1],destaddr[2],
                 destaddr[3],destaddr[4],destaddr[5]);

    ASSERT( iov[0].iov_len >= sizeof(struct ether_header));
    for(rcv=0;rcv<barrierNumSimul;rcv++) {
        for(j=0;j<ETHER_MAX_CONTROLLERS;j++) { 
            for(k=0;k<6;k++) {
                if( destaddr[k] != clusterArea->simul[rcv].etherAddr[j].etheraddr[k] ) 
                  break;
            }
            if( k == 6 ) break; /* found */
        }
        if( j!=ETHER_MAX_CONTROLLERS ) break;
    }
    if( rcv == barrierNumSimul ) { 
        CPUPrint("CLUSTER: Could not find etheraddr = %i:%i:%i:%i:%i:%i, going through ethersim \n",
                 destaddr[0],destaddr[1],destaddr[2],
                 destaddr[3],destaddr[4],destaddr[5]);
        return 0; /* can't find the etheraddr */
    }
    /* destination is node i */
    CPUPrint("CLUSTER: found at ID=%i contr=%i  etheraddr = %i:%i:%i:%i:%i:%i \n",
             rcv,j,
             clusterArea->simul[rcv].etherAddr[j].etheraddr[0],
             clusterArea->simul[rcv].etherAddr[j].etheraddr[1],
             clusterArea->simul[rcv].etherAddr[j].etheraddr[2],
             clusterArea->simul[rcv].etherAddr[j].etheraddr[3],
             clusterArea->simul[rcv].etherAddr[j].etheraddr[4],
             clusterArea->simul[rcv].etherAddr[j].etheraddr[5]);


 
    msg = ClusterAllocMsg();
    for(i=0;i<io_len;i++) {
        bcopy(iov[i].iov_base,msg->data+len,iov[i].iov_len);
        len += iov[i].iov_len;
    }
    msg->len = len;
    msg->if_num = j;
    ASSERT( len < SIMETHER_MAX_TRANSFER_SIZE );
    SendMsg(cpuNum,msg,rcv);
    return 1;
#else
    CPUPrint("SPARC can't send cluster packets\n");
    return 0;
#endif
}


/* **************************************************************************
 * Utility functions
 * **************************************************************************/
#ifdef sgi
static Message *
ClusterAllocMsg(void)
{
    Message *retval;
    LOCK_CLUSTER();
    if( clusterArea->freeList >= 0 ) { 
        retval = &clusterArea->msgPool[clusterArea->freeList];
        clusterArea->freeList = retval->next;
        retval->next = -1;
    } else { 
        if( clusterArea->numMsgsUsed >= CLUSTER_MAX_MSGS ) { 
            CPUError("Cluster: Message buffer overflow\n");
        } 
        retval = &clusterArea->msgPool[clusterArea->numMsgsUsed];
        retval->msgId = clusterArea->numMsgsUsed++;
        retval->next = -1;
    }
    retval->id = clusterArea->msgCounter++;
    UNLOCK_CLUSTER();
    return retval;
}
#endif

static void  ClusterFreeMsg(Message *msg)
{
    LOCK_CLUSTER();
    msg->next = clusterArea->freeList;
    clusterArea->freeList = msg->msgId;
    UNLOCK_CLUSTER();
}

/* ******************************************************************
 * SendMsg
 * The send and receive time is expressed in receiver clock 
 * cycles and not in local cycle count. This is important 
 * if we do not use barriers or if the clock rates are different
 * on the clients and on the servers
 * *****************************************************************/

#ifdef sgi
static void SendMsg(int cpuNum,Message *msg, int receiver)
{
    /* Propagation delays and preable latencies are ignored.
     * We model the latency as a linear function of the size
     * of the packet.
     */
   static int repeatWarning = 0;
   struct HostArea *destArea = &clusterArea->simul[receiver];
   SimTime sendTime = destArea->currentBarrier + 
      (( CPUVec.CycleCount(cpuNum) - myArea->currentBarrier ) * 
       destArea->cpuClock ) / myArea->cpuClock;
   SimTime latencyNs = clusterLatency * msg->len * 8;
   SimTime latency = (latencyNs * destArea->cpuClock)/1000;
   SimTime deliverTime = sendTime + latency;
   
    if( latency < destArea->barrierInterval &&  !repeatWarning) { 
       CPUWarning("Cluster: Simulation is not repeatable: \n");
       CPUWarning("          id=%i interval=%lld cycles) msg_len =%i  lat=%lld (cycles) \n", 
                  receiver,(uint64)destArea->barrierInterval, msg->len,
                  (uint64)latency);
       repeatWarning = 1;
    }

    msg->sendTime = sendTime; 
    msg->deliverTime = deliverTime;
    msg->source = barrierMyNum;
    msg->dest = receiver;

    ASSERT( (uint)msg < 
       (uint)&clusterArea->msgPool[clusterArea->numMsgsUsed]);

    LOCK_HOST(receiver);
    if( destArea->head < 0 || 
        deliverTime < clusterArea->msgPool[destArea->head].deliverTime  ) { 
        msg->next = destArea->head;
        destArea->head = msg->msgId;
    } else { 
        MessageId tId = destArea->head;
        Message *t = &clusterArea->msgPool[tId];
        while( t->next>=0 && 
               deliverTime >= clusterArea->msgPool[t->next].deliverTime ) {
           tId = t->next;
           t = &clusterArea->msgPool[tId];
        }
        msg->next = t->next;
        t->next = msg->msgId;
    }
    UNLOCK_HOST(receiver);
   LogEntry("CLUSTER-snd",cpuNum,"id=%i dest=%i len=%i \n",
            msg->id,msg->dest,msg->len);
 }
#endif /* sgi */

static Message *GetMsg(SimTime now) 
{
    Message *retval = &clusterArea->msgPool[myArea->head];
    LOCK_HOST(barrierMyNum);
    if( myArea->head>=0  && retval->deliverTime <= now ) {
        myArea->head = retval->next;
    }  else { 
       retval = 0;
    }
    UNLOCK_HOST(barrierMyNum);
    return retval;
}

        
